#' Enrichment & Pathway Analysis Server Module
#' 
#' This server module handles the enrichment and pathway analysis within a Shiny
#' application. Users can select pathway analysis databases and set parameters
#' for the analysis through the UI. This module processes those inputs to perform
#' pathway analysis on gene expression data and visualizes the results. Additionally,
#' it provides options for downloading the pathway analysis results.
#' 
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param deResults A reactive expression containing the results of differential
#'                  expression analysis.
#' @export
enrichmentServer <- function(id, deResults) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    # Gene selection parameters
    output$commonParameters <- renderUI({
      tagList(
        h4("Parameters for gene selection"),
        numericInput(ns("logFCThresholdGenes"), "Log2 Fold Change Cutoff", value = 1),
        numericInput(ns("adjPValThresholdGenes"), "False Discovery Rate (FDR) / Adjusted pValue Cutoff", value = 0.05)
      )
    })
    
    # Gene selection parameters
    output$geneMapOrg <- renderUI({
      selectInput(ns("geneMappingOrg"), "Select organism for gene mapping", 
                  c("Human" = "org.Hs.eg.db", "Mouse" = "org.Mm.eg.db"), 
                  selected = "org.Hs.eg.db")
    })
    
    output$specificParameters <- renderUI({
      analysis <- input$analysisSelector
      
      if (analysis == "GSEA: general interface for enrichment analysis") {
        fluidRow(
          column(
            width = 4,
            uiOutput(ns("commonParameters")),
            h4("Parameters for enrichment analysis"),
            uiOutput(ns("geneMapOrg")),
            bsPopover(ns("geneMapOrg"), "Organism selection", "Select the organism to map gene symbols to Entrez IDs for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gmtSource"), "Select GMT source", 
                        choices = c("Upload GMT file" = "upload", 
                                    "Pre-packaged GMT files" = "PreDownloaded"), 
                        selected = "PreDownloaded"),
            uiOutput(ns("gmtInput")),
            selectInput(ns("gse_padjustMethod"), "P-value adjustment method", 
                        choices = c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none"), 
                        selected = "BH"),
            bsPopover(ns("gse_padjustMethod"), "P-value adjustment method", "Choose the method for adjusting p-values", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gse_adjpvalthreshold"), "Adjusted p-value threshold", value = 0.05),
            bsPopover(ns("gse_adjpvalthreshold"), "Adjusted p-value threshold", "Cutoff for adjusted p-value in enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gse_Minsize"), "Minimum gene set size", value = 10),
            bsPopover(ns("gse_Minsize"), "Minimum gene set size", "Minimum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gse_Maxsize"), "Maximum gene set size", value = 500),
            bsPopover(ns("gse_Maxsize"), "Maximum gene set size", "Maximum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Visualizations"),
            selectInput(ns("gse_visualization"), "Select visualization", choices = c("dotplot"),
                        selected = "dotplot"),
            actionButton(ns("performgseenrichment"), "Perform GSEA enrichment analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0')
          ),
          column(
            width = 8,
            uiOutput(ns("gseplotType")),
            uiOutput(ns("downloadPlotGSEAUI")),
            plotOutput(ns("gse_plot"), width = 800, height = 600),
            uiOutput(ns("downloadResultsGSEAUI")),
            dataTableOutput(ns("table_gse"))
          )
        )
      } else if (analysis == "GSEA: GO enrichment analysis") {
        fluidRow(
          column(
            width = 4,
            uiOutput(ns("commonParameters")),
            uiOutput(ns("clusterSelectInput")),
            bsPopover(ns("clusterSelectInput"), "Cluster selection", "Select the cluster for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Parameters for enrichment analysis"),
            uiOutput(ns("geneMapOrg")),
            bsPopover(ns("geneMapOrg"), "Organism selection", "Select the organism to map gene symbols to Entrez IDs for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gse_OrgDb"), "Select organism for enrichment analysis",
                        choices = c("Human" = "org.Hs.eg.db", "Mouse" = "org.Mm.eg.db"),
                        selected = "org.Hs.eg.db"),
            bsPopover(ns("gse_OrgDb"), "Organism selection", "Select the organism for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gsego_padjustMethod"), "P-value adjustment method", 
                        choices = c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none"), 
                        selected = "BH"),
            bsPopover(ns("gsego_padjustMethod"), "P-value adjustment method", "Choose the method for adjusting p-values", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gsego_adjpvalthreshold"), "Adjusted p-value threshold", value = 0.05),
            bsPopover(ns("gsego_adjpvalthreshold"), "Adjusted p-value threshold", "Cutoff for adjusted p-value in enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gsego_Minsize"), "Minimum gene set size", value = 10),
            bsPopover(ns("gsego_Minsize"), "Minimum gene set size", "Minimum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gsego_Maxsize"), "Maximum gene set size", value = 500),
            bsPopover(ns("gsego_Maxsize"), "Maximum gene set size", "Maximum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("ontology_gseGO"), "Select ontology", choices = c("BP", "MF", "CC", "ALL"),
                        selected = "ALL"),
            bsPopover(ns("ontology_gseGO"), "Ontology", "Select the ontology for gene enrichment analysis. BP stands for Biological Process, MF stands for Molecular Function, CC stands for Cellular Component, and ALL represents all three ontologies combined.", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Visualizations"),
            selectInput(ns("gseGo_visualization"), "Select visualization", choices = c("dotplot"),
                        selected = "dotplot"),
            actionButton(ns("performgseGOenrichment"), "Perform GSEA GO enrichment analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0')
          ),
          column(
            width = 8,
            uiOutput(ns("gsegoplotType")),
            uiOutput(ns("downloadPlotGSEGOUI")),
            plotOutput(ns("gseGO_plot"), width = 800, height = 600),
            uiOutput(ns("downloadResultsGSEGOUI")),
            dataTableOutput(ns("table_gse_go"))
          )
        )
      } else if (analysis == "GSEA: WikiPathways enrichment analysis") {
        fluidRow(
          column(
            width = 4,
            uiOutput(ns("commonParameters")),
            uiOutput(ns("clusterSelectInput")),
            bsPopover(ns("clusterSelectInput"), "Cluster selection", "Select the cluster for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Parameters for Enrichment Analysis"),
            uiOutput(ns("geneMapOrg")),
            bsPopover(ns("geneMapOrg"), "Organism selection", "Select the organism to map gene symbols to Entrez IDs for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gse_Wiki_Org"), "Select organism for enrichment analysis",
                        choices = c("Human" = "Homo sapiens", "Mouse" = "Mus musculus"),
                        selected = "Homo sapiens"),
            bsPopover(ns("gse_Wiki_Org"), "Organism selection", "Select the organism for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gseWiki_padjustMethod"), "P-value adjustment method", 
                        choices = c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none"), 
                        selected = "BH"),
            bsPopover(ns("gseWiki_padjustMethod"), "P-value adjustment method", "Choose the method for adjusting p-values", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gseWiki_adjpvalthreshold"), "Adjusted p-value threshold", value = 0.05),
            bsPopover(ns("gseWiki_adjpvalthreshold"), "Adjusted p-value threshold", "Cutoff for adjusted p-value in enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gseWiki_Minsize"), "Minimum gene set size", value = 10),
            bsPopover(ns("gseWiki_Minsize"), "Minimum gene set size", "Minimum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gseWiki_Maxsize"), "Maximum gene set size", value = 500),
            bsPopover(ns("gseWiki_Maxsize"), "Maximum gene set size", "Maximum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Visualizations"),
            selectInput(ns("gseWP_visualization"), "Select visualization", choices = c("dotplot"),
                        selected = "dotplot"),
            actionButton(ns("performgsewpenrichment"), "Perform GSEA WikiPathways enrichment analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0')
          ),
          column(
            width = 8,
            uiOutput(ns("gseWPplotType")),
            uiOutput(ns("downloadPlotGSEWIKIUI")),
            plotOutput(ns("gseWP_plot"), width = 800, height = 600),
            uiOutput(ns("downloadResultsGSEWIKIUI")),
            dataTableOutput(ns("table_gseWiki"))
          )
        )
      } else if (analysis == "GSEA: Reactome pathway enrichment analysis") {
        fluidRow(
          column(
            width = 4,
            uiOutput(ns("commonParameters")),
            uiOutput(ns("clusterSelectInput")),
            bsPopover(ns("clusterSelectInput"), "Cluster selection", "Select the cluster for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Parameters for Enrichment Analysis"),
            uiOutput(ns("geneMapOrg")),
            bsPopover(ns("geneMapOrg"), "Organism selection", "Select the organism to map gene symbols to Entrez IDs for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gse_RE_Org"), "Select organism for enrichment analysis",
                        choices = c("Human" = "human", "Mouse" = "mouse"),
                        selected = "human"),
            bsPopover(ns("gse_RE_Org"), "Organism selection", "Select the organism for pathway/gene set enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("gseRE_padjustMethod"), "P-value adjustment method", 
                        choices = c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none"), 
                        selected = "BH"),
            bsPopover(ns("gseRE_padjustMethod"), "P-value adjustment method", "Choose the method for adjusting p-values", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gseRE_adjpvalthreshold"), "Adjusted p-value threshold", value = 0.05),
            bsPopover(ns("gseRE_adjpvalthreshold"), "Adjusted p-value threshold", "Cutoff for adjusted p-value in enrichment analysis", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gseRE_Minsize"), "Minimum gene set size", value = 10),
            bsPopover(ns("gseRE_Minsize"), "Minimum gene set size", "Minimum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            numericInput(ns("gseRE_Maxsize"), "Maximum gene set size", value = 500),
            bsPopover(ns("gseRE_Maxsize"), "Maximum gene set size", "Maximum number of genes required in a gene set", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            h4("Visualizations"),
            selectInput(ns("gseRE_visualization"), "Select visualization", choices = c("dotplot"),
                        selected = "dotplot"),
            actionButton(ns("performgseREenrichment"), "Perform GSEA reactome pathway enrichment analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0')
          ),
          column(
            width = 8,
            uiOutput(ns("gseREplotType")),
            uiOutput(ns("downloadPlotGSEReactomeUI")),
            plotOutput(ns("gseRE_plot"), width = 800, height = 600),
            uiOutput(ns("downloadResultsGSEREUI")),
            dataTableOutput(ns("table_gseRE"))
          )
        )
      }
    })
    
    # GMT input selection
    output$gmtInput <- renderUI({
      if (input$analysisSelector == "GSEA: general interface for enrichment analysis") {
        if (input$gmtSource == "upload") {
          tagList(
            fileInput(ns("file"), "Upload GMT file", accept = ".gmt"),
            selectInput(ns("TermID"), "Select term ID", choices = NULL),
            bsPopover(ns("TermID"), "Select term ID Column", "Select a column from GMT file representing term", 
                      placement = "right", trigger = "hover", options = list(container = "body")),
            selectInput(ns("GeneID"), "Select gene ID", choices = NULL),
            bsPopover(ns("GeneID"), "Select gene ID Column", "Select a column from GMT file representing gene", 
                      placement = "right", trigger = "hover", options = list(container = "body"))
          )
        } else if (input$gmtSource == "PreDownloaded") {
          req(input$geneMappingOrg)
          if (input$geneMappingOrg == "org.Hs.eg.db") {
            selectInput(ns("gmtCategory"), "Select GMT category",
                        choices = c("H: hallmark gene sets",
                                    "C1: positional gene sets",
                                    "C2: curated gene sets",
                                    "C3: regulatory target gene sets",
                                    "C4: computational gene sets",
                                    "C5: ontology gene sets",
                                    "C6: oncogenic signature gene sets",
                                    "C7: immunologic signature gene sets",
                                    "C8: cell type signature gene sets"))
          } else if (input$geneMappingOrg == "org.Mm.eg.db") {
            selectInput(ns("gmtCategory"), "Select GMT category",
                        choices = c("MH: hallmark gene sets",
                                    "M1: positional gene sets",
                                    "M2: curated gene sets",
                                    "M3: regulatory target gene sets",
                                    "M5: ontology gene sets",
                                    "M8: cell type signature gene sets"))
          }
        }
      }
    })
    
    gmt_data <- reactiveVal()
    
    observeEvent(input$file, {
      req(input$gmtSource == "upload")
      req(input$file)
      if (grepl(".gmt$", input$file$name)) {
        gmt <- clusterProfiler::read.gmt(input$file$datapath)
        columns_gmt <- colnames(gmt)
        updateSelectInput(session, "TermID", choices = columns_gmt)
        updateSelectInput(session, "GeneID", choices = columns_gmt)
        gmt_data(gmt)
      } else {
        stop("Unsupported file type")
      }
    })
    
    observeEvent(input$gmtCategory, {
      req(input$gmtSource == "PreDownloaded")
      req(input$geneMappingOrg)
      
      gmt_folder <- if (input$geneMappingOrg == "org.Hs.eg.db") {
        "data/GMT_Files/Human/"
      } else {
        "data/GMT_Files/Mouse/"
      }
      
      gmt_file <- switch(input$gmtCategory,
                         "H: hallmark gene sets" = "h.all.v2023.2.Hs.entrez.gmt",
                         "C1: positional gene sets" = "c1.all.v2023.2.Hs.entrez.gmt",
                         "C2: curated gene sets" = "c2.all.v2023.2.Hs.entrez.gmt",
                         "C3: regulatory target gene sets" = "c3.all.v2023.2.Hs.entrez.gmt",
                         "C4: computational gene sets" = "c4.all.v2023.2.Hs.entrez.gmt",
                         "C5: ontology gene sets" = "c5.all.v2023.2.Hs.entrez.gmt",
                         "C6: oncogenic signature gene sets" = "c6.all.v2023.2.Hs.entrez.gmt",
                         "C7: immunologic signature gene sets" = "c7.all.v2023.2.Hs.entrez.gmt",
                         "C8: cell type signature gene sets" = "c8.all.v2023.2.Hs.entrez.gmt",
                         "MH: hallmark gene sets" = "mh.all.v2023.2.Mm.entrez.gmt",
                         "M1: positional gene sets" = "m1.all.v2023.2.Mm.entrez.gmt",
                         "M2: curated gene sets" = "m2.all.v2023.2.Mm.entrez.gmt",
                         "M3: regulatory target gene sets" = "m3.all.v2023.2.Mm.entrez.gmt",
                         "M5: ontology gene sets" = "m5.all.v2023.2.Mm.entrez.gmt",
                         "M8: cell type signature gene sets" = "m8.all.v2023.2.Mm.entrez.gmt")
      
      gmt <- clusterProfiler::read.gmt(file.path(gmt_folder, gmt_file))
      gmt_data(gmt)
    })
    
    # GSEA: General Interface for Enrichment Analysis
    output$downloadPlotGSEAUI <- renderUI({
      if (!is.null(gse_plot())) {
        selectInput(ns("gseplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")
        downloadButton(ns('downloadPlot_gsea'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    output$gseplotType <- renderUI({
      if (!is.null(gse_plot())) {
        selectInput(ns("gseplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")
      }
    })
    
    output$downloadResultsGSEAUI <- renderUI({
      if (!is.null(gse_results()) && nrow(summary(gse_results())) > 0) {
        downloadButton(ns('download_gse'), 'Download results', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # GSEA: GO Enrichment Analysis
    output$downloadPlotGSEGOUI <- renderUI({
      if (!is.null(gsego_plot())) {
        downloadButton(ns('downloadPlot_gsego'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # GSEA: GO Enrichment Analysis
    output$gsegoplotType <- renderUI({
      if (!is.null(gsego_plot())) {
        selectInput(ns("gsegoplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")      }
    })
    
    output$downloadResultsGSEGOUI <- renderUI({
      if (!is.null(gsego_results()) && nrow(summary(gsego_results())) > 0) {
        downloadButton(ns('download_gse_go'), 'Download results', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # GSEA: WikiPathways Enrichment Analysis
    output$downloadPlotGSEWIKIUI <- renderUI({
      if (!is.null(gseWP_plot())) {
        downloadButton(ns('downloadPlot_gseWiki'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    output$gseWPplotType <- renderUI({
      if (!is.null(gseWP_plot())) {
        selectInput(ns("gseWPplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")
      }
    })
    
    output$downloadResultsGSEWIKIUI <- renderUI({
      if (!is.null(gseWP_results()) && nrow(summary(gseWP_results())) > 0) {
        downloadButton(ns('download_gseWiki'), 'Download results', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Reactome Pathway Enrichment Analysis
    output$downloadPlotGSEReactomeUI <- renderUI({
      if (!is.null(gseRE_plot())) {
        downloadButton(ns('downloadPlot_gseRE'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Reactome Pathway Enrichment Analysis
    output$gseREplotType <- renderUI({
      if (!is.null(gseRE_plot())) {
        selectInput(ns("gseREplot"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")
      }
    })
    
    output$downloadResultsGSEREUI <- renderUI({
      if (!is.null(gseRE_results()) && nrow(summary(gseRE_results())) > 0) {
        downloadButton(ns('download_gseRE'), 'Download results', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Initialize reactiveVal for MappedGenesList
    MappedGenesList <- reactiveVal(NULL)
    de_results <- reactive({
      res = deResults()[[3]]
      res = data.frame(res)
      return(res)
    })
    
    observe({
      if (is.null(de_results()) || is.null(de_results()$log2FoldChange) || is.null(de_results()$padj)) {
        MappedGenesList(NULL)
      } else {
        tryCatch({
          req(de_results())
          results <- de_results()
          
          #filtered_genes <- results[abs(as.numeric(results$log2FoldChange)) > input$logFCThresholdGenes & as.numeric(results$padj) < input$adjPValThresholdGenes, ]
          res_out <-  results %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFCThresholdGenes | log2FoldChange <= -1*(input$logFCThresholdGenes), padj <= input$adjPValThresholdGenes) %>% column_to_rownames('gene')
          #print(res_out)
          geneList <- rownames(res_out)
          
          checkIDs = grep(pattern = "^ENS", x = rownames(res_out), value = T)
          checkIDs_dot = NULL
          if(length(checkIDs) == length(rownames(res_out))) {
            message("Ensembl IDs found")
            #checkIDs_dot = grep(".[1.9]", x = rownames(res_out), value = T)
            checkIDs_dot = grep("\\.[1-9]", x = rownames(res_out), value = TRUE)
            if(length(checkIDs_dot) == length(rownames(res_out))) {
              id_split = strsplit(rownames(res_out), split = ".", fixed = T)
              final_ids = sapply(id_split, "[[", 1, simplify = TRUE)
              checkIDs_dot = data.frame(checkIDs_dot)
              checkIDs_dot$ENSEMBL = final_ids
              colnames(checkIDs_dot) = c("ID", "ENSEMBL")
              ids_use = clusterProfiler::bitr(final_ids, fromType = "ENSEMBL", toType = "ENTREZID", OrgDb = input$geneMappingOrg)
            } else {
              checkIDs_dot = NULL
              ids_use = clusterProfiler::bitr(rownames(res_out),fromType = "ENSEMBL", toType = "ENTREZID", OrgDb = input$geneMappingOrg)
            }
          } else {
            ids_use = clusterProfiler::bitr(rownames(res_out),fromType = "SYMBOL", toType = "ENTREZID", OrgDb = input$geneMappingOrg)
          }
          
          if(!is.null(ids_use)) {
            if((nrow(checkIDs_dot) > 0) && (!is.null(checkIDs_dot))) {
              checkIDs_dot_filtered = checkIDs_dot %>% dplyr::filter(ENSEMBL %in% ids_use[,1])
              rownames(checkIDs_dot_filtered) = checkIDs_dot_filtered$ENSEMBL
              checkIDs_dot_filtered = checkIDs_dot_filtered[ids_use[,1],]
              ids_use$Original = checkIDs_dot_filtered$ID
              res_out_pvlaue = data.frame(res_out) %>% dplyr::select(log2FoldChange)
              res_out_pvlaue = data.frame(res_out_pvlaue) %>% dplyr::filter(rownames(res_out_pvlaue) %in% ids_use[,3])
              res_out_pvlaue$ID = rownames(res_out_pvlaue)
              res_out_pvlaue = res_out_pvlaue[ids_use[,1],]
              ids_use$log2FoldChange = res_out_pvlaue$log2FoldChange
              ids_use = arrange(ids_use, desc(log2FoldChange))
              ordered_matched_genes = ids_use %>% dplyr::select(ENTREZID, log2FoldChange)
              ordered_matched_genes = deframe(ordered_matched_genes)
            } else {
              res_out_pvlaue = res_out %>% dplyr::select(log2FoldChange)
              res_out_pvlaue = res_out_pvlaue %>% dplyr::filter(rownames(res_out_pvlaue) %in% ids_use[,1])
              res_out_pvlaue$ID = rownames(res_out_pvlaue)
              res_out_pvlaue = res_out_pvlaue[ids_use[,1],]
              ids_use$log2FoldChange = res_out_pvlaue$log2FoldChange
              ids_use = arrange(ids_use, desc(log2FoldChange))
              ordered_matched_genes = ids_use %>% dplyr::select(ENTREZID, log2FoldChange)
              ordered_matched_genes = deframe(ordered_matched_genes)
            }
          }
          ####### Smriti's code 
          #convertedList <- clusterProfiler::bitr(geneID = geneList, fromType = "SYMBOL", toType = "ENTREZID", OrgDb = input$geneMappingOrg)
          #geneList <- convertedList$ENTREZID
          #matched_genes <- merge(filtered_genes, convertedList, by.x = "row.names", by.y = "SYMBOL", all.x = TRUE)
          #matched_genes <- matched_genes[!is.na(matched_genes$ENTREZID), ]
          #colnames(matched_genes)[1] <- "Symbols"
          #ordered_matched_genes <- matched_genes[order(matched_genes$log2FoldChange, decreasing = TRUE), ]
          #################
          
          MappedGenesList(ordered_matched_genes)
        }, error = function(e) {
          MappedGenesList(NULL)
        })
      }
    })
    
    # GSEA: General Interface for Enrichment Analysis
    gse_plot <- reactiveVal(NULL)
    gse_results <- reactiveVal(NULL)
    
    shiny::observeEvent(input$performgseenrichment, {
      
      if (is.null(gmt_data())) {
        shinyalert("Please upload a GMT file to proceed with the enrichment analysis.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (is.null(MappedGenesList())) {
        shinyalert("Please check input organism or run DE Analysis to proceed with the enrichment analysis.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(MappedGenesList())
      req(gmt_data())
      
      #gene_list <- MappedGenesList()
      
      gene_list <- MappedGenesList()
      
      term2gene <- if (input$gmtSource == "upload") {
        gmt_data()[, c(input$TermID, input$GeneID)]
      } else {
        gmt_data()
      }
      
      withProgress(
        message = 'Performing Pathway Analysis', 
        value = 0, {
          incProgress(0.5, detail = "Performing GSEA Enrichment Analysis")
          
          tryCatch({
            set.seed(12345)
            gse <- clusterProfiler::GSEA(gene = gene_list,
                                         exponent = 1,
                                         pvalueCutoff = input$gse_adjpvalthreshold,
                                         pAdjustMethod = input$gse_padjustMethod,
                                         eps = 1e-10,
                                         minGSSize = input$gse_Minsize,
                                         maxGSSize = input$gse_Maxsize,
                                         TERM2GENE = term2gene,
                                         TERM2NAME = NA,
                                         verbose = TRUE,
                                         seed = TRUE,
                                         by = "fgsea")
            gse_results(gse)
            print(gse_results())
            if (nrow(summary(gse)) == 0) {
              shinyalert("No Enrichment found for the selected parameters. Please adjust your parameters and try again.", type = "warning", confirmButtonCol = "#1c82e0")
            } else {
              plot <- if (input$gse_visualization == "dotplot") {
                if (input$gse_padjustMethod == "none") {
                  dotplot(gse, color = "pvalue", showCategory = 25)
                }
                else {
                  dotplot(gse, showCategory = 25)
                }
              }
              
              gse_plot(plot)
              output$gse_plot <- renderPlot({
                gse_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
              })
              
              output$downloadPlot_gsea <- downloadHandler(
                filename = function() {
                  paste0("CogentDS_v2.0_EnrichmentPlot_", Sys.Date(), input$gseplot)
                },
                content = function(file) {
                  gse_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
                  ggsave(file, plot = gse_plot() + ggtitle("GSEA Gene Set Enrichment Plot"), width = 10, height = 8)
                },
                contentType = 'image'
              )
              
              # Render table for results
              output$table_gse <- renderDataTable({
                datatable(as.data.frame(gse_results()@result), 
                          options = list(pageLength = 5, 
                                         scrollX = T,
                                         autoWidth = T))
              })
              
              output$download_gse <- downloadHandler(
                filename = function() { paste("CogentDS_v2.0_pathway_results_", Sys.Date(), ".csv", sep = "") },
                content = function(file) {
                  if (!is.null(gse_results()) && nrow(summary(gse_results())) > 0) {
                    write.csv(as.data.frame(gse_results()@result), file, row.names = T)
                  }
                }
              )
              
              shinyalert("GSEA Enrichment Analysis Completed", type = "success", confirmButtonCol = "#1c82e0")
            }
          }, error = function(e) {
            shinyalert("An error occurred during Enrichment Analysis. Please check your inputs and try again.", type = "error", confirmButtonCol = "#1c82e0")
          })
          
          incProgress(0.5, detail = "Finalizing...")
        }
      )
    })
    
    # GSEA: GO Enrichment Analysis
    gsego_plot <- reactiveVal(NULL)
    gsego_results <- reactiveVal(NULL)
    
    shiny::observeEvent(input$performgseGOenrichment, {
      
      if (is.null(MappedGenesList())) {
        shinyalert("Please check input organism or run DE Analysis to proceed with the enrichment analysis.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(MappedGenesList())
      
      gene_list <- MappedGenesList()
      
      withProgress(
        message = 'Performing Pathway Analysis', 
        value = 0, {
          incProgress(0.5, detail = "Performing GSEA GO Enrichment Analysis")
          
          tryCatch({
            set.seed(12345)
            gse_go <- clusterProfiler::gseGO(geneList = gene_list,
                                             ont = input$ontology_gseGO,
                                             OrgDb = input$gse_OrgDb,
                                             keyType = "ENTREZID",
                                             exponent = 1,
                                             minGSSize = input$gsego_Minsize,
                                             maxGSSize = input$gsego_Maxsize,
                                             eps = 1e-10,
                                             pvalueCutoff = input$gsego_adjpvalthreshold,
                                             pAdjustMethod = input$gsego_padjustMethod,
                                             verbose = TRUE,
                                             seed = TRUE,
                                             by = "fgsea")
            
            gsego_results(gse_go)
            if (nrow(summary(gse_go)) == 0) {
              shinyalert("No Enrichment found for the selected parameters. Please adjust your parameters and try again.", type = "warning", confirmButtonCol = "#1c82e0")
            } else {
              plot <- if (input$gseGo_visualization == "dotplot") {
                if (input$gsego_padjustMethod == "none") {
                  dotplot(gse_go, color = "pvalue", showCategory = 25)
                  
                } else{
                  dotplot(gse_go, showCategory = 25)
                }
              }
              
              gsego_plot(plot)
              
              output$gseGO_plot <- renderPlot({
                gsego_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
              })
              
              output$downloadPlot_gsego <- downloadHandler(
                filename = function() {
                  paste0("CogentDS_v2.0_EnrichmentPlot_", Sys.Date(), input$gsegoplot)
                },
                content = function(file) {
                  gsego_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
                  ggsave(file, plot = gsego_plot(), width = 10, height = 8)
                },
                contentType = 'image'
              )
              
              # Render table for results
              output$table_gse_go <- renderDataTable({
                datatable(as.data.frame(gsego_results()@result), 
                          options = list(pageLength = 5, 
                                         scrollX = T,
                                         autoWidth = T))
              })
              
              output$download_gse_go <- downloadHandler(
                filename = function() { paste("CogentDS_v2.0_pathway_results_", Sys.Date(), ".csv", sep = "") },
                content = function(file) {
                  if (!is.null(gsego_results()) && nrow(summary(gsego_results())) > 0) {
                    write.csv(as.data.frame(gsego_results()@result), file, row.names = T)
                  }
                }
              )
              
              shinyalert("GSEA GO Enrichment Analysis Completed", type = "success", confirmButtonCol = "#1c82e0")
            }
          }, error = function(e) {
            shinyalert("An error occurred during Enrichment Analysis. Please check your inputs and try again.", type = "error", confirmButtonCol = "#1c82e0")
          })
          
          incProgress(0.5, detail = "Finalizing...")
        }
      )
    })
    
    # GSEA: WikiPathways Enrichment Analysis
    gseWP_plot <- reactiveVal(NULL)
    gseWP_results <- reactiveVal(NULL)
    
    shiny::observeEvent(input$performgsewpenrichment, {
      
      if (is.null(MappedGenesList())) {
        shinyalert("Please check input organism or run DE Analysis to proceed with the enrichment analysis.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(MappedGenesList())
      
      gene_list <- MappedGenesList()
      
      withProgress(
        message = 'Performing Pathway Analysis', 
        value = 0, {
          incProgress(0.5, detail = "Performing GSEA WikiPathway Enrichment Analysis")
          
          tryCatch({
            set.seed(12345)
            gse_wp <- clusterProfiler::gseWP(gene = gene_list,
                                             exponent = 1,
                                             organism = input$gse_Wiki_Org,
                                             pvalueCutoff = input$gseWiki_adjpvalthreshold,
                                             pAdjustMethod = input$gseWiki_padjustMethod,
                                             eps = 1e-10,
                                             minGSSize = input$gseWiki_Minsize,
                                             maxGSSize = input$gseWiki_Maxsize,
                                             gson = NULL,
                                             verbose = TRUE,
                                             seed = TRUE,
                                             by = "fgsea")
            
            gseWP_results(gse_wp)
            if (nrow(summary(gse_wp)) == 0) {
              shinyalert("No Enrichment found for the selected parameters. Please adjust your parameters and try again.", type = "warning", confirmButtonCol = "#1c82e0")
            } else {
              plot <- if (input$gseWP_visualization == "dotplot") {
                if (input$gseWiki_padjustMethod == "none") {
                  dotplot(gse_wp, color = "pvalue", showCategory = 25) 
                } else {  
                  dotplot(gse_wp, showCategory = 25) 
                }
              }
              
              gseWP_plot(plot)
              
              output$gseWP_plot <- renderPlot({
                gseWP_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
              })
              
              output$downloadPlot_gseWiki <- downloadHandler(
                filename = function() {
                  paste0("CogentDS_v2.0_EnrichmentPlot_", Sys.Date(), input$gseWPplot)
                },
                content = function(file) {
                  gseWP_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
                  ggsave(file, plot = gseWP_plot() + ggtitle("GSEA Gene Set Enrichment Plot"), width = 10, height = 8)
                },
                contentType = 'image'
              )
              
              # Render table for results
              output$table_gseWiki <- renderDataTable({
                datatable(as.data.frame(gseWP_results()@result), 
                          options = list(pageLength = 5, 
                                         scrollX = T,
                                         autoWidth = T))
              })
              
              output$download_gseWiki <- downloadHandler(
                filename = function() { paste("CogentDS_v2.0_pathway_results_", Sys.Date(), ".csv", sep = "") },
                content = function(file) {
                  if (!is.null(gseWP_results()) && nrow(summary(gseWP_results())) > 0) {
                    write.csv(as.data.frame(gseWP_results()@result), file, row.names = T)
                  }
                }
              )
              
              shinyalert("GSEA WikiPathways Enrichment Analysis Completed", type = "success", confirmButtonCol = "#1c82e0")
            }
          }, error = function(e) {
            shinyalert("An error occurred during Enrichment Analysis. Please check your inputs and try again.", type = "error", confirmButtonCol = "#1c82e0")
          })
          
          incProgress(0.5, detail = "Finalizing...")
        }
      )
    })
    
    # Reactome pathway enrichment Enrichment Analysis
    gseRE_plot <- reactiveVal(NULL)
    gseRE_results <- reactiveVal(NULL)
    
    shiny::observeEvent(input$performgseREenrichment, {
      
      if (is.null(MappedGenesList())) {
        shinyalert("Please check input organism or run DE Analysis to proceed with the enrichment analysis.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(MappedGenesList())
      
      gene_list <- MappedGenesList()
      
      withProgress(
        message = 'Performing Pathway Analysis', 
        value = 0, {
          incProgress(0.5, detail = "Performing GSEA Reactome Pathway Enrichment Analysis")
          
          tryCatch({
            set.seed(12345)
            gse_re <- ReactomePA::gsePathway(gene = gene_list,
                                             exponent = 1,
                                             organism = input$gse_RE_Org,
                                             pvalueCutoff = input$gseRE_adjpvalthreshold,
                                             pAdjustMethod = input$gseRE_padjustMethod,
                                             eps = 1e-10,
                                             minGSSize = input$gseRE_Minsize,
                                             maxGSSize = input$gseRE_Maxsize,
                                             verbose = TRUE,
                                             seed = TRUE,
                                             by = "fgsea")
            
            gseRE_results(gse_re)
            if (nrow(summary(gse_re)) == 0) {
              shinyalert("No Enrichment found for the selected parameters. Please adjust your parameters and try again.", type = "warning", confirmButtonCol = "#1c82e0")
            } else {
              plot <- if (input$gseRE_visualization == "dotplot") {
                if (input$gseRE_padjustMethod == "none") {
                  dotplot(gse_re, color = "pvalue", showCategory = 25)
                } else {
                  dotplot(gse_re, showCategory = 25)
                }
              }
              
              gseRE_plot(plot)
              
              output$gseRE_plot <- renderPlot({
                gseRE_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
              })
              
              output$downloadPlot_gseRE <- downloadHandler(
                filename = function() {
                  paste0("CogentDS_v2.0_EnrichmentPlot_", Sys.Date(), input$gseREplot)
                },
                content = function(file) {
                  gseRE_plot() + ggtitle("GSEA Gene Set Enrichment Plot")
                  ggsave(file, plot = gseRE_plot() + ggtitle("GSEA Gene Set Enrichment Plot"), width = 10, height = 8)
                },
                contentType = 'image'
              )
              
              # Render table for results
              output$table_gseRE <- renderDataTable({
                datatable(as.data.frame(gseRE_results()@result), 
                          options = list(pageLength = 5, 
                                         scrollX = T,
                                         autoWidth = T))
              })
              
              output$download_gseRE <- downloadHandler(
                filename = function() { paste("CogentDS_v2.0_pathway_results_", Sys.Date(), ".csv", sep = "") },
                content = function(file) {
                  if (!is.null(gseRE_results()) && nrow(summary(gseRE_results())) > 0) {
                    write.csv(as.data.frame(gseRE_results()@result), file, row.names = T)
                  }
                }
              )
              
              shinyalert("GSEA Reactome Enrichment Analysis Completed", type = "success", confirmButtonCol = "#1c82e0")
            }
          }, error = function(e) {
            shinyalert("An error occurred during Enrichment Analysis. Please check your inputs and try again.", type = "error", confirmButtonCol = "#1c82e0")
          })
          
          incProgress(0.5, detail = "Finalizing...")
        }
      )
    })
  })
}