#' Data Upload Server Component
#'  
#' This server module handles the server-side logic for data upload,
#' including reading the uploaded file, creating a Seurat object,
#' and storing it in a reactive value.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the Seurat object created from
#'         the uploaded data.
#'         
#' Detailed Parameter Description:
#' - counts: A matrix of gene expression data where rows are genes and columns are cells/samples.
#' - project: Name of the project.
#' - min.cells: Minimum number of cells in which a gene must be detected to be kept in the dataset.
#' - min.features: Minimum number of features (genes) detected in a cell for it to be included in the analysis.
#'
#' @seealso \code{\link[Seurat]{Createdf_rds}}, \code{\link[Seurat]{PercentageFeatureSet}}
#' for details on Seurat object creation and estimation of % of all counts that belong to set of features. 
#'         
#' @export
#'
#' @examples
#' # In the server definition of a Shiny app
dataUploadServer <- function(id) {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns
    data <- reactiveVal(NULL)
    metadata_Collect <- reactiveVal(NULL)
    counts_data <- reactiveVal(NULL)
    AP_object <- reactiveVal(NULL)

    observe({
      req(input$file, input$fileoption)
      #metadata_check = read.csv(input$metadata_file$datapath, header = T, check.names=F)
      if(input$fileoption == 'Raw counts Matrix') {
        if(!is.null(input$file)) {
          if(grepl(".csv$", input$file$name)) {
            mat = read.csv(input$file$datapath, header = T, row.names = 1, check.names = F)
            df_rds = CreateSeuratObject(mat)
            counts_data(mat)
            AP_object(df_rds)
          } else {
            shinyalert("Require .csv format", "Please upload a .csv file with genes and samples", type = "error", confirmButtonCol = "#1c82e0")
          }
        } 
      }
      if(input$fileoption == 'Processed Data from CogentAP') {
        if(!is.null(input$file)) {
          if(grepl(".rds$", input$file$name)) {
            df_rds = readRDS(input$file$datapath)
            mat = df_rds[['RNA']]$counts
            counts_data(mat)
            AP_object(df_rds)
          } else {
            shinyalert("Require .rds format", "Please upload a .rds file", type = "error", confirmButtonCol = "#1c82e0")
          }
        }
      }
    })

    observe({
      req(counts_data(), AP_object(), input$metadata_file)
      metadata_check = read.csv(input$metadata_file$datapath, header = TRUE, check.names = F)
      if((nrow(metadata_check) > 0) && (nrow(metadata_check) <= ncol(counts_data()))) {
        if(any(is.na(metadata_check[,1]))) {
          metadata = NULL
          metadata_Collect(metadata)
          shinyalert("In Uploaded metadata First column has NAs", "Please remove NAs! Proceeding without uploaded metadata", type = "warning", confirmButtonCol = "#1c82e0")
        } else {
          metadata = read.csv(input$metadata_file$datapath, header = T, check.names=F, row.names = 1)
          if(!is.null(colnames(metadata))) {
            updateSelectInput(session, "condition_col", label = "Select Condition Column", choices = colnames(metadata), selected = NULL)
            metadata_Collect(metadata)
          }
        }
      } else {
        shinyalert("Uploaded metadata has more samples than counts data", "Please correct the metadata", type = "error", confirmButtonCol = "#1c82e0")
      }
    })

    shiny::observeEvent(input$upload, {
      shinyjs::disable("upload")
        timeStart <- Sys.time()
        mat = NULL
        if(input$fileoption == "Raw counts Matrix") {
          req(counts_data())
          tryCatch({
            if(!is.null(counts_data())) {
              mat = counts_data()
            } else {
              stop("Error file not Uoloaded!")
            }
            if (!is.null(mat)) {
              df_rds = CreateSeuratObject(mat)
              df_rds[["percent.mt"]] <- Seurat::PercentageFeatureSet(df_rds, pattern = "^MT-")
              output$bulkRNA_mat = DT::renderDataTable({
                avg_ngene = cbind('Average nFeature_RNA', formatC(round(mean(df_rds@meta.data$nFeature_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                avg_nCount = cbind('Average nCount_RNA', formatC(round(mean(df_rds@meta.data$nCount_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                #avg_percentmt = cbind('Average percent-mt', formatC(round(mean(df_rds@meta.data$percent.mt, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                numberofsamples = cbind('Number of Samples', formatC(as.numeric(ncol(df_rds)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                numberofgenes = cbind('Number of Genes', formatC(as.numeric(nrow(df_rds)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                df_show = rbind(numberofsamples, numberofgenes, avg_ngene, avg_nCount)
                df_show = data.frame(df_show)
                colnames(df_show) = c("Stats", "Value")
                df_show = data.frame(df_show)
                df_show[df_show == ""] <- NA
                df_show[df_show == "NaN"] <- NA
                df_show[df_show == 0] <- NA
                df_show = data.frame(df_show)
                DT::datatable(df_show, options = list(pageLength = 20, dom = "t"))
              })
            }
            if(!is.null(metadata_Collect())) {
              metadata = metadata_Collect()

              if(any(duplicated(colnames(metadata))) == TRUE) {
                metadata = subset(metadata, select=which(!duplicated(colnames(metadata))))
              } else {
                metadata = metadata
              }

              cols_not_needed_final = c('orig.ident', 'RNA_snn_res.0.8', 'seurat_clusters')
              find_not_needed_final = which(colnames(metadata) %in% cols_not_needed_final)
              if(length(find_not_needed_final) > 0) {
                metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed_final])
              }

              columns_remove = c()
              for(x in 1:length(colnames(metadata)))
              {
                number_of_NA = sum(is.na(metadata[[colnames(metadata)[x]]]))
                if(number_of_NA == length(metadata[[colnames(metadata)[x]]])) {
                  getcol = colnames(metadata)[x]
                  columns_remove = c(columns_remove, getcol)
                }
              }
              metadata = metadata %>% dplyr::select(-columns_remove)

              mat = mat[,rownames(metadata)]
              column_name <-  paste("~", input$condition_col)
              length_condition = length(unique(metadata[[input$condition_col]]))
              if(length_condition > 1) {
                condition_col_use = input$condition_col
                dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design= as.formula(column_name))
              } else {
                metadata$Samples_Found = rownames(metadata)
                metadata$conditionAddedbyCogentDS = rep("NoConditionSpecified", nrow(metadata))
                condition_col_use = 'conditionAddedbyCogentDS'
                dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design = ~1)
              }
              
              datalist = list(dds, metadata, input$countfilter, condition_col_use)
              data(datalist)
            } else {
              metadata = data.frame(colnames(mat))
              colnames(metadata) = "Samples_Found"
              rownames(metadata) = colnames(mat)

              if(any(duplicated(colnames(metadata))) == TRUE) {
                metadata = subset(metadata, select=which(!duplicated(colnames(metadata))))
              } else {
                metadata = metadata
              }

              cols_not_needed_final = c('orig.ident', 'RNA_snn_res.0.8', 'seurat_clusters')
              find_not_needed_final = which(colnames(metadata) %in% cols_not_needed_final)
              if(length(find_not_needed_final) > 0) {
                metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed_final])
              }

              columns_remove = c()
              for(x in 1:length(colnames(metadata)))
              {
                number_of_NA = sum(is.na(metadata[[colnames(metadata)[x]]]))
                if(number_of_NA == length(metadata[[colnames(metadata)[x]]])) {
                  getcol = colnames(metadata)[x]
                  columns_remove = c(columns_remove, getcol)
                }
              }
              metadata = metadata %>% dplyr::select(-columns_remove)
              metadata$conditionAddedbyCogentDS = rep("NoConditionSpecified", nrow(metadata))
              mat = mat[,rownames(metadata)]
              dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design= ~1)
              condition_col = 'conditionAddedbyCogentDS'
              datalist = list(dds, metadata, input$countfilter, condition_col)
              data(datalist)
            }
            shinyalert("Data Upload Successful", paste("Number of samples:", ncol(mat), "\n" ,"Number of genes:", nrow(mat), "\n" ,"Let's Get Started with QC"), type = "success", confirmButtonCol = "#1c82e0")
          }, error = function(e) {
              shinyalert("Data Upoad Failed", "Check input parameters", type = "error", confirmButtonCol = "#1c82e0")
          }) 
        }

        if(input$fileoption == "Processed Data from CogentAP") {
          req(counts_data(), AP_object())
          tryCatch({
            inFile <- input$file
            if(!is.null(AP_object())) {
              df_rds = AP_object()
            } else {
              stop("Error file not Uoloaded!")
            }
            if(!is.null(AP_object())) {
              mat = counts_data()
              df_rds[["percent.mt"]] <- Seurat::PercentageFeatureSet(df_rds, pattern = "^MT-")
              metadata_AP = data.frame(df_rds@meta.data)
              output$bulkRNA_mat = DT::renderDataTable({
               metadata = data.frame(df_rds@meta.data)
                cols_not_needed = c('Barcode', 'orig.ident', 'No_of_Genes_nUMIs_USSs', 'No_of_Transcripts_nUMIs_USSs', 'No_of_Genes', 'No_of_Transcripts')
                find_not_needed = which(colnames(metadata) %in% cols_not_needed)

                if(length(find_not_needed) > 0) {
                  metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed])
                }

                cols = c('Barcoded_Reads', 'Trimmed_Reads', 'Unmapped_Reads', 'Mapped_Reads', 'Uniquely_Mapped_Reads', 'Multimapped_Reads', 'Chimeric_Reads', 'No_of_Reads_w_UMIs', 'Reads_After_Dedup_nUMIs_USSs', 'Exon_nUMIs_USSs', 'Intron_nUMIs_USSs', 'Gene_nUMIs_USSs', 'Intergenic_Reads', 'Ribosomal_Reads', 'Mitochondrial_Reads', 'Exon_Reads', 'Intron_Reads', 'Gene_Reads')
                find = which(colnames(metadata) %in% cols)
                if(length(find) > 0) {
  
                  metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
  
  
                  columns_remove = c()
                  for(x in 1:length(colnames(metadata_select)))
                  {
                    number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                    if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                      getcol = colnames(metadata_select)[x]
                      columns_remove = c(columns_remove, getcol)
                    }
                  }
  
                  metadata_select = metadata_select %>% dplyr::select(-columns_remove)
  
                  average_stats_cells = round(colMeans(metadata_select, na.rm = T))
                  average_stats_cells = data.frame(average_stats_cells)
                  colnames(average_stats_cells) = "Average Stats across Cells"
                  average_stats_cells$`Average Stats across Cells` = formatC(as.numeric(average_stats_cells$`Average Stats across Cells`), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  total_stats_cells = round(colSums(metadata_select, na.rm = T))
                  total_stats_cells = data.frame(total_stats_cells)
                  colnames(total_stats_cells) = "Total Stats across Cells"
                  total_stats_cells$Percentage = round((total_stats_cells$`Total Stats across Cells`/total_stats_cells$`Total Stats across Cells`[1])*100, 2)
                  total_stats_cells_forplot = total_stats_cells
                  total_stats_cells_forplot$values = as.numeric(total_stats_cells_forplot$`Total Stats across Cells`)
                  total_stats_cells$`Total Stats across Cells` = formatC(as.numeric(total_stats_cells$`Total Stats across Cells`), big.mark = ",", big.interval = 3, format = "f", digits = 0)

                  if('Strand_Specificity' %in% colnames(metadata)) {
                    Strand_Specificity_avg = mean(metadata$Strand_Specificity, na.rm = T)
                    Strand_Specificity_avg = data.frame(Strand_Specificity_avg)
                    colnames(Strand_Specificity_avg) = "Average Stats across Cells"
                    rownames(Strand_Specificity_avg) = "Strand_Specificity"
                    average_stats_cells = rbind(average_stats_cells, Strand_Specificity_avg)

                    Strand_Specificity_total = cbind('NA', 'NA')
                    Strand_Specificity_total = data.frame(Strand_Specificity_total)
                    colnames(Strand_Specificity_total) = c("Total Stats across Cells", "Percentage")
                    rownames(Strand_Specificity_total) = "Strand_Specificity"
                    total_stats_cells = rbind(total_stats_cells, Strand_Specificity_total)
                  }
  

                  numberofcells = cbind(formatC(as.numeric(ncol(df_rds)), big.mark = ",", big.interval = 3, format = "f", digits = 0), 'NA')
                  numberofcells = data.frame(numberofcells)
                  colnames(numberofcells) = c("Total Stats across Cells", "Percentage")
                  rownames(numberofcells) = "Number of Cells/Samples"
                  numberofcells_avg = 'NA'
                  numberofcells_avg = data.frame(numberofcells_avg)
                  colnames(numberofcells_avg) = "Average Stats across Cells"
                  rownames(numberofcells_avg) = "Number of Cells/Samples"

                  numberoffeatures = cbind(formatC(as.numeric(nrow(df_rds)), big.mark = ",", big.interval = 3, format = "f", digits = 0), 'NA')
                  numberoffeatures = data.frame(numberoffeatures)
                  colnames(numberoffeatures) = c("Total Stats across Cells", "Percentage")
                  rownames(numberoffeatures) = "Number of Genes/Features"
                  numberoffeatures_avg = 'NA'
                  numberoffeatures_avg = data.frame(numberoffeatures_avg)
                  colnames(numberoffeatures_avg) = "Average Stats across Cells"
                  rownames(numberoffeatures_avg) = "Number of Genes/Features"

                  average_stats_cells = rbind(average_stats_cells, numberoffeatures_avg, numberofcells_avg)
                  total_stats_cells = rbind(total_stats_cells, numberoffeatures, numberofcells)
                  df_show = cbind(total_stats_cells, average_stats_cells)
                  df_temp = df_show
                  df_temp$check_values = df_show$`Total Stats across Cells`
                  colnames(df_show) = c("Total_stats", "Percentage_of_Barcoded_Reads", "Average_stats")
                } else {
  
                  avg_nfeature = cbind('Average nFeature_RNA', formatC(round(mean(df_rds@meta.data$nFeature_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  avg_nCount = cbind('Average nCount_RNA', formatC(round(mean(df_rds@meta.data$nCount_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  #avg_percentmt = cbind('Average percent-mt', formatC(round(mean(df_rds@meta.data$percent.mt, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  numberofcells = cbind('Number of Cells/Samples', formatC(ncol(df_rds), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  numberoffeatures = cbind('Number of Genes/Features', formatC(nrow(df_rds), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  df_show = rbind(numberofcells, numberoffeatures, avg_nfeature, avg_nCount)
                  df_show = data.frame(df_show)
                  colnames(df_show) = c("Seurat_Object_Stats", "Value")
                  df_temp = df_show
                }
                df_show = data.frame(df_show)
                df_show[df_show == ""] <- NA
                df_show[df_show == "NaN"] <- NA
                df_show[df_show == 0] <- NA
                df_show = data.frame(df_show)
                DT::datatable(df_show, options = list(pageLength = 20, dom = "t"))
              })
            }
            if(!is.null(metadata_Collect())) {
              metadata_user = metadata_Collect()
              metadata_AP = metadata_AP[rownames(metadata_user),]
              metadata = cbind(metadata_user, metadata_AP)

              if(any(duplicated(colnames(metadata))) == TRUE) {
                metadata = subset(metadata, select=which(!duplicated(colnames(metadata))))
              } else {
                metadata = metadata
              }

              cols_not_needed_final = c('orig.ident', 'RNA_snn_res.0.8', 'seurat_clusters')
              find_not_needed_final = which(colnames(metadata) %in% cols_not_needed_final)
              if(length(find_not_needed_final) > 0) {
                metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed_final])
              }

              columns_remove = c()
              for(x in 1:length(colnames(metadata)))
              {
                number_of_NA = sum(is.na(metadata[[colnames(metadata)[x]]]))
                if(number_of_NA == length(metadata[[colnames(metadata)[x]]])) {
                  getcol = colnames(metadata)[x]
                  columns_remove = c(columns_remove, getcol)
                }
              }
              metadata = metadata %>% dplyr::select(-columns_remove)

              mat = mat[,rownames(metadata)]
              column_name <-  paste("~", input$condition_col)
              length_condition = length(unique(metadata[[input$condition_col]]))
              if(length_condition > 1) {
                condition_col_use = input$condition_col
                dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design= as.formula(column_name))
              } else {
                metadata$Samples_Found = rownames(metadata)
                metadata$conditionAddedbyCogentDS = rep("NoConditionSpecified", nrow(metadata))
                condition_col_use = 'conditionAddedbyCogentDS'
                dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design = ~1)
              }
              datalist = list(dds, metadata, input$countfilter, condition_col_use)
              data(datalist)
            } else {
              metadata_user = data.frame(colnames(mat))
              colnames(metadata_user) = "Samples_Found"
              rownames(metadata_user) = colnames(mat)
              metadata_AP = metadata_AP[rownames(metadata_user),]
              metadata = cbind(metadata_user, metadata_AP)

              if(any(duplicated(colnames(metadata))) == TRUE) {
                metadata = subset(metadata, select=which(!duplicated(colnames(metadata))))
              } else {
                metadata = metadata
              }

              cols_not_needed_final = c('orig.ident', 'RNA_snn_res.0.8', 'seurat_clusters')
              find_not_needed_final = which(colnames(metadata) %in% cols_not_needed_final)
              if(length(find_not_needed_final) > 0) {
                metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed_final])
              }

              columns_remove = c()
              for(x in 1:length(colnames(metadata)))
              {
                number_of_NA = sum(is.na(metadata[[colnames(metadata)[x]]]))
                if(number_of_NA == length(metadata[[colnames(metadata)[x]]])) {
                  getcol = colnames(metadata)[x]
                  columns_remove = c(columns_remove, getcol)
                }
              }
              metadata = metadata %>% dplyr::select(-columns_remove)

              metadata$conditionAddedbyCogentDS = rep("NoConditionSpecified", nrow(metadata))
              mat = mat[,rownames(metadata)]
              dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design= ~1)
              condition_col = 'conditionAddedbyCogentDS'
              datalist = list(dds, metadata, input$countfilter, condition_col)
              data(datalist)
            }
            shinyalert("Data Upload Successful", paste("Number of samples:", ncol(mat), "\n" ,"Number of genes:", nrow(mat), "\n" ,"Let's Get Started with QC"), type = "success", confirmButtonCol = "#1c82e0")
          }, error = function(e) {
              shinyalert("Data Upoad Failed", "Check Input Parameters!", type = "error", confirmButtonCol = "#1c82e0")
          })
        }

        if(input$fileoption == "Example Data") {
          df_counts = read.csv('data/example_counts.csv', header = TRUE, row.names = 1, check.names = FALSE)
          metadata = read.csv('data/example_metadata.csv', header = TRUE, row.names = 1, check.names = FALSE)
          mat = df_counts
          tryCatch({
            if (!is.null(mat)) {
              output$bulkRNA_mat = DT::renderDataTable({
                numberofcells = formatC(ncol(df_counts), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                numberofcells = data.frame(numberofcells)
                colnames(numberofcells) = "Total Stats across Samples"
                rownames(numberofcells) = "Number of Samples"
                numberofcells_avg = 'NA'
                numberofcells_avg = data.frame(numberofcells_avg)
                colnames(numberofcells_avg) = "Average Stats across Samples"
                rownames(numberofcells_avg) = "Number of Samples"

                numberoffeatures = formatC(nrow(df_counts), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                numberoffeatures = data.frame(numberoffeatures)
                colnames(numberoffeatures) = "Total Stats across Samples"
                rownames(numberoffeatures) = "Number of Genes"
                numberoffeatures_avg = 'NA'
                numberoffeatures_avg = data.frame(numberoffeatures_avg)
                colnames(numberoffeatures_avg) = "Average Stats across Samples"
                rownames(numberoffeatures_avg) = "Number of Genes"

                total_stats_cells = rbind(numberoffeatures, numberofcells)
                df_show = cbind(total_stats_cells)
                #DT::datatable(df_show, options = list(pageLength = 20, dom = "t", rowCallback = JS(rowCallback)))
                DT::datatable(df_show, options = list(pageLength = 20, dom = "t"))
              })
            }
            if(!is.null(metadata)) {
              mat = mat[,rownames(metadata)]
              dds <- DESeqDataSetFromMatrix(countData = round(as.matrix(mat)), colData = metadata, design= ~condition)
              datalist = list(dds, metadata, 0, 'condition')
              print (datalist)
              data(datalist)
            }
            shinyalert("Data Upload Successful", paste("Number of samples:", ncol(mat), "\n" ,"Number of genes:", nrow(mat), "\n" ,"Let's Get Started with QC"), type = "success", confirmButtonCol = "#1c82e0")
          })
        } 
    })
    return(data)
  })
}
        