#' Data Upload UI Component
#'
#' This function creates a set of UI elements for data uploading in a Shiny app. It includes 
#' a file input for dataset upload, a text input for specifying a project name, and numeric 
#' inputs for defining minimum cell and feature thresholds. The function is designed to be 
#' flexible, supporting both .csv and .rda file formats for dataset uploads.
#'
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return A UI element comprising a file input, a project name text input, and numeric inputs 
#'         for minimum cells and features. These elements are wrapped in a `tagList` for easy 
#'         integration into the app's UI.
#' @export
#'
#' @examples
#' In the UI definition of a Shiny app
#' dataUploadUI("dataUploadModule")
#' 
#' @note The 'minCells' and 'minFeatures' inputs allow users to specify filtering criteria 
#'       for the dataset post-upload, offering a basic level of data preprocessing.
#'       
dataUploadUI <- function(id) {
  ns <- NS(id)
  
  tagList(
    fluidRow(column(4,
      h2('Upload Data for Analysis'),
    radioButtons(ns("fileoption"), "Select type of Data Upload", choices = c("Processed Data from CogentAP", "Raw counts Matrix", "Example Data"), selected = "Processed Data from CogentAP"),
    bsPopover(ns("fileoption"), "File Upload options", "CogentDS v2.0 Bulk RNA app accepts Processed data from CogentAP, Raw counts matrix in .csv and has Example Data which is subset of GSE147507", placement = "right", trigger = "hover", options = list(container = "body")),
    conditionalPanel(
      condition = "input.fileoption == 'Raw counts Matrix' || input.fileoption == 'Processed Data from CogentAP'",
      fileInput(ns("file"), "Upload your dataset", accept = c(".csv", ".rds")),
      numericInput(ns("countfilter"), "Counts Filter", value = 0, min = 0, max = 100, step = 5),
      selectInput(ns("metadata_question"), label = "Sample Metadata", choices = c("Yes", "No"), selected = "No"),
      bsPopover(ns("metadata_question"), "Metadata", "Sample Metadata with condition, batch etc. with samples as first column", placement = "right", trigger = "hover", options = list(container = "body")),
      conditionalPanel(
        condition = "input.metadata_question == 'Yes'",
        fileInput(ns("metadata_file"), label = "Upload Metadata File", accept = c(".csv")),
        selectInput(ns("condition_col"), "Select Condition Column", choices = NULL, selected = NULL), ns=NS(id)), ns=NS(id)),
    actionButton(ns("upload"), label = "Upload Raw Data", class = "btn-primary", icon = icon("upload"), style='color:#fff; background-color:#1c82e0')),
    column(6,
      conditionalPanel(
        condition = "input.upload > 0",
        shinycssloaders::withSpinner(DT::dataTableOutput(ns("bulkRNA_mat")), size = 1), ns=NS(id))))
  )
}