#' QC Server Component
#'  
#' This server module handles the server-side logic for de visualization using ggplot2
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the data with matrix, metadata and counts filter
#'         
#' @export
#'

qcServer <- function(id, dataObject) {
  shiny::moduleServer(id, function(input, output, session) {

    ns <- session$ns
    
    plotOutPCA <- reactiveVal(NULL)
    plotOutSD <- reactiveVal(NULL)
    plotOutMA <- reactiveVal(NULL)

    metadataColumns = reactiveVal(NULL)

    shinyjs::hide("downloadPlot_pca")
    shinyjs::hide("downloadPlot_sd")
    shinyjs::hide("downloadPlot_ma")

    metadataColumns <- reactive({
      req(dataObject())

      metadata_use = colData(dataObject()[[1]])
      
      character_cols <- sapply(metadata_use, is.character)
      metadata_use[character_cols] <- lapply(metadata_use[character_cols], as.factor)

      # Get the names of  metadata columns
      all_metadata <- colnames(metadata_use)
      
      # Filter out numeric columns
      relevant_columns <- all_metadata[sapply(metadata_use[, all_metadata], is.factor)]

      relevant_columns <- relevant_columns[!is.na(relevant_columns)]
    
      # Return the filtered relevant columns
      return(relevant_columns)
    })

    observe({
      req(dataObject(), metadataColumns())
      updateSelectInput(session, "colorBy", label = "Color by", choices = metadataColumns())
    })

    plotOutPCA <- eventReactive(input$performvqcis_pca, {
      req(dataObject(), input$colorBy)
      dds = dataObject()[[1]]
      metadata = dataObject()[[2]]
      countsfilter = dataObject()[[3]]
      condition_column = dataObject()[[4]]

      if((!is.null(condition_column)) && (condition_column != "Samples_Found")) {
        table_condition = data.frame(table(metadata[[condition_column]]))
        colnames(table_condition) = c("condition", "number_of_samples")
        list_number_of_samples = c()
        for(x in 1:length(table_condition$condition))
        {
          val = table_condition$number_of_samples[x]
          list_number_of_samples = c(list_number_of_samples, val)
        }
        keep <- rowSums( counts(dds) >= countsfilter ) >= min(list_number_of_samples)
        dds <- dds[keep,]
      } else {
        keep <- rowSums( counts(dds) >= countsfilter ) >= 1
        dds <- dds[keep,]
      }
      if(input$qcvisualization == "PCA Plot") {
        colData(dds)$Samples = rownames(colData(dds))
        vsd <- vst(dds, blind=FALSE)
        if(!is.null(input$colorBy)) {
          p5 = plotPCA(vsd, intgroup=c(input$colorBy)) + geom_point(size = 10) + ggtitle("Principal Component Analysis Plot (PC1 vs PC2)") + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
        } #else {
          #p5 = plotPCA(vsd, intgroup=c('Samples')) + geom_point(size = 10) + ggtitle("Principal Component Analysis Plot (PC1 vs PC2)") + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
        #}
        return(p5)
      }
    })

    plotOutSD <- eventReactive(input$performvqcis_sd, {
      req(dataObject(), input$colorBy)
      dds = dataObject()[[1]]
      metadata = dataObject()[[2]]
      countsfilter = dataObject()[[3]]
      condition_column = dataObject()[[4]]

      if(!is.null(condition_column)) {
        table_condition = data.frame(table(metadata[[condition_column]]))
        colnames(table_condition) = c("condition", "number_of_samples")
        list_number_of_samples = c()
        for(x in 1:length(table_condition$condition))
        {
          val = table_condition$number_of_samples[x]
          list_number_of_samples = c(list_number_of_samples, val)
        }
        keep <- rowSums( counts(dds) >= countsfilter ) >= min(list_number_of_samples)
        dds <- dds[keep,]
        vsd <- vst(dds, blind=FALSE)
        annotate_col = data.frame(colData(dds)) %>% dplyr::select(input$colorBy)
        sampleDists <- dist(t(assay(vsd)))
        sampleDistMatrix <- as.matrix(sampleDists)
        rownames(sampleDistMatrix) <- colnames(assay(vsd))
        colnames(sampleDistMatrix) <- colnames(assay(vsd))
        colors <- colorRampPalette( rev(brewer.pal(9, "Blues")) )(255)
        if(nrow(annotate_col) > 0) {
          p5 = pheatmap(sampleDistMatrix, clustering_distance_rows = sampleDists, clustering_distance_cols = sampleDists, col = colors, annotation_col = annotate_col, main = "Sample Distance Plot - Distance across samples")
        } else {
          p5 = pheatmap(sampleDistMatrix, clustering_distance_rows = sampleDists, clustering_distance_cols = sampleDists, col = colors, main = "Sample Distance Plot - Distance across samples")
        }
        return(p5)
      } else {
        keep <- rowSums( counts(dds) >= countsfilter ) >= 1
        dds <- dds[keep,]
        vsd <- vst(dds, blind=FALSE)
        annotate_col = data.frame(colData(dds)) %>% dplyr::select(input$colorBy)
        sampleDists <- dist(t(assay(vsd)))
        sampleDistMatrix <- as.matrix(sampleDists)
        rownames(sampleDistMatrix) <- colnames(assay(vsd))
        colnames(sampleDistMatrix) <- colnames(assay(vsd))
        colors <- colorRampPalette( rev(brewer.pal(9, "Blues")) )(255)
        if(nrow(annotate_col) > 0) {
          p5 = pheatmap(sampleDistMatrix, clustering_distance_rows = sampleDists, clustering_distance_cols = sampleDists, col = colors, annotation_col = annotate_col, main = "Sample Distance Plot - Distance across samples")
        } else {
          p5 = pheatmap(sampleDistMatrix, clustering_distance_rows = sampleDists, clustering_distance_cols = sampleDists, col = colors, main = "Sample Distance Plot - Distance across samples")
        }
        return(p5)
      }
    })

    plotOutMA <- eventReactive(input$performvqcis_ma, {
      req(dataObject())
      dds = dataObject()[[1]]
      metadata = dataObject()[[2]]
      countsfilter = dataObject()[[3]]
      condition_column = dataObject()[[4]]

      if(!is.null(condition_column)) {
        table_condition = data.frame(table(metadata[[condition_column]]))
        colnames(table_condition) = c("condition", "number_of_samples")
        list_number_of_samples = c()
        for(x in 1:length(table_condition$condition))
        {
          val = table_condition$number_of_samples[x]
          list_number_of_samples = c(list_number_of_samples, val)
        }
        keep <- rowSums( counts(dds) >= countsfilter ) >= min(list_number_of_samples)
        dds <- dds[keep,]
        dds <- DESeq(dds)
        #p6 = DESeq2::plotMA(dds)
        return(dds)
        
      } else {
        keep <- rowSums( counts(dds) >= countsfilter ) >= 1
        dds <- dds[keep,]
        dds <- DESeq(dds)
        #p6 = DESeq2::plotMA(dds)
        return(dds)
      }
    })

    observe({
      if(!is.null(plotOutPCA())) {
        if(!is.null(dev.list())) {
          dev.off()
          message("Found an open Device! setting it off")
        } else {
          message("no open device found, internal test!")
        }
      } else {
        message("plot not found")
      }
    })

    observe({
      if(!is.null(plotOutSD())) {
        if(!is.null(dev.list())) {
          dev.off()
          message("Found an open Device! setting it off")
        } else {
          message("no open device found, internal test!")
        }
      } else {
        message("plot not found")
      }
    })

    observe({
      if(!is.null(plotOutMA())) {
        if(!is.null(dev.list())) {
          dev.off()
          message("Found an open Device! setting it off")
        } else {
          message("no open device found, internal test!")
        }
      } else {
        message("plot data not found")
      }
    })

    observe({
      if(!is.null(plotOutPCA())) {
        shinyjs::show("downloadPlot_pca")
      }
    })

    observe({
      if(!is.null(plotOutSD())) {
        shinyjs::show("downloadPlot_sd")
      }
    })

    observe({
      if(!is.null(plotOutMA())) {
        shinyjs::show("downloadPlot_ma")
      }
    })

    output$pca_plot <- renderPlot({
      req(plotOutPCA())
      #ggplotly(plotOutPCA())
      plotOutPCA()
    })

    output$sd_plot <- renderPlot({
      req(plotOutSD())
      plotOutSD()
    })

    output$ma_plot <- renderPlot({
      req(plotOutMA())

      DESeq2::plotMA(plotOutMA(), main = "MA plot - Log2 fold change vs Average Counts")
    })

    output$downloadPlot_pca <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_QC_PCAPlot", "-", Sys.Date(), ".png")
      },
      content = function(file) {
        plotOutPCA()
        ggsave(file, plot = plotOutPCA(), width = 10, height = 10)
      },
      contentType = 'image/png'
    )

    output$downloadPlot_sd <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_QC_SampleDistancePlot", "-", Sys.Date(), ".png")
      },
      content = function(file) {
        plotOutSD()
        ggsave(file, plot = plotOutSD(), width = 10, height = 10)
      },
      contentType = 'image/png'
    ) 

     output$downloadPlot_ma <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_QC_MAPlot", "-", Sys.Date(), ".png")
      },
      content = function(file) {
        png(file, width=800, height=800)
        DESeq2::plotMA(plotOutMA(), main = "MA plot - Log2 fold change vs Average Counts")
        dev.off()
      },
      contentType = 'image/png'
    )

  })
}