#' Visualization Server Component
#'  
#' This server module handles the server-side logic for de visualization using ggplot2
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the data with matrix, metadata and counts filter
#'         
#' @export
#'

visualizationServer <- function(id, de_results) {
  shiny::moduleServer(id, function(input, output, session) {

    ns <- session$ns
    
    plotOut_volcano <- reactiveVal(NULL)

    plotOut_heatmap <- reactiveVal(NULL)

    plotOut_pe <- reactiveVal(NULL)


    shinyjs::hide("downloadPlot_volcano")
    shinyjs::hide("downloadPlot_heamtmap")
    shinyjs::hide("downloadPlot_pe")
    shinyjs::hide("downloadtable_pe")

    metadataColumns <- reactive({
      req(de_results())

      metadata_use = colData(de_results()[[2]])
      
      character_cols <- sapply(metadata_use, is.character)
      metadata_use[character_cols] <- lapply(metadata_use[character_cols], as.factor)

      # Get the names of  metadata columns
      all_metadata <- colnames(metadata_use)
      
      # Filter out numeric columns
      relevant_columns <- all_metadata[sapply(metadata_use[, all_metadata], is.factor)]
    
      # Return the filtered relevant columns
      return(relevant_columns)
    })

    observe({
      req(de_results(), metadataColumns())
      updateSelectInput(session, "annotate_samples", label = "Annotate Samples By", choices = metadataColumns())
    })

    observe({
      if(!is.null(plotOut_volcano())) {
        if(!is.null(dev.list())) {
          dev.off()
          message("Found an open Device! setting it off")
        } else {
          message("no open device found, internal test!")
        }
      } else {
        message("plot not found")
      }
    })

    observe({
      if(!is.null(plotOut_heatmap())) {
        if(!is.null(dev.list())) {
          dev.off()
          message("Found an open Device! setting it off")
        } else {
          message("no open device found, internal test!")
        }
      } else {
        message("plot not found")
      }
    })

    observe({
      if(!is.null(plotOut_pe())) {
        if(!is.null(dev.list())) {
          dev.off()
          message("Found an open Device! setting it off")
        } else {
          message("no open device found, internal test!")
        }
      } else {
        message("plot not found")
      }
    })

    plotOut_volcano <- eventReactive(input$performvis_volcano, {
      req(de_results())
      dds_original = de_results()[[1]]
      dds = de_results()[[2]]
      res = de_results()[[3]]
      comparison = de_results()[[7]]
      res_out = data.frame(res)
      res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter | log2FoldChange <= -1*(input$logFC_filter), padj <= input$fdr_filter) %>% column_to_rownames('gene')
      if(input$visualization == "Volcano Plot") {
        req(res)
        res_use = data.frame(res)
        results = res_use
        results$gene <- rownames(res_use)
        Title <- paste("Volcano Plot for ", comparison)
          
        results$category <- ifelse(results$padj <= input$fdr_filter & results$log2FoldChange >= input$logFC_filter, "Upregulated",
                                     ifelse(results$padj <= input$fdr_filter & results$log2FoldChange <= -1*(input$logFC_filter), "Downregulated", "Non-significant"))
        results = results %>% dplyr::filter(category %in% c('Non-significant', 'Upregulated', 'Downregulated'))

        results$hover_info <- paste("Gene:", results$gene, "<br>Log2 FC:", round(results$log2FoldChange, 3), "<br>Adj P-value:", results$padj)

        results$category <- factor(results$category, levels = c("Non-significant", "Downregulated", "Upregulated"))
        p1 <- ggplot(results, aes(x = log2FoldChange, y = -log10(padj), color = category, text = hover_info)) +
            geom_point(alpha = 0.5) +
            scale_color_manual(values = c("Upregulated" = "red", "Downregulated" = "blue", "Non-significant" = "grey")) +
            labs(x = "Log2 Fold Change", y = "-Log10 Adjusted P-value", title = Title) + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
        return(p1)
      }
    })

    plotOut_heatmap <- eventReactive(input$performvis_heatmap, {
      req(de_results())
      dds_original = de_results()[[1]]
      dds = de_results()[[2]]
      res = de_results()[[3]]
      comparison = de_results()[[7]]
      res_out = data.frame(res)
      res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter | log2FoldChange <= -1*(input$logFC_filter), padj <= input$fdr_filter) %>% column_to_rownames('gene')
      if(input$visualization == "Heatmap") {
        req(dds, res_out)
        vsd <- vst(dds, blind=FALSE)
        counts = assay(vsd)
        counts = counts[rownames(res_out),]
        if(!is.null(input$annotate_samples)) {
          annotate_col = data.frame(colData(dds)) %>% dplyr::select(input$annotate_samples)
        }
        if(nrow(annotate_col) > 0) {
          p3 = pheatmap(counts, cluster_rows = TRUE, cluster_cols = TRUE, scale = "row", annotation_col = annotate_col, fontsize_row = 1, main = "Differential Gene Expression Heatmap")
        } else {
          p3 = pheatmap(counts, cluster_rows = TRUE, cluster_cols = TRUE, scale = "row", fontsize_row = 1, main = "Differential Gene Expression Heatmap")
        }
        return(p3)
      }
    })

    plotOut_pe <- eventReactive(input$performvis_pe, {
      req(de_results())
      dds_original = de_results()[[1]]
      dds = de_results()[[2]]
      res = de_results()[[3]]
      comparison = de_results()[[7]]
      res_out = data.frame(res)
      res_out <-  res_out %>% rownames_to_column('gene') %>% filter(log2FoldChange >= input$logFC_filter | log2FoldChange <= -1*(input$logFC_filter), padj <= input$fdr_filter) %>% column_to_rownames('gene')
      if(input$visualization == "Pathway Enrichment") {
        req(res_out)
        checkIDs = grep(pattern = "^ENS", x = rownames(res_out), value = T)
        checkIDs_dot = NULL
        if(length(checkIDs) == length(rownames(res_out))) {
          message("Ensembl IDs found")
          checkIDs_dot = grep(".[1.9]", x = rownames(res_out), value = T)
          if(length(checkIDs_dot) == length(rownames(res_out))) {
            id_split = strsplit(rownames(res_out), split = ".", fixed = T)
            final_ids = sapply(id_split, "[[", 1, simplify = TRUE)
            checkIDs_dot = data.frame(checkIDs_dot)
            checkIDs_dot$ENSEMBL = final_ids
            colnames(checkIDs_dot) = c("ID", "ENSEMBL")
            ids_use = bitr(final_ids, fromType = "ENSEMBL", toType = "ENTREZID", OrgDb = "org.Hs.eg.db")
          } else {
            ids_use = bitr(rownames(res_out),fromType = "ENSEMBL", toType = "ENTREZID", OrgDb = "org.Hs.eg.db")
          }
        } else {
          ids_use = bitr(rownames(res_out),fromType = "SYMBOL", toType = "ENTREZID", OrgDb = "org.Hs.eg.db")
        }

        if(!is.null(ids_use)) {
          if((nrow(checkIDs_dot) > 0) && (!is.null(checkIDs_dot))) {
            checkIDs_dot_filtered = checkIDs_dot %>% dplyr::filter(ENSEMBL %in% ids_use[,1])
            rownames(checkIDs_dot_filtered) = checkIDs_dot_filtered$ENSEMBL
            checkIDs_dot_filtered = checkIDs_dot_filtered[ids_use[,1],]
            ids_use$Original = checkIDs_dot_filtered$ID
            res_out_pvlaue = data.frame(res_out) %>% dplyr::select(log2FoldChange)
            res_out_pvlaue = data.frame(res_out_pvlaue) %>% dplyr::filter(rownames(res_out_pvlaue) %in% ids_use[,3])
            res_out_pvlaue$ID = rownames(res_out_pvlaue)
            res_out_pvlaue = res_out_pvlaue[ids_use[,1],]
            ids_use$log2FoldChange = res_out_pvlaue$log2FoldChange
            ids_use = arrange(ids_use, desc(log2FoldChange))
            ids_use_analysis = ids_use %>% dplyr::select(ENTREZID, log2FoldChange)
            ids_use_analysis = deframe(ids_use_analysis)
          } else {
            res_out_pvlaue = res_out %>% dplyr::select(log2FoldChange)
            res_out_pvlaue = res_out_pvlaue %>% dplyr::filter(rownames(res_out_pvlaue) %in% ids_use[,1])
            res_out_pvlaue$ID = rownames(res_out_pvlaue)
            res_out_pvlaue = res_out_pvlaue[ids_use[,1],]
            ids_use$log2FoldChange = res_out_pvlaue$log2FoldChange
            ids_use = arrange(ids_use, desc(log2FoldChange))
            ids_use_analysis = ids_use %>% dplyr::select(ENTREZID, log2FoldChange)
            ids_use_analysis = deframe(ids_use_analysis)
          }
          
        }

        if(!is.null(ids_use_analysis)) {
          if(input$pathway_option == "KEGG Pathway Gene signature") {
            result_enrichr <- gseKEGG(geneList = ids_use_analysis, organism = 'hsa', pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Reactome Pathway Gene signature") {
            result_enrichr <- gsePathway(ids_use_analysis, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Hallmark Gene sets") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "H") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Positional Gene sets") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C1") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Curated Gene sets") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C2") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Motif Gene sets") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C3") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Compuatational Gene sets") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C4") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "GO gene sets") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C5") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Oncogenic signatures") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C6") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
          if(input$pathway_option == "Immunologic signatures") {
            m_df <- msigdbr(species = "Homo sapiens")
            t2g <- msigdbr(species = "Homo sapiens", category = "C7") %>% dplyr::select(gs_name, entrez_gene)
            result_enrichr <- GSEA(ids_use_analysis, TERM2GENE = t2g, pvalueCutoff = input$pathway_option_pvalue, verbose = TRUE, minGSSize = 5, pAdjustMethod = "BH")
          }
        } else {
          shinyalert("No mapped genes found", "Check if your data has Gene Symbols or Ensembl IDs! Other IDs not accepted", type = "warning", confirmButtonCol = "#1c82e0")
        }
        
        if(nrow(result_enrichr) > 0) {
          p4 = clusterProfiler::dotplot(object = result_enrichr, color = "pvalue", showCategory = 25)
          result_enrichr_table = data.frame(result_enrichr@result)
          result_enrichr_table = arrange(result_enrichr_table, pvalue)
          result_enrichr_table = result_enrichr_table %>% dplyr::select(ID, Description, setSize, enrichmentScore, NES, pvalue, p.adjust, qvalue, rank)
          list_res = list(p4, result_enrichr_table)
          return(list_res)
        } else {
          shinyalert("No Enriched Terms Found", "Please check your Differential Expression LogFC, FDR cutoffs and GSEA pathway Enrichment pvlaue cutoff", type = "warning", confirmButtonCol = "#1c82e0")
        }
      }
    })

    output$de_plot_volvano <- renderPlotly({
      req(plotOut_volcano())
      ggplotly(plotOut_volcano(), tooltip = "text")
    })

    output$de_plot_heatmap <- renderPlot({
      req(plotOut_heatmap())
      plotOut_heatmap()
    })

    output$de_plot_pe <- renderPlot({
      req(plotOut_pe())
      plotOut_pe()[[1]]
    })

    output$pe_table <- DT::renderDataTable({
      req(plotOut_pe())
      table = data.frame(plotOut_pe()[[2]])
      if((!is.null(table)) && (nrow(table) > 0)) {
        DT::datatable(table)
      } else {
        shinyalert("No Enriched Terms Found", "Please check your Differential Expression LogFC, FDR cutoffs and GSEA pathway Enrichment pvlaue cutoff", type = "warning", confirmButtonCol = "#1c82e0")
      }
      
    })

    observe({
      req(plotOut_volcano())
      if(!is.null(plotOut_volcano())) {
        shinyjs::show("downloadPlot_volcano")
      }
    })

    observe({
      req(plotOut_heatmap())
      if(!is.null(plotOut_heatmap())) {
        shinyjs::show("downloadPlot_heamtmap")
      }
    })

    observe({
      req(plotOut_pe())
      if(!is.null(plotOut_pe())) {
        shinyjs::show("downloadPlot_pe")
        shinyjs::show("downloadtable_pe")
      }
    })

    output$downloadPlot_volcano <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_DE_VolcanoPlot", "-", Sys.Date(), ".png")
      },
      content = function(file) {
        plotOut_volcano()
        ggsave(file, plot = plotOut_volcano(), width = 10, height = 10)
      },
      contentType = 'image/png'
    ) 

    output$downloadPlot_heamtmap <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_DE_HeatmapPlot", "-", Sys.Date(), ".png")
      },
      content = function(file) {
        plotOut_heatmap()
        ggsave(file, plot = plotOut_heatmap(), width = 10, height = 10)
      },
      contentType = 'image/png'
    ) 

    output$downloadPlot_pe <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_DE_GSEAEnrichmentPlot", "-", Sys.Date(), ".png")
      },
      content = function(file) {
        plotOut_pe()[[1]]
        ggsave(file, plot = plotOut_pe()[[1]], width = 10, height = 10)
      },
      contentType = 'image/png'
    ) 

    output$downloadtable_pe <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_bulkRNA_DE_GSEAEnrichmentData", "-", Sys.Date(), ".csv")
      },
      content = function(file) {
        write.csv(plotOut_pe()[[2]], file, quote = F)
      },
      contentType = 'image/png'
    ) 

  })
}