#' Visualization UI Component
#'
#' This function creates a set of UI elements for data uploading in a Shiny app. It includes 
#' de analysis based on user selected control and treatment samples. 
#'
#' @param id A unique identifier for the UI module. This is necessary for distinguishing 
#'           this module's input from others in the same Shiny application.
#'
#' @return A UI element comprising a file input, a project name text input, and numeric inputs 
#'         for minimum cells and features. These elements are wrapped in a `tagList` for easy 
#'         integration into the app's UI.
#' @export
#'
#' @examples
#' In the UI definition of a Shiny app
#' deanalysisUI("DEModule")
#' 
#'       
visualizationUI <- function(id) {
  ns <- NS(id)
  
  tagList(
    fluidRow(
      column(4,
        h2('Differential Expression Visualization'),
        radioButtons(ns("visualization"), label = "Select Visualization Type", choices = c("Volcano Plot", "Heatmap"), selected = NULL),
        conditionalPanel(
          condition = "input.visualization == 'Volcano Plot' || input.visualization == 'Heatmap' || input.visualization == 'Pathway Enrichment'",
          numericInput(ns("logFC_filter"), label = "Log2 Fold Change Cutoff", value = 0.5, min = 0, max = 100, step = 0.5),
          numericInput(ns("fdr_filter"), label = "False Discovery Rate (FDR) / Adjusted pValue Cutoff", value = 0.05, min = 0, max = 1, step = 0.01), ns = NS(id)),
        conditionalPanel(
          condition = "input.visualization == 'Pathway Enrichment'",
          selectInput(ns("pathway_option"), "Choose GSEA Enrichment Database", choices = c("KEGG Pathway Gene signature", "Reactome Pathway Gene signature", "Hallmark Gene sets", "Positional Gene sets", "Curated Gene sets", "Motif Gene sets", "Compuatational Gene sets", "GO gene sets", "Oncogenic signatures", "Immunologic signatures"), selected = "Reactome Pathway Gene signature"),
          numericInput(ns("pathway_option_pvalue"), label = "GSEA Enrichment pvalue cutoff", value = 0.5, min = 0, max = 1, step = 0.1), ns = NS(id)),
        conditionalPanel(
          condition = "input.visualization == 'Heatmap'",
          selectInput(ns("annotate_samples"), "Annotate Samples By", choices = NULL, selected = NULL), ns = NS(id)),
        conditionalPanel(
          condition = "input.visualization == 'Volcano Plot'",
          actionButton(ns("performvis_volcano"), label = "Generate Plot", class = "btn-primary", icon = icon("upload"), style='color:#fff; background-color:#1c82e0'), ns = NS(id)),
        conditionalPanel(
          condition = "input.visualization == 'Heatmap'",
          actionButton(ns("performvis_heatmap"), label = "Generate Plot", class = "btn-primary", icon = icon("upload"), style='color:#fff; background-color:#1c82e0'), ns = NS(id)),
        conditionalPanel(
          condition = "input.visualization == 'Pathway Enrichment'",
          actionButton(ns("performvis_pe"), label = "Generate Plot", class = "btn-primary", icon = icon("upload"), style='color:#fff; background-color:#1c82e0'), ns = NS(id))),
      column(8,
        conditionalPanel(
          condition = "input.performvis_volcano > 0 && input.visualization == 'Volcano Plot'",
          downloadButton(ns('downloadPlot_volcano'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
          shinycssloaders::withSpinner(plotlyOutput(ns("de_plot_volvano"), width = "700", height = "600"), size = 1), ns = NS(id)),
        conditionalPanel(
          condition = "input.performvis_heatmap > 0 && input.visualization == 'Heatmap'",
          downloadButton(ns('downloadPlot_heamtmap'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
          shinycssloaders::withSpinner(plotOutput(ns("de_plot_heatmap"), width = "700", height = "600"), size = 1), ns = NS(id)),
        conditionalPanel(
          condition = "input.performvis_pe > 0 && input.visualization == 'Pathway Enrichment'",
          downloadButton(ns('downloadPlot_pe'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
          shinycssloaders::withSpinner(plotOutput(ns("de_plot_pe"), width = "700", height = "600"), size = 1),
          downloadButton(ns('downloadtable_pe'), 'Save Data', style='color:#fff; background-color:#1c82e0'),
          DT::dataTableOutput(ns("pe_table")), style = "height:800px; overflow-y: scroll;overflow-x: scroll;", ns = NS(id))
        )
    )
  )
}