#' Additional Plots Server
#' 
#' Handles server-side logic for the additional plots module. This allows users 
#' to visualize features using different plots including violin, ridge, feature and 
#' dot plot 
#'
#' @param id A unique identifier for the server module. This is necessary for 
#'           distinguishing this module's input from others in the same Shiny 
#'           application.
#'           
#' @param seuratObj A reactive expression containing the Seurat object to be 
#'                 processed.
#'
#' @return
#' @export
#' 
#' @seealso \code{\link[Seurat]{FeaturePlot}}, \code{\link[Seurat]{DotPlot}}, 
#'           \code{\link[Seurat]{VlnPlot}}, \code{\link[Seurat]{RidgePlot}} for 
#'           more information on specific plot functions.
#' 
#' @examples
#' additionalPlotsServer("additionalPlotsModule", seuratObject)
#'
additionalPlotsServer <- function(id, seuratObject) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    shinyjs::hide("downloadPlot_feature_genes")
    shinyjs::hide("downloadPlot_dot")
    shinyjs::hide("downloadPlot_vln")
    shinyjs::hide("downloadPlot_ridge")
    shinyjs::hide("feature_genesPlt")
    shinyjs::hide("featuredotPlt")
    shinyjs::hide("VlnPlt")
    shinyjs::hide("RidgePlt")
    
    # Track tab state
    tab_state <- reactiveVal(NULL)
    observeEvent(input$Tabs,{
      tab_state(input$Tabs)
    })
    
    # # Check if the Seurat object has a data slot present or not
    # check_data_slot <- function() {
    #   req(seuratObject())
    #   if (!"data" %in% Layers(seuratObject())) {
    #     shinyalert::shinyalert("No Data Layer Found", "Please run the normalization followed by dimension reduction & clustering module to proceed with additional visualization.",
    #                            type = "error"
    #     )
    #     return(FALSE)
    #   }
    #   return(TRUE)
    # }
    
    # Reactive value to hold metadata columns (factor and character)
    metadataColumns <- reactive({
      req(seuratObject())
      
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObject()@meta.data)
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(seuratObject()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Columns to be potentially removed
      column_not_needed <- c("Barcode")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% relevant_columns)) {
        relevant_columns <- relevant_columns[!(relevant_columns %in% column_not_needed)]
      }
      
      # Return the filtered relevant columns
      relevant_columns
    })
    
    # Reactive value to hold numerical metadata column
    metadataColumnsFeaturePlot <- reactive({
      req(seuratObject())
      
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObject()@meta.data)
      
      # Filter out numeric columns
      numerical_columns <- all_metadata[sapply(seuratObject()@meta.data[, all_metadata], is.numeric)]
      
      # Columns to be potentially removed
      column_not_needed <- c("nCount_BCRl", "nFeature_BCRl", "nCount_BCRh", "nFeature_BCRh",
                             "nCount_TCRa", "nFeature_TCRa", "nCount_TCRb", "nFeature_TCRb", "nCount_TCRdg",
                             "nFeature_TCRdg", "nCount_junction_reads", "nFeature_junction_reads", "nCount_span_frags",
                             "nFeature_span_frags")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% numerical_columns)) {
        numerical_columns <- numerical_columns[!(numerical_columns %in% column_not_needed)]
      }
      
      # Return the filtered columns
      numerical_columns
    })
    
    # Clear selections when switching between gene and metadata
    observeEvent(input$selection_type, {
      updateSelectizeInput(session, "plotFeaturePlot_options_features", choices =  "", selected = "")
    })
    
    observe({
      req(seuratObject(), input$selection_type)
      if (input$selection_type == "gene") {
        updateSelectizeInput(session, "plotFeaturePlot_options_features", choices = rownames(seuratObject()), server = TRUE)
      } else if (input$selection_type == "metadata") {
        updateSelectizeInput(session, "plotFeaturePlot_options_features", choices = metadataColumnsFeaturePlot())
      } 
    })
    
    observe({
      req(seuratObject())
      updateSelectInput(session, "plotDotPlot_options_group", choices = metadataColumns())
    })
    
    observe({
      req(seuratObject())
      updateSelectInput(session, "plotViolinPlot_group", choices = metadataColumns())
    })
    
    observe({
      req(seuratObject())
      updateSelectInput(session, "plotRidgePlot_options_group", choices = metadataColumns())
    })
    
    observe({
      if (input$selection_type == "gene") {
        showTab(inputId = "Tabs", target = "Dot Plot", session = session)
      } else {
        hideTab(inputId = "Tabs", target = "Dot Plot", session = session)
      }
    })
    
    # Feature plots (Gene based)
    featureplot_genes <- reactive({
      req(seuratObject(), input$plotFeaturePlot_options_features)
      # if (!check_data_slot()) 
      #   return(NULL)
      
      tryCatch({
        FeaturePlot(seuratObject(), pt.size = 2,
                    features = input$plotFeaturePlot_options_features,   
                    keep.scale = NULL,
                    cols = c("lightgrey", "blue"),
                    order = FALSE) & RotatedAxis() & 
          theme(axis.text = element_text(colour = "black", size = 15, face = "bold"),
                axis.title = element_text(colour = "black", size = 15, face = "bold"),
                legend.position = "right",
                legend.box = "vertical",
                panel.background = element_rect(colour = "white"))
      } , error = function(e) {
        # Custom error message for duplicates or other errors
        if (!"data" %in% Layers(seuratObject()) && tab_state() == "Feature Plot"){
          shinyalert::shinyalert("No Data Layer Found", "Please run the normalization followed by dimension reduction & clustering module to proceed with additional visualization.",
                                 type = "error"
          )        
        } 
      })
    })
    
    # Feature Plots (Metadata based)
    featureplot_metadata <- reactive({
      # if (!check_data_slot()) 
      #   return(NULL)
      req(seuratObject(), input$plotFeaturePlot_options_features)
      
      tryCatch({
        FeaturePlot(seuratObject(), pt.size = 2,
                    features = input$plotFeaturePlot_options_features,
                    keep.scale = NULL,
                    cols = c("lightgrey", "blue"),
                    order = FALSE) & RotatedAxis() & 
          theme(axis.text = element_text(colour = "black", size = 15, face = "bold"),
                axis.title = element_text(colour = "black", size = 15, face = "bold"),
                legend.position = "right",
                legend.box = "vertical",
                panel.background = element_rect(colour = "white"))
      } , error = function(e) {
        # Custom error message for duplicates or other errors
        if (!"data" %in% Layers(seuratObject()) && tab_state() == "Feature Plot"){
          shinyalert::shinyalert("No Data Layer Found", "Please run the normalization followed by dimension reduction & clustering module to proceed with additional visualization.",
                                 type = "error"
          )        
        } 
      })
    })
    
    observeEvent(input$plotFeaturePlot_options_features, {
      if (input$selection_type == "gene") {
        output$interactivePlot_feature_genes <- renderPlot({
          req(featureplot_genes())
          featureplot_genes()
        })
        
        if (!is.null(featureplot_genes()) && !is.null(input$plotFeaturePlot_options_features)) {
          shinyjs::show("downloadPlot_feature_genes")
          shinyjs::show("feature_genesPlt")
        }
        
        output$downloadPlot_feature_genes <- downloadHandler(
          filename = function() {
            paste0("CogentDS_v2.0_FeaturePlot_Genes", "-", Sys.Date(), input$feature_genesPlt)
          },
          content = function(file) {
            ggsave(file, plot = featureplot_genes(), width = 10, height = 10)
          },
          contentType = 'image'
        )
      } else {
        output$interactivePlot_feature_genes <- renderPlot({
          req(featureplot_metadata())
          featureplot_metadata()
        })
        
        if (!is.null(featureplot_metadata())) {
          shinyjs::show("downloadPlot_feature_genes")
          shinyjs::show("feature_genesPlt")
        }
        
        output$downloadPlot_feature_genes <- downloadHandler(
          filename = function() {
            paste0("CogentDS_v2.0_FeaturePlot_Metadata", "-", Sys.Date(), input$feature_genesPlt)
          },
          content = function(file) {
            ggsave(file, plot = featureplot_metadata(), width = 10, height = 10)
          },
          contentType = 'image'
        )
      }
    })
    
    # Dot Plot
    dotplot = reactive({
      req(seuratObject(), input$plotFeaturePlot_options_features, input$plotDotPlot_options_group)
      
      # Extract column values for selected group
      metaCol <- seuratObject()@meta.data[[input$plotDotPlot_options_group]]
      
      if(any(is.na(metaCol))) {
        # Remove rows with NA values
        req_indices <- !is.na(metaCol)
        seuratObject_filtered <- subset(seuratObject(), cells = Cells(seuratObject())[req_indices])
        
        p <- DotPlot(seuratObject_filtered, features = input$plotFeaturePlot_options_features, group.by = input$plotDotPlot_options_group) + 
          RotatedAxis() + 
          theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), 
                axis.title = element_text(colour = "black", size = 15, face = "bold"), 
                legend.position = "right", 
                legend.box = "vertical", 
                panel.background = element_rect(colour = "white"))
      } else {
        p <- DotPlot(seuratObject(), features = input$plotFeaturePlot_options_features, group.by = input$plotDotPlot_options_group) + 
          RotatedAxis() + 
          theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), 
                axis.title = element_text(colour = "black", size = 15, face = "bold"), 
                legend.position = "right", 
                legend.box = "vertical", 
                panel.background = element_rect(colour = "white"))
      }
      return(p)
    })
    
    observeEvent(input$plotFeaturePlot_options_features, {
      output$interactivePlot_dot <- renderPlot({
        dotplot()
      })
      
      if(!is.null(dotplot())) {
        shinyjs::show("downloadPlot_dot")
        shinyjs::show("featuredotPlt")
      }
      
      output$downloadPlot_dot <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_DotPlot", "-", Sys.Date(), input$featuredotPlt)
        },
        content = function(file) {
          dotplot()
          ggsave(file, plot = dotplot(), width = 10, height = 10)
        },
        contentType = 'image'
      )  
    })
    
    # Violin Plot
    vlnplot <- reactive({
      req(seuratObject(), input$plotFeaturePlot_options_features, input$plotViolinPlot_group)
      
      # Extract column values for selected group
      metaCol <- seuratObject()@meta.data[[input$plotViolinPlot_group]]
      
      if(any(is.na(metaCol))) {
        # Remove rows with NA values
        req_indices <- !is.na(metaCol)
        seuratObject_filtered <- subset(seuratObject(), cells = Cells(seuratObject())[req_indices])
        p <- VlnPlot(seuratObject_filtered, features = input$plotFeaturePlot_options_features, group.by = input$plotViolinPlot_group) & 
          RotatedAxis() & 
          guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & 
          theme(axis.text = element_text(colour = "black", size = 10, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), panel.background = element_rect(colour = "white"))
      } 
      else {
        p <- VlnPlot(seuratObject(), features = input$plotFeaturePlot_options_features, group.by = input$plotViolinPlot_group) & 
          RotatedAxis() & 
          guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & 
          theme(axis.text = element_text(colour = "black", size = 10, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), panel.background = element_rect(colour = "white"))
        
      }
      return(p)
    })
    
    observeEvent(input$plotFeaturePlot_options_features, {
      output$interactivePlot_vln <- renderPlot({
        vlnplot()
      })
      
      if(!is.null(vlnplot())) {
        shinyjs::show("downloadPlot_vln")
        shinyjs::show("VlnPlt")
      }
      
      output$downloadPlot_vln <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_VlnPlot", "-", Sys.Date(), input$VlnPlt)
        },
        content = function(file) {
          vlnplot()
          ggsave(file, plot = vlnplot(), width = 10, height = 10)
        },
        contentType = 'image'
      )  
    })
    
    # Ridge Plot
    ridgeplot <- reactive({
      req(seuratObject(), input$plotFeaturePlot_options_features, input$plotRidgePlot_options_group)
      
      # Extract column values for selected group
      metaCol <- seuratObject()@meta.data[[input$plotRidgePlot_options_group]]
      tryCatch({
        if(any(is.na(metaCol))) {
          # Remove rows with NA values
          req_indices <- !is.na(metaCol)
          seuratObject_filtered <- subset(seuratObject(), cells = Cells(seuratObject())[req_indices])
          p <- RidgePlot(seuratObject_filtered, features = input$plotFeaturePlot_options_features, group.by = input$plotRidgePlot_options_group) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & theme(axis.text = element_text(colour = "black", size = 10, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), panel.background = element_rect(colour = "white"))
          
        } 
        else {
          p <- RidgePlot(seuratObject(), features = input$plotFeaturePlot_options_features, group.by = input$plotRidgePlot_options_group) & RotatedAxis() & guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) & theme(axis.text = element_text(colour = "black", size = 10, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), panel.background = element_rect(colour = "white"))
        }
        return(p)
      }, error = function(e) {
        # Custom error message for duplicates or other errors
        if (!"data" %in% Layers(seuratObject()) && tab_state() == "Ridge Plot"){
          shinyalert::shinyalert("No Data Layer Found", "Please run the normalization followed by dimension reduction & clustering module to proceed with additional visualization.",
                                 type = "error"
          )        
        } 
      })
    })
    
    observeEvent(input$plotFeaturePlot_options_features, {
      output$interactivePlot_ridge <- renderPlot({
        ridgeplot()
      })
      
      if(!is.null(ridgeplot())) {
        shinyjs::show("downloadPlot_ridge")
        shinyjs::show("RidgePlt")
      }
      
      output$downloadPlot_ridge <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_RidgePlot", "-", Sys.Date(), input$RidgePlt)
        },
        content = function(file) {
          ridgeplot()
          ggsave(file, plot = ridgeplot(), width = 10, height = 10)
        },
        contentType = 'image'
      )  
    })
  })
}