#' Server logic for Clustering and Visualization Module
#' 
#' Implements server-side functionality for clustering and visualization within
#' a Shiny application, leveraging Seurat for comprehensive single-cell data analysis.
#' This module allows users to perform clustering , and
#' visualize the results using dimensionality reduction techniques such as UMAP and t-SNE.
#' Parameters for each analysis can be dynamically set by the user, with visualizations
#' rendered interactively.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#' 
#' @seealso \code{\link[Seurat]{FindClusters}}, \code{\link[Seurat]{RunUMAP}}, and
#'          \code{\link[Seurat]{RunTSNE}} for more details on the underlying analysis
#'          methods used.
#'
#' @examples
#' clusteringVisualizationServer("clusteringVisualizationModule", seuratObject)
clusteringVisualizationServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    shinyjs::hide("downloadPlot")
    shinyjs::hide("DimPlot")
    
    static_dimredplot <- reactiveVal(NULL)
    clustering_data <- reactiveVal(NULL)
    
    # Clustering logic
    clustering_data <- eventReactive(input$runClusteringanddimRed, {
      req(seuratObj())
      updatedSeuratObject <- NULL
      shinyjs::disable("runClusteringanddimRed")
      shinyjs::disable("slect_nondim")
      tryCatch({
        withProgress(message = 'Performing Clustering & Dimension reduction...', value = 0, {
          incProgress(0.5, detail = "Clustering...")
          updatedSeuratObject = seuratObj()
          parameter_list = updatedSeuratObject@assays$user_parameter_tracker
          dimensions_use = paste0('Number of dimensions for clustering: ', input$dimsForClustering)
          resolution_use = paste0('Clustering Resolution: ', input$resolutionForClustering)
          updatedSeuratObject@assays$user_parameter_tracker = NULL
          updatedSeuratObject <- Seurat::FindNeighbors(updatedSeuratObject, dims = 1:input$dimsForClustering)
          updatedSeuratObject <- Seurat::FindClusters(updatedSeuratObject, resolution = input$resolutionForClustering)
          if(input$slect_nondim == 'UMAP') {
            umapDims_use = paste0('Number of dimensions for UMAP: ', input$umapDims)
            nneighbors_use = paste0('Number of neighbors: ', input$nneighbors)
            mindist_use = paste0('Mininum distance: ', input$mindist)
            parameter_list = c(parameter_list, dimensions_use, resolution_use, umapDims_use, nneighbors_use, mindist_use)
            updatedSeuratObject <- Seurat::RunUMAP(updatedSeuratObject, dims = 1:input$umapDims, n.neighbors = input$nneighbors, min.dist = input$mindist)
            updatedSeuratObject@assays$user_parameter_tracker = parameter_list
            seuratObj(updatedSeuratObject) # Update the main Seurat object
          }
          if(input$slect_nondim == 'tSNE') {
             tsneDims_use = paste0('Number of dimensions for tSNE: ', input$tsneDims)
             perplexity_use = paste0('Perplexity for tsne: ', input$perplexity)
             parameter_list = c(parameter_list, dimensions_use, resolution_use, tsneDims_use, perplexity_use)
             updatedSeuratObject <- Seurat::RunTSNE(updatedSeuratObject, dims = 1:input$tsneDims, perplexity = input$perplexity, check_duplicates = FALSE)
             updatedSeuratObject@assays$user_parameter_tracker = parameter_list
             seuratObj(updatedSeuratObject) # Update the main Seurat object
          }
          incProgress(0.5, detail = "Finalizing...")
        })
        shinyalert("Clustering & Non-linear reduction completed", "Proceed with celltype annotation", type = "success", confirmButtonCol = "#1c82e0")
      }, error = function(e) {
        shinyalert("An error occurred during Clustering & Non-linear reduction", "Please check your inputs and try again", type = "error", confirmButtonCol = "#1c82e0")
      })
      shinyjs::enable("runClusteringanddimRed")
      return(updatedSeuratObject)
    })

    static_dimredplot <- reactive({
      req(clustering_data())
      if ("seurat_clusters" %in% colnames(seuratObj()@meta.data)){
      plot <- DimPlot(seuratObj(), group.by = "seurat_clusters", pt.size = input$ptSize, label = TRUE, label.size = 5) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
      return(plot)
      } else NULL
    })

    observe({
      req(static_dimredplot())
      if(!is.null(static_dimredplot())) {
        shinyjs::show("downloadPlot")
        shinyjs::show("DimPlot")
      }
    })

    output$dimreductionPlot <- renderPlotly({
      req(static_dimredplot())
      ggplotly(static_dimredplot())
    })

    output$downloadPlot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_DimensionReductionPlot", "-", Sys.Date(), input$DimPlot)
      },
      content = function(file) {
        req(static_dimredplot())
        static_dimredplot()
        ggsave(file, plot = static_dimredplot(), width = 10, height = 10)
      },
      contentType = 'image'
    )
  })
}