#' Data Upload Server Component
#'  
#' This server module handles the server-side logic for data upload,
#' including reading the uploaded file, creating a Seurat object,
#' and storing it in a reactive value.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the Seurat object created from
#'         the uploaded data.
#'         
#' Detailed Parameter Description:
#' - counts: A matrix of gene expression data where rows are genes and columns are cells/samples.
#' - project: Name of the project.
#' - min.cells: Minimum Number of cells in which a gene must be detected to be kept in the dataset.
#' - min.features: Minimum Number of features (genes) detected in a cell for it to be included in the analysis.
#'
#' @seealso \code{\link[Seurat]{CreateSeuratObject}}, \code{\link[Seurat]{PercentageFeatureSet}}
#' for details on Seurat object creation and estimation of % of all counts that belong to set of features. 
#'         
#' @export
#'
#' @examples
#' # In the server definition of a Shiny app
#' seuratObject <- dataUploadServer("dataUploadModule")
dataUploadServer <- function(id, value) {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns
    data <- reactiveVal(NULL)
    
    observe({ 
      tryCatch({
        req(input$metadata_file)
        metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1, check.names = F)
        updateSelectInput(session, "samples_col", choices = colnames(metadata))
      })
    })
    
    observe({ 
      tryCatch({
        req(input$metadata_file, input$samples_col)
        metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1, check.names = F)
        sampletypes = unique(metadata[[input$samples_col]])
        sampletypes = c('Analyze all samples', sampletypes)
        updateSelectizeInput(session, "samples_pick", choices = sampletypes, selected = sampletypes[1], server = T)
      })
    })
    
    shiny::observeEvent(input$upload, {
      shinyjs::disable("upload")
      tryCatch({
        withProgress(message = 'Reading User inputs:', {
          incProgress(0.1, detail = 'tracking inputs and parameters')
          timeStart <- Sys.time()
          parameter_list = NULL
          if(input$fileoption == "Raw count matrix")
          {
            seuratObject = NULL
            req(input$file_raw, input$geneInfofile)
            seed_use = paste0('Seed Setting: set.seed(123456)')
            raw_counts_use = paste0('Raw count matrix: ', input$file_raw$name)
            geneinfo_use = paste0('Gene Info file: ', input$geneInfofile$name)
            metadata_use = paste0('Metadata file: ', input$metadata_file$name)
            projectName_use = paste0('Project Name: ', input$projectName)
            minCells_use = paste0('Minimum number of cells: ', input$minCells)
            minFeatures_use = paste0('Minimum number of features: ', input$minFeatures)
            parameter_list = c(seed_use, raw_counts_use, geneinfo_use, metadata_use, projectName_use, minCells_use, minFeatures_use)
            incProgress(0.1, detail = 'executing..')
            if((grepl(".csv$", input$file_raw$name)) | (grepl(".csv.gz$", input$file_raw$name)))
            {
              message('running check for number of cells')
              ch_check = fread(file = input$file_raw$datapath, sep = ",", data.table = F, verbose = F, check.names = FALSE) ## read the file using fread
              number_of_cells_in_data = length(colnames(ch_check)) - 1
              if(number_of_cells_in_data < 20) {
                seuratObject = NULL
                shinyalert("Too few cells in a single cell RNA experiment", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
                stop ()
              }
              if(input$metadata_question == "Yes")
              {
                req(input$metadata_file)
                metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1, check.names = F)
                message("read metadata")
                if(!is.null(input$geneInfofile)) {
                  if(is.null(input$samples_pick))
                  {
                    seuratObject = NULL
                    seuratObject_out = NULL
                    shinyalert("Sample Selection not done", "Please select atleast one sample", type = "error", confirmButtonCol = "#1c82e0")
                    stop()
                  }
                  if(length(input$samples_pick) > 1)
                  {
                    if('Analyze all samples' %in% input$samples_pick) {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      message("object ready")
                    } else {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                      message("object ready")
                    }
                  }
                  if(length(input$samples_pick) == 1)
                  {
                    if(input$samples_pick == 'Analyze all samples') {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      message("object ready")
                    } else {
                      source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                      seuratObject = seuratObject_out
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                      message("object ready")
                    }
                  }
                } else {
                  shinyalert('Gene Info file is missing or set to NULL', 'Please refresh app and upload Gene Info file', type = "warning", confirmButtonCol = "#1c82e0")
                  stop ()
                }
              }
              if(input$metadata_question == "No")
              {
                metadata = NULL
                if(!is.null(input$geneInfofile)) {
                  source('R/Setup_Funtions/bin/FastRead_largeCSV.R', local = TRUE)
                  seuratObject = seuratObject_out
                  message("object ready")
                } else {
                  seuratObject = NULL
                  seuratObject_out = NULL
                  shinyalert('Gene Info file is missing or set to NULL', 'Please refresh app and upload Gene Info file', type = "warning", confirmButtonCol = "#1c82e0")
                }
              }
              incProgress(0.3, detail = 'seurat object creation completed')
              rm(seuratObject_out)
            }
          }
          if(input$fileoption == "Processed Data from CogentAP")
          {
            seuratObject = NULL
            junction_reads <- NULL
            span_frags <- NULL
            TCRb <- NULL
            TCRa <- NULL
            TCRdg <- NULL
            BCRh <- NULL
            BCRl <- NULL
            df_trans <- NULL
            req(input$file)
            seed_use = paste0('Seed Setting: set.seed(123456)')
            cogentAP_use = paste0('CogentAP processed data: ', input$file$name)
            metadata_use = paste0('Metadata file: ', input$metadata_file$name)
            projectName_use = paste0('Project Name: ', input$projectName)
            minCells_use = paste0('Minimum number of cells: ', input$minCells)
            minFeatures_use = paste0('Minimum number of features: ', input$minFeatures)
            parameter_list = c(seed_use, cogentAP_use, metadata_use, projectName_use, minCells_use, minFeatures_use)
            incProgress(0.1, detail = 'executing...')
            if (grepl(".rda$", input$file$name)) {
              load(input$file$datapath)
              df <- get("CogentDS_data")[["raw_data"]]$gm
              
              number_of_cells_in_data = length(colnames(df))
              if(number_of_cells_in_data < 20) {
                seuratObject = NULL
                shinyalert("Too few cells in a single cell RNA experiment", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
                stop ()
              }
              
              metadata_import <-  CogentDS_data[["raw_data"]][["metadata"]]
              # Get fusion genes
              # Check if "overlay_raw_data" exists in the loaded data
              if ("overlay_raw_data" %in% names(get("CogentDS_data"))) {
                sce <- get("CogentDS_data")[["overlay_raw_data"]]
                
                if ("gene_fusion" %in% altExpNames(sce)) {
                  alt_exp <- altExp(sce, "gene_fusion")
                  junction_reads <- assay(alt_exp, "junction_reads")
                  span_frags <- assay(alt_exp, "span_frags")
                }
                
                # Get clonotypes data
                if (any(c("TCRb", "TCRa", "BCRh", "BCRl") %in% altExpNames(sce))) {
                  if ("TCRb" %in% altExpNames(sce)) TCRb <- assay(altExp(sce, "TCRb"))
                  if ("TCRa" %in% altExpNames(sce)) TCRa <- assay(altExp(sce, "TCRa"))
                  if ("BCRh" %in% altExpNames(sce)) BCRh <- assay(altExp(sce, "BCRh"))
                  if ("BCRl" %in% altExpNames(sce)) BCRl <- assay(altExp(sce, "BCRl"))
                }
              }
              seuratObject <- Seurat::CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures)
              if(input$metadata_question == "Yes") {
                req(input$metadata_file)
                metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1)
                check_cells = intersect(rownames(metadata), Cells(seuratObject))
                if(length(check_cells) == length(Cells(seuratObject))) {
                  message("read metadata")
                  if(length(input$samples_pick) == 1) {
                    if(input$samples_pick == 'Analyze all samples') {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    }
                    else {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                  if(length(input$samples_pick) > 1) {
                    if('Analyze all samples' %in% input$samples_pick) {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    } else {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                } else {
                  message ("Fatal Error: Cells in metadata do not match user provided seurat object; Metadata should have cells in the same order and equal length as in user provided seurat object")
                }
              }
              if(input$metadata_question == "No") {
                metadata = NULL
                seuratObject <- Seurat::CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures, meta.data = metadata_import)
              }
              rm(df)
            }
            if (grepl(".rds$", input$file$name)) {       ### added the .rds acceptance
              df_temp <- readRDS(input$file$datapath)
              ###### check number of cells ########
              number_of_cells_in_data = length(Cells(df_temp))
              if(number_of_cells_in_data < 20) {
                shinyalert("Too few cells in a single cell RNA experiment", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
                stop ()
              }
              message('read object')
              df <- df_temp[["RNA"]]$counts
              
              metadata_import = data.frame(df_temp@meta.data)
              cols_not_needed_import = c('orig.ident', 'nCount_RNA', 'nFeature_RNA', 'seurat_clusters', 'RNA_snn_res.0.5', 'RNA_snn_res.0.8', 'RNA_snn_res.1')
              find_not_needed_import = which(colnames(metadata_import) %in% cols_not_needed_import)
              
              if(length(find_not_needed_import) > 0) {
                metadata_import = metadata_import %>% dplyr::select(!colnames(metadata_import)[find_not_needed_import])
                message('removed unwanted colums')
              }
              
              if (any(c("junction_reads", "span_frags") %in%  names(df_temp))){
                junction_reads <- df_temp[["junction_reads"]]$counts
                span_frags <- df_temp[["span_frags"]]$counts
                message("read fusion")
              }
              
              if (any(c("BCRl", "BCRh", "TCRa", "TCRb", "TCRdg") %in%  names(df_temp))){
                BCRh <- df_temp[["BCRh"]]$counts
                BCRl <- df_temp[["BCRl"]]$counts
                TCRa <- df_temp[["TCRa"]]$counts
                TCRb <- df_temp[["TCRb"]]$counts
                TCRdg <- df_temp[["TCRdg"]]$counts
                message("read clonotype")
              }
              
              if(any('Transcript_RNA' %in%  names(df_temp)))
              {
                df_trans <- df_temp[["Transcript_RNA"]]$counts
                message("read isoform")
              }
              
              seuratObject <- Seurat::CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures, meta.data = metadata_import)
              if(input$metadata_question == "Yes") {
                req(input$metadata_file)
                metadata = read.csv(input$metadata_file$datapath, header = TRUE, row.names = 1)
                check_cells = intersect(rownames(metadata), Cells(seuratObject))
                if(length(check_cells) == length(Cells(seuratObject))) {
                  message("read metadata")
                  if(length(input$samples_pick) == 1) {
                    if(input$samples_pick == 'Analyze all samples') {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    }
                    else {
                      if(length(colnames(metadata)) == 1) {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                  if(length(input$samples_pick) > 1) {
                    if('Analyze all samples' %in% input$samples_pick) {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                    } else {
                      if(length(colnames(metadata)) == 1) {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata, col.name = input$samples_col)
                      } else {
                        message("object ready")
                        metadata = metadata[Cells(seuratObject),]
                        seuratObject = AddMetaData(seuratObject, metadata = metadata)
                      }
                      Idents(seuratObject) = input$samples_col
                      seuratObject$orig.ident = seuratObject[[input$samples_col]]
                      seuratObject = subset(seuratObject, idents = input$samples_pick)
                    }
                  }
                } else {
                  message ("Fatal Error: Cells in metadata do not match user provided seurat object; Metadata should have cells in the same order and equal length as in user provided seurat object")
                }
              }
              if(input$metadata_question == "No") {
                metadata = NULL
                seuratObject <- Seurat::CreateSeuratObject(counts = df, project = input$projectName, min.cells = input$minCells, min.features = input$minFeatures, meta.data = metadata_import)
              }
              rm(df)
            }
            if(!is.null(seuratObject)) {
              if((!is.null(df_trans)) && (nrow(df_trans) > 0)) {
                intersect_col = intersect(colnames(df_trans), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  df_trans = df_trans[,intersect_col]
                  seuratObject[["Transcript_RNA"]] <- Seurat::CreateAssayObject(counts = df_trans)
                  message("isoform ready")
                }
              }
              if ((!is.null(junction_reads)) && (nrow(junction_reads) > 0)) {
                intersect_col = intersect(colnames(junction_reads), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  junction_reads = junction_reads[,intersect_col]
                  seuratObject[["junction_reads"]] <- Seurat::CreateAssayObject(counts = junction_reads)
                  message("junction ready")
                }
              }
              if ((!is.null(span_frags)) && (nrow(span_frags) > 0)) {
                intersect_col = intersect(colnames(span_frags), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  span_frags = span_frags[,intersect_col]
                  seuratObject[["span_frags"]] <- Seurat::CreateAssayObject(counts = span_frags)
                  message("span ready")
                }
              }
              if ((!is.null(BCRh)) && (nrow(BCRh) > 0)) {
                intersect_col = intersect(colnames(BCRh), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  BCRh = BCRh[,intersect_col]
                  seuratObject[["BCRh"]] <- Seurat::CreateAssayObject(counts = BCRh)
                  message("BCRh ready")
                }
              } 
              if ((!is.null(BCRl)) && (nrow(BCRl) > 0)) {
                intersect_col = intersect(colnames(BCRl), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  BCRl = BCRl[,intersect_col]
                  seuratObject[["BCRl"]] <- Seurat::CreateAssayObject(counts = BCRl)
                  message("BCRl ready")
                }
              }
              if ((!is.null(TCRa)) && (nrow(TCRa) > 0)) {
                intersect_col = intersect(colnames(TCRa), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRa = TCRa[,intersect_col]
                  seuratObject[["TCRa"]] <- Seurat::CreateAssayObject(counts = TCRa)
                  message("TCRa ready")
                }
              }
              if ((!is.null(TCRb)) && (nrow(TCRb) > 0)) {
                intersect_col = intersect(colnames(TCRb), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRb = TCRb[,intersect_col]
                  seuratObject[["TCRb"]] <- Seurat::CreateAssayObject(counts = TCRb)
                  message("TCRb ready")
                }
              }
              if ((!is.null(TCRdg)) && (nrow(TCRdg) > 0)) {
                intersect_col = intersect(colnames(TCRdg), Cells(seuratObject))
                if(length(intersect_col) > 0) {
                  TCRdg = TCRdg[,intersect_col]
                  seuratObject[["TCRdg"]] <- Seurat::CreateAssayObject(counts = TCRdg)
                  message("TCRdg ready")
                }
              }
              incProgress(0.3, detail = 'seurat object creation completed')
            } else {
              message ("Fatal: Seurat object creation failed; Please check your inputs")
              stop("Error: Seurat object creation failed: Fatal")
            }
          }
          if(input$fileoption == "Example data") {
            seed_use = paste0('Seed Setting: set.seed(123456)')
            seuratObject <- readRDS('data/exampleData_PBMC_2k_with_Isoform-Fusion-Clonotype.rds')
            examplefile = paste0('Example data: ', 'exampleData_PBMC_2k_with_Isoform-Fusion-Clonotype.rds')
            projectName_use = paste0('Project Name: ', 'SeuratProject')
            minCells_use = paste0('Minimum number of cells: ', 0)
            minFeatures_use = paste0('Minimum number of features: ', 0)
            incProgress(0.1, detail = 'executing...')
            sample_not_needed = 'Sample'
            find_not_needed_sample = which(colnames(seuratObject@meta.data) %in% sample_not_needed)
            
            if(length(find_not_needed_sample) > 0) {
              seuratObject@meta.data = seuratObject@meta.data %>% dplyr::select(!colnames(seuratObject@meta.data)[find_not_needed_sample])
              message('Removed unwanted columns from the metadata')
            }
            parameter_list = c(seed_use, examplefile, projectName_use, minCells_use, minFeatures_use)
            incProgress(0.3, detail = 'seurat object creation completed')
            message("object ready") 
          }
          
          if (!is.null(seuratObject)) {
            incProgress(0.2, detail = 'seurat object checks in progress')
            seuratObject@assays$user_parameter_tracker = parameter_list  ######## created a tracker for user args
            
            number_of_cells_in_data = length(Cells(seuratObject))
            if(number_of_cells_in_data < 20) {
              shinyalert("Too few cells; < 20 found, please adjust the filtering parameters", "CogentDS v2.0 expects >= 20 cells", type = "warning", confirmButtonCol = "#1c82e0")
              stop ()
            }
            
            ###### add dry run paramerters
            dryrun1 = paste0('dry run paramerters; normalization method: LogNormalize')
            dryrun2 = paste0('dry run paramerters; nfeatures: 2000')
            dryrun3 = paste0('dry run paramerters; feature selection method: vst')
            dryrun4 = paste0('dry run paramerters; npcs: 10')
            dryrun5 = paste0('dry run paramerters; number of dimensions: 10')
            dryrun6 = paste0('dry run paramerters; clustering resolution: 0.5')
            parameter_list = c(parameter_list, dryrun1, dryrun2, dryrun3, dryrun4, dryrun5, dryrun6)
            seuratObject@assays$user_parameter_tracker = parameter_list
            
            message("parameter tracking is on")
            message("Object ready for stats!")
            
            # Estimating the percentage of mitochondrial genes as an initial quality control metric
            ##seuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(seuratObject, pattern = "^MT-")
            check_Mito_presence_human = grep("^MT-", rownames(seuratObject), value = TRUE, ignore.case = FALSE)
            check_Mito_presence_mouse = grep("^mt-", rownames(seuratObject), value = TRUE, ignore.case = FALSE)
            if(length(check_Mito_presence_human) > 0) {
              seuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(seuratObject, pattern = "^MT-")
            }
            if(length(check_Mito_presence_mouse) > 0) {
              seuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(seuratObject, pattern = "^mt-")
            }
            if((length(check_Mito_presence_human) == 0) && (length(check_Mito_presence_mouse) == 0)) {
              seuratObject[["percent.mt"]] = rep(0, ncol(seuratObject))
            }
            metadata = data.frame(seuratObject@meta.data)
            cols_imp = c('Barcoded_Reads', 'Trimmed_Reads', 'Unmapped_Reads', 'Mapped_Reads', 'Uniquely_Mapped_Reads', 'Multimapped_Reads', 'Chimeric_Reads', 'No_of_Reads_w_UMIs', 'Reads_After_Dedup_nUMIs_USSs', 'Exon_nUMIs_USSs', 'Intron_nUMIs_USSs', 'Gene_nUMIs_USSs', 'Intergenic_Reads', 'Ribosomal_Reads', 'Mitochondrial_Reads', 'Exon_Reads', 'Intron_Reads', 'Gene_Reads')
            find_imp = which(colnames(metadata) %in% cols_imp)
            
            if(length(find_imp) > 0) {
              if('Intergenic_Reads' %in% colnames(metadata)) {
                seuratObject$IntergenicReadFraction = as.numeric(seuratObject$Intergenic_Reads)/as.numeric(seuratObject$Mapped_Reads)
                seuratObject$Intergenic_reads_percentage = (as.numeric(seuratObject$Intergenic_Reads)/as.numeric(seuratObject$Mapped_Reads))*100
              } else {
                seuratObject$IntergenicReadFraction = rep(0, nrow(seuratObject@meta.data))
                seuratObject$Intergenic_reads_percentage = rep(0, nrow(seuratObject@meta.data))
                
              }
              if('Ribosomal_Reads' %in% colnames(metadata)) {
                seuratObject$RibosomalReadFraction = as.numeric(seuratObject$Ribosomal_Reads)/as.numeric(seuratObject$Trimmed_Reads)
                seuratObject$Ribosomal_reads_percentage = (as.numeric(seuratObject$Ribosomal_Reads)/as.numeric(seuratObject$Trimmed_Reads))*100
              } else {
                seuratObject$RibosomalReadFraction = rep(0, nrow(seuratObject@meta.data))
                seuratObject$Ribosomal_reads_percentage = rep(0, nrow(seuratObject@meta.data))
              }
            } else {
              seuratObject$IntergenicReadFraction = rep(0, nrow(seuratObject@meta.data))
              seuratObject$RibosomalReadFraction = rep(0, nrow(seuratObject@meta.data))
              seuratObject$Intergenic_reads_percentage = rep(0, nrow(seuratObject@meta.data))
              seuratObject$Ribosomal_reads_percentage = rep(0, nrow(seuratObject@meta.data))
            }
            output$seurat <- DT::renderDataTable({
              
              rowCallback <- c(
                "function(row, data){",
                "  for(var i=0; i<data.length; i++){",
                "    if(data[i] === null){",
                "      $('td:eq('+i+')', row).html('NA')",
                "        .css({'color': 'rgb(151,151,151)', 'font-style': 'italic'});",
                "    }",
                "  }",
                "}"  
              )
              
              cols_not_needed = c('orig.ident', 'No_of_Genes_nUMIs_USSs', 'No_of_Transcripts_nUMIs_USSs', 'No_of_Genes', 'No_of_Transcripts')
              find_not_needed = which(colnames(metadata) %in% cols_not_needed)
              
              if(length(find_not_needed) > 0) {
                metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed])
              }
              cols = c('Barcoded_Reads', 'Trimmed_Reads', 'Unmapped_Reads', 'Mapped_Reads', 'Uniquely_Mapped_Reads', 'Multimapped_Reads', 'Chimeric_Reads', 'No_of_Reads_w_UMIs', 'Reads_After_Dedup_nUMIs_USSs', 'Exon_nUMIs_USSs', 'Intron_nUMIs_USSs', 'Gene_nUMIs_USSs', 'Intergenic_Reads', 'Ribosomal_Reads', 'Mitochondrial_Reads', 'Exon_Reads', 'Intron_Reads', 'Gene_Reads')
              
              find = which(colnames(metadata) %in% cols)
              if(length(find) > 0) {
                
                metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
                
                
                columns_remove = c()
                for(x in 1:length(colnames(metadata_select)))
                {
                  number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                  if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                    getcol = colnames(metadata_select)[x]
                    columns_remove = c(columns_remove, getcol)
                  }
                }
                
                metadata_select = metadata_select %>% dplyr::select(-columns_remove)
                
                average_stats_cells = round(colMeans(metadata_select, na.rm = T))
                average_stats_cells = data.frame(average_stats_cells)
                colnames(average_stats_cells) = "Average Stats across Cells"
                average_stats_cells$`Average Stats across Cells` = formatC(as.numeric(average_stats_cells$`Average Stats across Cells`), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                total_stats_cells = round(colSums(metadata_select, na.rm = T))
                total_stats_cells = data.frame(total_stats_cells)
                colnames(total_stats_cells) = "Total Stats across Cells"
                total_stats_cells$Percentage = round((total_stats_cells$`Total Stats across Cells`/total_stats_cells$`Total Stats across Cells`[1])*100, 2)
                total_stats_cells_forplot = total_stats_cells
                total_stats_cells_forplot$values = as.numeric(total_stats_cells_forplot$`Total Stats across Cells`)
                total_stats_cells$`Total Stats across Cells` = formatC(as.numeric(total_stats_cells$`Total Stats across Cells`), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                
                if('Strand_Specificity' %in% colnames(metadata)) {
                  Strand_Specificity_avg = mean(metadata$Strand_Specificity, na.rm = T)
                  Strand_Specificity_avg = data.frame(Strand_Specificity_avg)
                  colnames(Strand_Specificity_avg) = "Average Stats across Cells"
                  rownames(Strand_Specificity_avg) = "Strand_Specificity"
                  average_stats_cells = rbind(average_stats_cells, Strand_Specificity_avg)
                  
                  Strand_Specificity_total = cbind(NA, NA)
                  Strand_Specificity_total = data.frame(Strand_Specificity_total)
                  colnames(Strand_Specificity_total) = c("Total Stats across Cells", "Percentage")
                  rownames(Strand_Specificity_total) = "Strand_Specificity"
                  total_stats_cells = rbind(total_stats_cells, Strand_Specificity_total)
                }
                
                
                numberofcells = cbind(formatC(as.numeric(ncol(seuratObject)), big.mark = ",", big.interval = 3, format = "f", digits = 0), NA)
                numberofcells = data.frame(numberofcells)
                colnames(numberofcells) = c("Total Stats across Cells", "Percentage")
                rownames(numberofcells) = "Number of Cells/Samples"
                numberofcells_avg = NA
                numberofcells_avg = data.frame(numberofcells_avg)
                colnames(numberofcells_avg) = "Average Stats across Cells"
                rownames(numberofcells_avg) = "Number of Cells/Samples"
                
                numberoffeatures = cbind(formatC(as.numeric(nrow(seuratObject)), big.mark = ",", big.interval = 3, format = "f", digits = 0), NA)
                numberoffeatures = data.frame(numberoffeatures)
                colnames(numberoffeatures) = c("Total Stats across Cells", "Percentage")
                rownames(numberoffeatures) = "Number of Genes/Features"
                numberoffeatures_avg = NA
                numberoffeatures_avg = data.frame(numberoffeatures_avg)
                colnames(numberoffeatures_avg) = "Average Stats across Cells"
                rownames(numberoffeatures_avg) = "Number of Genes/Features"
                
                average_stats_cells = rbind(average_stats_cells, numberoffeatures_avg, numberofcells_avg)
                total_stats_cells = rbind(total_stats_cells, numberoffeatures, numberofcells)
                df_show = cbind(total_stats_cells, average_stats_cells)
                df_temp = df_show
                df_temp$check_values = df_show$`Total Stats across Cells`
                colnames(df_show) = c("Total_stats", "Percentage_of_Barcoded_Reads", "Average_stats")
              } else {
                
                avg_nfeature = cbind('Average nFeature_RNA', formatC(round(mean(seuratObject@meta.data$nFeature_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                avg_nCount = cbind('Average nCount_RNA', formatC(round(mean(seuratObject@meta.data$nCount_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                avg_percentmt = cbind('Average percent-mt', formatC(round(mean(seuratObject@meta.data$percent.mt, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                numberofcells = cbind('Number of Cells/Samples', formatC(ncol(seuratObject), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                numberoffeatures = cbind('Number of Genes/Features', formatC(nrow(seuratObject), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                df_show = rbind(numberofcells, numberoffeatures, avg_nfeature, avg_nCount, avg_percentmt)
                df_show = data.frame(df_show)
                colnames(df_show) = c("Seurat_Object_Stats", "Value")
                df_temp = df_show
              }
              
              df_show = data.frame(df_show)
              df_show[df_show == ""] <- NA
              df_show[df_show == "NaN"] <- NA
              df_show[df_show == 0] <- NA
              df_show = data.frame(df_show)
              DT::datatable(df_show, options = list(pageLength = 20, dom = "t", rowCallback = JS(rowCallback)))
            })
            
            shiny::observeEvent(input$reset_btn, {
              
              output$seurat <- DT::renderDataTable({
                
                rowCallback <- c(
                  "function(row, data){",
                  "  for(var i=0; i<data.length; i++){",
                  "    if(data[i] === null){",
                  "      $('td:eq('+i+')', row).html('NA')",
                  "        .css({'color': 'rgb(151,151,151)', 'font-style': 'italic'});",
                  "    }",
                  "  }",
                  "}"  
                )
                
                cols_not_needed = c('orig.ident', 'Sample', 'No_of_Genes')
                find_not_needed = which(colnames(metadata) %in% cols_not_needed)
                
                if(length(find_not_needed) > 0) {
                  metadata = metadata %>% dplyr::select(!colnames(metadata)[find_not_needed])
                }
                
                cols = c('Barcoded_Reads', 'Trimmed_Reads', 'Unmapped_Reads', 'Mapped_Reads', 'Uniquely_Mapped_Reads', 'Multimapped_Reads', 'Chimeric_Reads', 'No_of_Reads_w_UMIs', 'Reads_After_Dedup_nUMIs_USSs', 'Exon_nUMIs_USSs', 'Intron_nUMIs_USSs', 'Gene_nUMIs_USSs', 'Intergenic_Reads', 'Ribosomal_Reads', 'Mitochondrial_Reads', 'Exon_Reads', 'Intron_Reads', 'Gene_Reads')
                
                find = which(colnames(metadata) %in% cols)
                if(length(find) > 0) {
                  
                  metadata_select = metadata %>% dplyr::select(colnames(metadata)[find])
                  
                  
                  columns_remove = c()
                  for(x in 1:length(colnames(metadata_select)))
                  {
                    number_of_NA = sum(is.na(metadata_select[[colnames(metadata_select)[x]]]))
                    if(number_of_NA == length(metadata_select[[colnames(metadata_select)[x]]])) {
                      getcol = colnames(metadata_select)[x]
                      columns_remove = c(columns_remove, getcol)
                    }
                  }
                  
                  metadata_select = metadata_select %>% dplyr::select(-columns_remove)
                  
                  average_stats_cells = round(colMeans(metadata_select, na.rm = T))
                  average_stats_cells = data.frame(average_stats_cells)
                  colnames(average_stats_cells) = "Average Stats across Cells"
                  average_stats_cells$`Average Stats across Cells` = formatC(as.numeric(average_stats_cells$`Average Stats across Cells`), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  total_stats_cells = round(colSums(metadata_select, na.rm = T))
                  total_stats_cells = data.frame(total_stats_cells)
                  colnames(total_stats_cells) = "Total Stats across Cells"
                  total_stats_cells$Percentage = round((total_stats_cells$`Total Stats across Cells`/total_stats_cells$`Total Stats across Cells`[1])*100, 2)
                  total_stats_cells_forplot = total_stats_cells
                  total_stats_cells_forplot$values = as.numeric(total_stats_cells_forplot$`Total Stats across Cells`)
                  total_stats_cells$`Total Stats across Cells` = formatC(as.numeric(total_stats_cells$`Total Stats across Cells`), big.mark = ",", big.interval = 3, format = "f", digits = 0)
                  
                  if('Strand_Specificity' %in% colnames(metadata)) {
                    Strand_Specificity_avg = mean(metadata$Strand_Specificity, na.rm = T)
                    Strand_Specificity_avg = data.frame(Strand_Specificity_avg)
                    colnames(Strand_Specificity_avg) = "Average Stats across Cells"
                    rownames(Strand_Specificity_avg) = "Strand_Specificity"
                    average_stats_cells = rbind(average_stats_cells, Strand_Specificity_avg)
                    
                    Strand_Specificity_total = cbind(NA, NA)
                    Strand_Specificity_total = data.frame(Strand_Specificity_total)
                    colnames(Strand_Specificity_total) = c("Total Stats across Cells", "Percentage")
                    rownames(Strand_Specificity_total) = "Strand_Specificity"
                    total_stats_cells = rbind(total_stats_cells, Strand_Specificity_total)
                  }
                  
                  
                  numberofcells = cbind(formatC(as.numeric(ncol(seuratObject)), big.mark = ",", big.interval = 3, format = "f", digits = 0), NA)
                  numberofcells = data.frame(numberofcells)
                  colnames(numberofcells) = c("Total Stats across Cells", "Percentage")
                  rownames(numberofcells) = "Number of Cells/Samples"
                  numberofcells_avg = NA
                  numberofcells_avg = data.frame(numberofcells_avg)
                  colnames(numberofcells_avg) = "Average Stats across Cells"
                  rownames(numberofcells_avg) = "Number of Cells/Samples"
                  
                  numberoffeatures = cbind(formatC(as.numeric(nrow(seuratObject)), big.mark = ",", big.interval = 3, format = "f", digits = 0), NA)
                  numberoffeatures = data.frame(numberoffeatures)
                  colnames(numberoffeatures) = c("Total Stats across Cells", "Percentage")
                  rownames(numberoffeatures) = "Number of Genes/Features"
                  numberoffeatures_avg = NA
                  numberoffeatures_avg = data.frame(numberoffeatures_avg)
                  colnames(numberoffeatures_avg) = "Average Stats across Cells"
                  rownames(numberoffeatures_avg) = "Number of Genes/Features"
                  
                  average_stats_cells = rbind(average_stats_cells, numberoffeatures_avg, numberofcells_avg)
                  total_stats_cells = rbind(total_stats_cells, numberoffeatures, numberofcells)
                  df_show = cbind(total_stats_cells, average_stats_cells)
                  df_temp = df_show
                  df_temp$check_values = df_show$`Total Stats across Cells`
                  colnames(df_show) = c("Total_stats", "Percentage_of_Barcoded_Reads", "Average_stats")
                } else {
                  
                  avg_nfeature = cbind('Average nFeature_RNA', formatC(round(mean(seuratObject@meta.data$nFeature_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  avg_nCount = cbind('Average nCount_RNA', formatC(round(mean(seuratObject@meta.data$nCount_RNA, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  avg_percentmt = cbind('Average percent-mt', formatC(round(mean(seuratObject@meta.data$percent.mt, na.rm = T)), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  numberofcells = cbind('Number of Cells/Samples', formatC(ncol(seuratObject), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  numberoffeatures = cbind('Number of Genes/Features', formatC(nrow(seuratObject), big.mark = ",", big.interval = 3, format = "f", digits = 0))
                  df_show = rbind(numberofcells, numberoffeatures, avg_nfeature, avg_nCount, avg_percentmt)
                  df_show = data.frame(df_show)
                  colnames(df_show) = c("Seurat_Object_Stats", "Value")
                  df_temp = df_show
                }
                
                df_show = data.frame(df_show)
                df_show[df_show == ""] <- NA
                df_show[df_show == "NaN"] <- NA
                df_show[df_show == 0] <- NA
                df_show = data.frame(df_show)
                DT::datatable(df_show, options = list(pageLength = 20, dom = "t", rowCallback = JS(rowCallback)))
              })
              
            })
            message("Seurat v5 object creation is completed!")
            data(seuratObject) # Update reactive value
          } else {
            message ("Fatal: Seurat object creation failed; Please check your inputs")
            stop("Error: Seurat object creation failed: Fatal")
          }
          incProgress(0.1, detail = 'checks complete, stats ready')
          timeEnd <- Sys.time()
          elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')
          gc()
        })
      }, error = function(e) {
        shinyjs::enable("upload")
        shinyalert("Data preparation failed", "Please check your inputs and parameters; Some possible errors: Metadata cells | data format | required input is empty", type = "error", confirmButtonCol = "#1c82e0")
      })
    })
    return(data)
  })
}