#' Decontamination Server Logic
#' 
#' Handles server-side logic for performing decontamination of ambient RNA 
#' using the DecontX package within a Shiny module. It triggers the decontamination 
#' process upon user action, updates the Seurat object with decontamination results,
#' and visualizes the output.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @seealso \code{\link[decontX]{decontX}}: For more information about the DecontX package, see its documentation.
#'
#' @return
#' @export
#' 
#' @examples
#' decontaminationServer("DecontaminationModule", seuratObject)
#' Decontamination Server Logic
#' 
#' Handles server-side logic for performing decontamination of ambient RNA 
#' using the DecontX package within a Shiny module. It triggers the decontamination 
#' process upon user action, updates the Seurat object with decontamination results,
#' and visualizes the output.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @seealso \code{\link[decontX]{decontX}}: For more information about the DecontX package, see its documentation.
#'
#' @return
#' @export
#' 
#' @examples
#' decontaminationServer("DecontaminationModule", seuratObject)
decontaminationServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    # Reactive values to hold plot data
    before_decontamination <- reactiveVal(NULL)
    after_decontamination <- reactiveVal(NULL)
    barcode_plot_before_filter <- reactiveVal(NULL)
    barcode_plot_after_filter <- reactiveVal(NULL)
    decontX_UMAP <- reactiveVal(NULL)
    temp_object <- reactiveVal(NULL)
    
    # Reactive value to keep track of the current action (skip or perform)
    current_action <- reactiveVal("")
    
    # Reactive value to store decontamination results
    decontamination_results <- reactiveVal(NULL)
    
    observe({
      req(seuratObj())
      ############### Dry run to estimate clusters ###############
      withProgress(message = 'Starting dry run for ambient RNA estimation:', {
        seurat_nochange = seuratObj()
        # Check if any of the columns is completely zero
        indx <- which(seurat_nochange$nCount_RNA > 0)
        seurat_nochange <- seurat_nochange[ ,indx]
        DefaultAssay(seurat_nochange) = "RNA"
        Idents(seurat_nochange) = "orig.ident"
        parameterList = seurat_nochange@assays$user_parameter_tracker
        seurat_nochange@assays$user_parameter_tracker = NULL
        seurat_nochange_sub = seurat_nochange
        DefaultAssay(seurat_nochange_sub) = "RNA"
        incProgress(0.4, detail = 'estimating clusters..please wait')
        seurat_nochange_sub = NormalizeData(seurat_nochange_sub)
        seurat_nochange_sub = FindVariableFeatures(seurat_nochange_sub)
        seurat_nochange_sub = ScaleData(seurat_nochange_sub)
        seurat_nochange_sub = RunPCA(seurat_nochange_sub, assay = "RNA", npcs = 10)
        seurat_nochange_sub = FindNeighbors(seurat_nochange_sub, dims = 1:10)
        seurat_nochange_sub = FindClusters(seurat_nochange_sub, resoulution = 0.5)
        if(ncol(seurat_nochange_sub) <= 50) {
          seurat_nochange_sub = RunUMAP(seurat_nochange_sub, dims = 1:10, n.neighbors = 5)
        } else {
          seurat_nochange_sub = RunUMAP(seurat_nochange_sub, dims = 1:10)
        }
        incProgress(0.2, detail = 'estimating clusters: finished')
        seurat_nochange[['Original_RNA']] <- CreateAssay5Object(counts = seurat_nochange_sub[['RNA']]$counts, data = seurat_nochange_sub[['RNA']]$data)
        VariableFeatures(seurat_nochange[['Original_RNA']]) = VariableFeatures(seurat_nochange_sub)
        seurat_nochange[['Original_RNA']]@misc$clusters_orig = data.frame(seurat_nochange_sub@meta.data$seurat_clusters)
        rownames(seurat_nochange[['Original_RNA']]@misc$clusters_orig) = Cells(seurat_nochange_sub)
        colnames(seurat_nochange[['Original_RNA']]@misc$clusters_orig) = "clusters_orig"
        shinyalert("Data preparation is complete", paste("Number of cell:", ncol(seuratObj()), "\n" ,"Number of genes:", nrow(seuratObj()), "\n" ,"Let's Get Started with Ambient RNA Filtering"), type = "success", confirmButtonCol = "#1c82e0")
        incProgress(0.4, detail = 'dry run complete')

        DefaultAssay(seurat_nochange) = "RNA"
        Idents(seurat_nochange) = "orig.ident"
        seurat_nochange@assays$user_parameter_tracker <- parameterList
        before_decontamination(seurat_nochange)
        after_decontamination(seurat_nochange)
        temp_object(seurat_nochange_sub)
        # Initialize and render the skip plot without button press
        current_action("skip")
        barcode_plot_before_filter({  
          DotPlot(temp_object(), VariableFeatures(temp_object())[1:10], dot.scale = 10) + RotatedAxis() + ggtitle('Before Ambient RNA Filtering')
        })
      })
      
      output$barcode_inflection_before_filtering_with_no <- renderPlot({
        barcode_plot_before_filter()
      })

      # Show/hide relevant download buttons
      shinyjs::show("downloadPlot_kneenofilter")
      shinyjs::hide("downloadPlot_kneefilter")
      shinyjs::hide("downloadPlot_decontxumap")
      
      output$downloadPlot_kneenofilter <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_DotPlot_HVFs_Before_AmbientRNA_Correction", "-", Sys.Date(), input$kneenofilter)
        },
        content = function(file) {
          ggsave(file, plot = barcode_plot_before_filter(), width = 10, height = 10)
        },
        contentType = 'image'
      )
      
      # Reset the plots for decontamination
      barcode_plot_after_filter(NULL)
      decontX_UMAP(NULL)
    })
    
    observeEvent(input$skipDecontamination, {
      
      # Alert to notify user that no decontamination have been performed
      shinyalert("Skip ambient rna decontamination", "No decontamination have been performed.", type = "info", confirmButtonCol = "#1c82e0")  # Notify user that QC was skipped
      
      current_action("skip")
      req(seuratObj(), temp_object())
      #sc_nochange <- seuratObj()
      sc_nochange <- before_decontamination()
      parameterList = sc_nochange@assays$user_parameter_tracker
      ambient_RNA = paste0('Ambient RNA correction: ', "Skipped")
      parameterList = c(parameterList, ambient_RNA)
      sc_nochange@assays$user_parameter_tracker = parameterList
      before_decontamination(sc_nochange)
      after_decontamination(sc_nochange)
      
      # Generate plot for skipping decontamination
      barcode_plot_before_filter({
        DotPlot(temp_object(), VariableFeatures(temp_object())[1:10], dot.scale = 10) + RotatedAxis() + ggtitle('Before Ambient RNA Filtering')
      })
      
      output$barcode_inflection_before_filtering_with_no <- renderPlot({
        barcode_plot_before_filter()
      })
      
      # Show/hide relevant download buttons
      shinyjs::show("downloadPlot_kneenofilter")
      shinyjs::show("kneenofilter")
      shinyjs::hide("downloadPlot_kneefilter")
      shinyjs::hide("downloadPlot_decontxumap")
      shinyjs::hide("kneefilter")
      shinyjs::hide("decontxumap")
      
      output$downloadPlot_kneenofilter <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_DotPlot_HVFs_Before_AmbientRNA_Correction", "-", Sys.Date(), input$kneenofilter)
        },
        content = function(file) {
          ggsave(file, plot = barcode_plot_before_filter(), width = 10, height = 10)
        },
        contentType = 'image'
      )
      
      # Reset the plots for decontamination
      barcode_plot_after_filter(NULL)
      decontX_UMAP(NULL)
    })

    observeEvent(input$performDecontamination, {
      current_action("perform")
      req(before_decontamination(), seuratObj(), temp_object())
      if (is.null(decontamination_results())) {
        withProgress(
          message = 'Performing Decontamination using DecontX:',
          value = 0, {
            incProgress(0.1, detail = 'initializing decontX')
            
            updatedSeuratObject <- before_decontamination()
            updatedSeuratObject@assays$user_parameter_tracker = NULL
            query <- as.SingleCellExperiment(updatedSeuratObject, assay = "RNA")
            estimated_clusters = updatedSeuratObject[['Original_RNA']]@misc$clusters_orig$clusters_orig
            incProgress(0.4, detail = 'Starting decontamination')
            if((length(unique(estimated_clusters)) > 1)&&(!is.null(estimated_clusters))) {
              query <- decontX::decontX(query, z = estimated_clusters)
              updatedSeuratObject_use = seuratObj()
              parameterList = updatedSeuratObject_use@assays$user_parameter_tracker
              ambient_RNA = paste0('Ambient RNA correction: ', "Performed")
              parameterList = c(parameterList, ambient_RNA)
              updatedSeuratObject[['RNA']] <- CreateAssay5Object(counts = round(decontX::decontXcounts(query)))
              DefaultAssay(updatedSeuratObject) <- "RNA"
              Idents(updatedSeuratObject) = "orig.ident"
              updatedSeuratObject@meta.data$DecontX_contamination <- query@metadata$decontX$contamination
              check_Mito_presence_human = grep("^MT-", rownames(updatedSeuratObject), value = TRUE, ignore.case = FALSE)
              check_Mito_presence_mouse = grep("^mt-", rownames(updatedSeuratObject), value = TRUE, ignore.case = FALSE)
              if(length(check_Mito_presence_human) > 0) {
                updatedSeuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(updatedSeuratObject, pattern = "^MT-")
              }
              if(length(check_Mito_presence_mouse) > 0) {
                updatedSeuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(updatedSeuratObject, pattern = "^mt-")
              }
              if((length(check_Mito_presence_human) == 0) && (length(check_Mito_presence_mouse) == 0)) {
                updatedSeuratObject[["percent.mt"]] = rep(0, ncol(updatedSeuratObject))
              }
              updatedSeuratObject@assays$user_parameter_tracker = parameterList
              decontamination_results(list(query = query, updatedSeuratObject = updatedSeuratObject))
              incProgress(0.5, detail = 'DecontX: decontamination complete')
              after_decontamination(updatedSeuratObject)
            } else {
              updatedSeuratObject_use = before_decontamination()
              parameterList = updatedSeuratObject_use@assays$user_parameter_tracker
              ambient_RNA = paste0('Ambient RNA correction: ', "Skipped")
              parameterList = c(parameterList, ambient_RNA)
              updatedSeuratObject@meta.data$DecontX_contamination <- rep(0, nrow(updatedSeuratObject@meta.data))
              query@metadata$decontX$contamination = rep(0, nrow(updatedSeuratObject@meta.data))
              check_Mito_presence_human = grep("^MT-", rownames(updatedSeuratObject), value = TRUE, ignore.case = FALSE)
              check_Mito_presence_mouse = grep("^mt-", rownames(updatedSeuratObject), value = TRUE, ignore.case = FALSE)
              if(length(check_Mito_presence_human) > 0) {
                updatedSeuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(updatedSeuratObject, pattern = "^MT-")
              }
              if(length(check_Mito_presence_mouse) > 0) {
                updatedSeuratObject[["percent.mt"]] <- Seurat::PercentageFeatureSet(updatedSeuratObject, pattern = "^mt-")
              }
              if((length(check_Mito_presence_human) == 0) && (length(check_Mito_presence_mouse) == 0)) {
                updatedSeuratObject[["percent.mt"]] = rep(0, ncol(updatedSeuratObject))
              }
              updatedSeuratObject@assays$user_parameter_tracker = parameterList
              #colData(query)$decontX_contamination = rep(0, nrow(colData(query)))
              #query@metadata$decontX$contamination = rep(0, nrow(colData(query)))
              updatedSeuratObject@meta.data$DecontX_contamination <- rep(0, nrow(updatedSeuratObject@meta.data))
              decontamination_results(list(query = query, updatedSeuratObject = updatedSeuratObject))
              incProgress(0.5, detail = 'DecontX: decontamination complete')
              after_decontamination(updatedSeuratObject)
              shinyalert("Decontamination process not needed, DecontX needs ≥2 clusters", "Only 1 cluster found", type = "warning", confirmButtonCol = "#1c82e0")
            }
            
          }
        )
      } else {
        after_decontamination(decontamination_results()$updatedSeuratObject)
      }
      
      barcode_plot_after_filter({
        withProgress(
          message = 'Generating plots:',
          value = 0, {
            incProgress(0.4, detail = 'getting before and after decontamination data')
            updatedSeuratObject_new <- after_decontamination()
            updatedSeuratObject_new@assays$user_parameter_tracker = NULL
            temp_object = temp_object()
            DefaultAssay(updatedSeuratObject_new) = "RNA"
            updatedSeuratObject_new = NormalizeData(updatedSeuratObject_new)
            VariableFeatures(updatedSeuratObject_new) = VariableFeatures(temp_object)
            updatedSeuratObject_new = ScaleData(updatedSeuratObject_new)
            updatedSeuratObject_new$clusters_temp = temp_object$seurat_clusters
            #p1 + p2
            Idents(updatedSeuratObject_new) = "clusters_temp"
            Idents(temp_object) = "seurat_clusters"
            if(length(unique(temp_object$seurat_clusters)) > 1) {
              p1=DotPlot(temp_object, VariableFeatures(temp_object)[1:10], dot.scale = 10) + RotatedAxis() + guides(size = guide_legend(order = 1, title = "Percent Expressed"),
                                                                                                                    color = guide_colorbar(order = 2, title = "Average Expression")) + ggtitle('Before Ambient RNA Filtering') 
              p2=DotPlot(updatedSeuratObject_new, VariableFeatures(temp_object)[1:10], dot.scale = 10) + RotatedAxis() + guides(size = guide_legend(order = 1, title = "Percent Expressed"),
                                                                                                                                color = guide_colorbar(order = 2, title = "Average Expression")) + ggtitle('After Ambient RNA Filtering') 
              incProgress(0.4, detail = 'combining plots')
              incProgress(0.2, detail = 'finished')
              p1 + p2
            } else {
              current_action("skip")
            }
        })
      })
 
      output$barcode_inflection_before_and_after_filtering <- renderPlot({
        barcode_plot_after_filter()
      })
      
      decontX_UMAP({
        #if('decontX_contamination' %in% colnames(colData(decontamination_results()$query))) {
          #plotDecontXContamination(decontamination_results()$query) + 
          #theme(axis.text = element_text(colour = "black", size = 15, face = "bold"),
                #axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right",
                #legend.box = "vertical", panel.background = element_rect(colour = "white"))
        #}
        updatedSeuratObject_new <- after_decontamination()
        temp_object = temp_object()
        temp_object$DecontX_contamination = updatedSeuratObject_new$DecontX_contamination
        range_use <- function(x){
          if((min(x) == 0) && (max(x) == 0)) {
            x
          } else {
            (x-min(x))/(max(x)-min(x))
          }
        }
        temp_object$DecontX_contamination = range_use(temp_object$DecontX_contamination)
        Idents(temp_object) = "seurat_clusters"
        #cols = c("blue", "green", "yellow", "orange", "red")
        FeaturePlot(temp_object, "DecontX_contamination") + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
      })

     
      output$decontX_UMAP <- renderPlot({
        decontX_UMAP()
      })
      
      shinyjs::hide("downloadPlot_kneenofilter")
      shinyjs::show("downloadPlot_kneefilter")
      shinyjs::show("downloadPlot_decontxumap")
      shinyjs::show("kneefilter")
      shinyjs::show("decontxumap")
      
      output$downloadPlot_kneefilter <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_DotPlot_HVFs_After_AmbientRNA_Correction", "-", Sys.Date(), input$kneefilter)
        },
        content = function(file) {
          ggsave(file, plot = barcode_plot_after_filter(), width = 20, height = 10)
        },
        contentType = 'image/png'
      )
      
      output$downloadPlot_decontxumap <- downloadHandler(
        filename = function() {
          paste0("CogentDS_v2.0_DecontX_UMAP_AmbientRNA_Contamination", "-", Sys.Date(), input$decontxumap)
        },
        content = function(file) {
          ggsave(file, plot = decontX_UMAP(), width = 10, height = 10)
        },
        contentType = 'image/png'
      )
    })
    
    output$dynamicPlots <- renderUI({
      if (current_action() == "skip") {
        tagList(
          tabsetPanel(type = "tabs",
                      tabPanel("Ambient RNA plot",
                               downloadButton(ns('downloadPlot_kneenofilter'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                               selectInput(ns("kneenofilter"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                               shiny::plotOutput(ns("barcode_inflection_before_filtering_with_no"),
                                                 width = "500", height = "400"),
                               h5("Dot plots are generated based on clustering from all cells. 
                                  Highly variable features are estimated, and the top 10 features are visualized.")
                      )
          )
        )
      } else if (current_action() == "perform") {
        tagList(
          tabsetPanel(type = "tabs",
                      tabPanel("Ambient RNA plot",
                               selectInput(ns("kneefilter"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                               downloadButton(ns('downloadPlot_kneefilter'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                               shiny::plotOutput(ns("barcode_inflection_before_and_after_filtering"),
                                                 width = "900", height = "400")),
                      h5("Dot plots are generated based on clustering from all cells. 
                                  Highly variable features are estimated, and the top 10 features are visualized."),
                      tabPanel("Contamination Fraction Plot",
                               selectInput(ns("decontxumap"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                               downloadButton(ns('downloadPlot_decontxumap'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                               shiny::plotOutput(ns("decontX_UMAP"), width = "700", height = "600"))
          )
        )
      }
    })
    return(after_decontamination)
  })
}
