#' Differential Gene Expression Analysis Server Module
#' 
#' This server module handles the differential expression (DE) analysis within a Shiny 
#' application, leveraging Seurat for the analysis. Users can select DE analysis methods 
#' and parameters through a UI, and this module processes those inputs to perform DE 
#' analysis on a Seurat object. It supports various analysis methods, including finding 
#' markers for individual identities and all markers within the dataset.
#' Results can be downloaded and visualizations such as volcano plots and heatmaps can 
#' be generated based on the selected method. 
#' 
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @param seuratObject A reactive expression containing the Seurat object.
#'
#' @seealso \code{\link[Seurat]{FindMarkers}}, and  \code{\link[Seurat]{FindAllMarkers}}
#'          for more information on the methods used for DE analysis within Seurat.
#' 
#' @examples
# Differential Expression Analysis Server Logic
diffExpressionServer <- function(id, seuratObject) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    # Store original identities
    originalIdents <- reactiveVal(NULL)
    
    # Store clusters/Celltypes being compared
    comparedIdents <- reactiveVal(NULL)
    
    # Define seurat copy
    DESeuratObj <- reactiveVal(NULL)
    
    observe({
      DESeuratObj(seuratObject())
    })
    
    # Reset DE results when identity type changes
    observeEvent(input$identityType, {
      deResults(NULL)
      volcanoplt(NULL)  
      heatmapplt(NULL)
    })
    
    observeEvent(input$deAnalysisMethod, {
      volcanoplt(NULL)  
      heatmapplt(NULL)
    })
    
    # Render DE analysis parameters based on selected method
    shiny::observe({
      req(DESeuratObj())
      req(input$identityType)
      req(input$deAnalysisMethod)
      UpdateSeuratObject <- DESeuratObj()
      
      # Check if "Clusters", "Cell Types", or "Custom Selection" is selected
      if (input$identityType == "Clusters") {
        if (!is.null(originalIdents())) {
          Idents(UpdateSeuratObject) <- originalIdents()
          originalIdents(NULL)  # Reset originalIdents after reverting
        } else {
          Idents(UpdateSeuratObject) <- DESeuratObj()@active.ident
        }
        
        # Store clusters/Celltypes being compared
        comparedIdents(c(input$ident1, input$ident2))
        
      } else if (input$identityType == "Cell Type" && "SingleR_Celltypes" %in% colnames(UpdateSeuratObject@meta.data)) {
        # Store original identities before updating
        if (is.null(originalIdents())) {
          originalIdents(DESeuratObj()@active.ident)
        }
        Idents(UpdateSeuratObject) <- "SingleR_Celltypes"
        
        # Store clusters/Celltypes being compared
        comparedIdents(c(input$ident1, input$ident2))
        
      }
      else if (input$identityType == "Custom Selection Clusters" && 
               "custom_clusters" %in% colnames(UpdateSeuratObject@meta.data)) {
        
        # Store original identities before updating
        if (is.null(originalIdents())) {
          originalIdents(DESeuratObj()@active.ident)
        }
        Idents(UpdateSeuratObject) <- "custom_clusters"
        
        # Store clusters/Celltypes being compared
        comparedIdents(c(input$ident1, input$ident2))
        
      } else if (input$identityType == "Custom Selection Cell Types" && 
                 "custom_cell_types" %in% colnames(UpdateSeuratObject@meta.data)) {
        if (is.null(originalIdents())) {
          originalIdents(DESeuratObj()@active.ident)
        }
        Idents(UpdateSeuratObject) <- "custom_cell_types"
        
        # Store clusters/Celltypes being compared
        comparedIdents(c(input$ident1, input$ident2))
      }
      DESeuratObj(UpdateSeuratObject)
    })
    
    # Reactive value to hold metadata columns
    metadataColumns <- reactive({
      req(DESeuratObj())
      
      # Get the names of  metadata columns
      all_metadata <- colnames(DESeuratObj()@meta.data)
      
      # Filter out numeric columns
      relevant_columns <- all_metadata[sapply(DESeuratObj()@meta.data[, all_metadata], is.factor)]
      
      # Return the filtered relevant columns
      relevant_columns
    })
    
    # Conditional download DE results button
    output$downloadDEResultsUI <- renderUI({
      if (!is.null(deResults()) && nrow(deResults()) > 0) {
        downloadButton(ns("downloadDEResults"), "Download DE Results", style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Reactive value to ensure alerts is trigger only once when 
    # custom clusters not present or if they are equal to Seurat clusters
    alert_triggered <- reactiveVal(FALSE)
    
    # Reset alert trigger whenever identity type or method changes
    observeEvent(list(input$identityType, input$deAnalysisMethod), {
      alert_triggered(FALSE)
    })
    
    
    # Initialize deResults
    deResults <- reactiveVal(list())
    
    # Initialize last executed method
    LastExecutedMethod <- reactiveVal(NULL)
    
    # Render DE analysis parameters based on selected method
    output$deParams <- renderUI({
      
      # Check if Cell Type is selected but no SingleR cell types are available
      if (input$identityType == "Cell Type" && !"SingleR_Celltypes" %in% colnames(DESeuratObj()@meta.data)) {
        shinyalert("Error", "Run cell type annotation annotation module", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Check if clusters is selected but no data available
      if (!alert_triggered()) {
        if (input$deAnalysisMethod == "FindMarkers" && input$identityType == "Custom Selection Clusters") {
          if (!"custom_clusters" %in% colnames(DESeuratObj()@meta.data)) {
            shinyalert("Error", "Run cluster module and custom lasso selection", type = "error", confirmButtonCol = "#1c82e0")
            alert_triggered(TRUE)
            return(NULL)
          }
          if ("custom_clusters" %in% colnames(DESeuratObj()@meta.data) && all(DESeuratObj()$seurat_clusters == DESeuratObj()$custom_clusters)) {
            shinyalert("Error", "Run cluster module and custom lasso selection", type = "error", confirmButtonCol = "#1c82e0")
            alert_triggered(TRUE)
            return(NULL)
          }
        }
      }
      
      # Check if Cell Type is selected but no SingleR cell types are available
      if (input$deAnalysisMethod == "FindMarkers" 
          && input$identityType == "Custom Selection Cell Types" 
          && !"custom_cell_types" %in% colnames(DESeuratObj()@meta.data)) {
        shinyalert("Error", "Run cell type annotation and custom lasso selection", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      # Check if DE analysis method is not supported for custom selections
      if (input$deAnalysisMethod %in% c("FindAllMarkers") && 
          input$identityType %in% c("Custom Selection Clusters", "Custom Selection Cell Types")) {
        shinyalert("Error", "Custom cell and clusters only supported by FindMarkers", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      req(DESeuratObj(), input$deAnalysisMethod, input$deAnalysisMethod)
      LastExecutedMethod(input$deAnalysisMethod)
      
      # Collect metadata column names for grouping options
      metadataChoices <- c("No Grouping" = "", metadataColumns())
      
      # Determine if ident1 and ident2 should be displayed
      showIdents <- input$deAnalysisMethod != "FindAllMarkers"
      
      # Conditional UI elements based on selected DE analysis method
      if (input$identityType == "Clusters") {
        tagList(
          if (showIdents) {
            popify(selectInput(ns("ident1"), "Identity 1", choices = levels(DESeuratObj()@active.ident), multiple = T), "First Identity Class", "Specify the first identity class you want to compare", placement = "right", options = list(container = "body"))
          },
          if (showIdents) {
            popify(selectInput(ns("ident2"), "Identity 2", choices = levels(DESeuratObj()@active.ident), multiple = T), "second Identity Class", "Specify the second identity class you want to compare", placement = "right", options = list(container = "body"))
          },
          selectInput(ns("deMethod"), "DE analysis method", choices = c("wilcox" = "wilcox", "bimod" = "bimod", "roc" = "roc", 
                                                                        "LR" = "Logistic Regression", "t" = "Student's t", 
                                                                        "negbinom" = "Negative Binomial", 
                                                                        "poisson" = "Poisson", "MAST" = "MAST", "DESeq2" = "DESeq2")),
          bsPopover(ns("deMethod"), "DE method", "Test to use for performing DE Analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("logfcThreshold"), "Log Fold Change threshold", value = 0.1),
          bsPopover(ns("logfcThreshold"), "Log FC threshold", "Minimum fold difference (log-scale) required between two groups of cells to be included in the analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("minPct"), "Minimum percentage", value = 0.01),
          bsPopover(ns("minPct"), "Minimum percentage of cells", "Minimum fraction of cells in either of the groups where gene must be detected to be considered for testing", placement = "right", trigger = "hover", options = list(container = "body")),
          
          checkboxInput(ns("onlyPos"), "Only positive markers", value = FALSE),
          bsPopover(ns("onlyPos"), "Only positive markers", "Genes with only positive log FC will be considered", placement = "right", trigger = "hover", options = list(container = "body")),
          
          if (input$deAnalysisMethod == "FindMarkers") {
            selectInput(ns("groupby"), "Group by:", choices = metadataChoices, selected = "")
          },
          
          bsPopover(ns("groupby"), "Group by identity classes", "Allows to specify a metadata column to group cells prior to DE Analysis.", placement = "right", trigger = "hover", options = list(container = "body")),
          actionButton(ns("performDEAnalysisNow"), "Perform DE analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
          br(), br(),
          uiOutput(ns("downloadDEResultsUI"))
          
        )
      } else if (input$identityType == "Cell Type" && "SingleR_Celltypes" %in% colnames(DESeuratObj()@meta.data)) {
        tagList(
          if (showIdents) {
            popify(selectInput(ns("ident1"), "Identity 1 (Cell Types)", choices = levels(DESeuratObj()@active.ident), multiple = T), "First Identity Class", "Specify the first identity class you want to compare", placement = "right", options = list(container = "body"))
          },
          if (showIdents) {
            popify(selectInput(ns("ident2"), "Identity 2 (Cell Types)", choices = levels(DESeuratObj()@active.ident), multiple = T), "Second Identity Class", "Specify the second identity class you want to compare", placement = "right", options = list(container = "body"))
          },
          selectInput(ns("deMethod"), "DE analysis method", choices = c("wilcox" = "wilcox", "bimod" = "bimod", "roc" = "roc", 
                                                                        "LR" = "Logistic Regression", "t" = "Student's t", 
                                                                        "negbinom" = "Negative Binomial", 
                                                                        "poisson" = "Poisson", "MAST" = "MAST", "DESeq2" = "DESeq2")),
          bsPopover(ns("deMethod"), "DE method", "Test to use for performing DE Analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("logfcThreshold"), "Log Fold Change threshold", value = 0.1),
          bsPopover(ns("logfcThreshold"), "Log FC threshold", "Minimum fold difference (log-scale) required between two groups of cells to be included in the analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("minPct"), "Minimum percentage", value = 0.01),
          bsPopover(ns("minPct"), "Minimum percentage of cells", "Minimum fraction of cells in either of the groups where gene must be detected to be considered for testing", placement = "right", trigger = "hover", options = list(container = "body")),
          
          checkboxInput(ns("onlyPos"), "Only positive markers", value = FALSE),
          bsPopover(ns("onlyPos"), "Only positive markers", "Genes with only positive log FC will be considered", placement = "right", trigger = "hover", options = list(container = "body")),
          
          if (input$deAnalysisMethod == "FindMarkers") {
            selectInput(ns("groupby"), "Group by:", choices = metadataChoices, selected = "")
          },
          
          actionButton(ns("performDEAnalysisNow"), "Perform DE Analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
          br(), br(),
          uiOutput(ns("downloadDEResultsUI"))
          
        )
      } else if (input$deAnalysisMethod == "FindMarkers" && input$identityType == "Custom Selection Clusters" && "custom_clusters" %in% colnames(DESeuratObj()@meta.data)) {
        tagList(
          popify(selectInput(ns("ident1"), "Identity 1", choices = sort(levels(DESeuratObj()@active.ident)), multiple = T), "First Identity Class", "Specify the first identity class you want to compare", placement = "right", options = list(container = "body")),
          popify(selectInput(ns("ident2"), "Identity 2", choices = sort(levels(DESeuratObj()@active.ident)), multiple = T), "Second Identity Class", "Specify the second identity class you want to compare", placement = "right", options = list(container = "body")),
          selectInput(ns("deMethod"), "DE analysis method", choices = c("wilcox" = "wilcox", "bimod" = "bimod", "roc" = "roc", 
                                                                        "LR" = "Logistic Regression", "t" = "Student's t", 
                                                                        "negbinom" = "Negative Binomial", 
                                                                        "poisson" = "Poisson", "MAST" = "MAST", "DESeq2" = "DESeq2")),
          bsPopover(ns("deMethod"), "DE method", "Test to use for performing DE Analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("logfcThreshold"), "Log Fold Change threshold", value = 0.1),
          bsPopover(ns("logfcThreshold"), "Log FC threshold", "Minimum fold difference (log-scale) required between two groups of cells to be included in the analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("minPct"), "Minimum percentage", value = 0.01),
          bsPopover(ns("minPct"), "Minimum percentage of cells", "Minimum fraction of cells in either of the groups where gene must be detected to be considered for testing", placement = "right", trigger = "hover", options = list(container = "body")),
          
          checkboxInput(ns("onlyPos"), "Only positive markers", value = FALSE),
          bsPopover(ns("onlyPos"), "Only positive markers", "Genes with only positive log FC will be considered", placement = "right", trigger = "hover", options = list(container = "body")),
          
          selectInput(ns("groupby"), "Group by:", choices = metadataChoices, selected = ""),
          bsPopover(ns("groupby"), "Group by identity classes", "Allows to specify a metadata column to group cells prior to DE Analysis.", placement = "right", trigger = "hover", options = list(container = "body")),
          actionButton(ns("performDEAnalysisNow"), "Perform DE analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
          br(), br(),
          uiOutput(ns("downloadDEResultsUI"))
        )
        
      } else if (input$deAnalysisMethod == "FindMarkers" && input$identityType == "Custom Selection Cell Types" && "custom_cell_types" %in% colnames(DESeuratObj()@meta.data)) {
        tagList(
          popify(selectInput(ns("ident1"), "Identity 1 (Cell Types)", choices = sort(levels(DESeuratObj()@active.ident)), multiple = T), "First Identity Class", "Specify the first identity class you want to compare", placement = "right", options = list(container = "body")),
          popify(selectInput(ns("ident2"), "Identity 2 (Cell Types)", choices = sort(levels(DESeuratObj()@active.ident)), multiple = T), "Second Identity Class", "Specify the second identity class you want to compare", placement = "right", options = list(container = "body")),
          selectInput(ns("deMethod"), "DE analysis method", choices = c("wilcox" = "wilcox", "bimod" = "bimod", "roc" = "roc", 
                                                                        "LR" = "Logistic Regression", "t" = "Student's t", 
                                                                        "negbinom" = "Negative Binomial", 
                                                                        "poisson" = "Poisson", "MAST" = "MAST", "DESeq2" = "DESeq2")),
          bsPopover(ns("deMethod"), "DE method", "Test to use for performing DE Analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("logfcThreshold"), "Log Fold Change threshold", value = 0.1),
          bsPopover(ns("logfcThreshold"), "Log FC threshold", "Minimum fold difference (log-scale) required between two groups of cells to be included in the analysis", placement = "right", trigger = "hover", options = list(container = "body")),
          
          numericInput(ns("minPct"), "Minimum percentage", value = 0.01),
          bsPopover(ns("minPct"), "Minimum percentage of cells", "Minimum fraction of cells in either of the groups where gene must be detected to be considered for testing", placement = "right", trigger = "hover", options = list(container = "body")),
          
          checkboxInput(ns("onlyPos"), "Only positive markers", value = FALSE),
          bsPopover(ns("onlyPos"), "Only positive markers", "Genes with only positive log FC will be considered", placement = "right", trigger = "hover", options = list(container = "body")),
          
          selectInput(ns("groupby"), "Group by:", choices = metadataChoices, selected = ""),
          bsPopover(ns("groupby"), "Group by identity classes", "Allows to specify a metadata column to group cells prior to DE Analysis.", placement = "right", trigger = "hover", options = list(container = "body")),
          actionButton(ns("performDEAnalysisNow"), "Perform DE analysis", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
          br(), br(),
          uiOutput(ns("downloadDEResultsUI"))
        )
      }
    })
    
    # Handle DE analysis process
    shiny::observeEvent(input$performDEAnalysisNow, {
      
      if (input$performDEAnalysisNow > 0) {
        if (input$deAnalysisMethod == "FindMarkers" && (is.null(input$ident1) || is.null(input$ident2))) {
          shinyalert("Error", "Please select identities to perform differential gene expression analysis.", type = "error", confirmButtonCol = "#1c82e0")
          return()
        }
      }
      
      # Check if there are enough clusters for DE Analysis
      unique_clusters <- length(unique(Idents(DESeuratObj())))
      if (unique_clusters < 2) {
        shinyalert("Error", "Not enough clusters or groups to perform differential expression analysis.", type = "error", confirmButtonCol = "#1c82e0")
        return()
      }
      
      req(DESeuratObj())
      
      # Start timing
      timeStart <- Sys.time()
      
      # Initiate the progress bar
      withProgress(message = 'Performing Differential Expression Analysis', value = 0, {
        tryCatch({
          if (input$deAnalysisMethod == "FindMarkers") {
            shiny::setProgress(message = "Analyzing markers...", value = 0.5)
            
            groupByParam <- if (input$groupby == "") NULL else input$groupby
            
            # Handle multiple selections
            ident1 <- as.character(unlist(input$ident1, ","))
            ident2 <- as.character(unlist(input$ident2, ","))
            
            results <- Seurat::FindMarkers(
              object = DESeuratObj(),
              ident.1 = ident1,
              ident.2 = ident2,
              logfc.threshold = input$logfcThreshold,
              min.pct = input$minPct,
              test.use = input$deMethod,
              only.pos = input$onlyPos,
              group.by = groupByParam
            )
            results$compared_idents <- paste(paste(ident1, collapse = ","), if (is.null(ident2)) "All" else paste(ident2, collapse = ","), sep = " vs. ")
          } else if (input$deAnalysisMethod == "FindAllMarkers") {
            shiny::setProgress(message = "Analyzing all markers...", value = 0.5)
            results <- Seurat::FindAllMarkers(
              object = DESeuratObj(),
              logfc.threshold = input$logfcThreshold,
              min.pct = input$minPct,
              test.use = input$deMethod
            )
          }
          shiny::setProgress(message = "Finalizing analysis...", value = 1)
          deResults(results) # Update reactive value with DE results
          LastExecutedMethod(input$deAnalysisMethod) # Update last executed method
          shinyalert("DE completed", "Lets get started with Pathway Analysis", type = "success", confirmButtonCol = "#1c82e0")
          
        }, error = function(e) {
          # Custom error message for duplicates or other errors
          if (!is.null(ident2) && any(ident1 %in% ident2)){
            shinyalert("Error", "No duplicate cells allowed.", type = "error", confirmButtonCol = "#1c82e0")
          } else{
            shinyalert(paste("Error during DE analysis:", e$message), type = "error", confirmButtonCol = "#1c82e0")
          }
        })
      })
    })
    
    # Download DE Results
    output$downloadDEResults <- downloadHandler(
      filename = function() { paste("CogentDS_v2.0_DE_results-", Sys.Date(), ".csv", sep = "") },
      content = function(file) {
        if (!is.null(deResults()) && nrow(deResults()) > 0) {
          write.csv(deResults(), file, row.names = TRUE)
        }
      }
    )
    
    # Render plots based on analysis method and user input
    output$PlotSettings <- renderUI({
      req(deResults(), !is.null(LastExecutedMethod()), LastExecutedMethod() == input$deAnalysisMethod, input$performDEAnalysisNow > 0)
      if (input$deAnalysisMethod %in% c("FindMarkers")) {
        # Render volcano plot settings
        output$volcanoSettings <- renderUI({
          tagList(
            h4("Volcano Plot Parameters"),
            numericInput(ns("adjPValThreshold"), "Adjusted p-value threshold", value = 0.05),
            bsPopover(ns("adjPValThreshold"), "Adjusted p-value threshold", "Select genes based on adjusted P value threshold to be visualized on Volcano Plot ", placement = "right", trigger = "hover", options = list(container = "body")),
            
            numericInput(ns("logFCThreshold_volcano"), "Log Fold Change threshold", value = 1),
            bsPopover(ns("logFCThreshold_volcano"), "Log Fold Change threshold", "Select genes based on Log Fold change threshold to be visualized on Volcano Plot", placement = "right", trigger = "hover", options = list(container = "body")),
            
            numericInput(ns("pointSize"), "Point Size", value = 1),
            bsPopover(ns("pointSize"), "Point Size", "Set the point size for the volcano plot", placement = "right", trigger = "hover", options = list(container = "body")),
            
            textInput(ns("GenesofInterest"), "Genes of interest (comma-separated)", value = ""),
            bsPopover(ns("GenesofInterest"), "Genes of interest (comma-separated)", "Enter genes of interest (comma separated) to be displayed on volcano plot", placement = "right", trigger = "hover", options = list(container = "body")),
            
            fileInput(ns("FileUpload"), "Genes of interest (comma-separated)", accept = c(".csv")),
            bsPopover(ns("FileUpload"), "Upload file", "Upload file containing comma separated genes of interest to be displayed on volcano Plot", placement = "right", trigger = "hover", options = list(container = "body")),
            actionButton(ns("GeneratePlots"), "Generate plots", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
            
          )
        })
      } else if (input$deAnalysisMethod == "FindAllMarkers") {
        # Render heatmap settings
        output$heatmapSettings <- renderUI({
          
          tagList(
            h4("Heatmap Parameters"),
            
            numericInput(ns("nHeatmap"), "Number of features for heatmap", value = 5),
            bsPopover(ns("nHeatmap"), "Number of features for heatmap", "Select number of genes from each cluster to be visulised on Heatmap" , placement = "right", trigger = "hover", options = list(container = "body")),
            
            numericInput(ns("logFCThreshold_heatmap"), "Log Fold Change threshold", value = 1),
            bsPopover(ns("logFCThreshold_heatmap"), "Log Fold Change threshold", "Select genes based on Log Fold change threshold to be visualized on Heatmap", placement = "right", trigger = "hover", options = list(container = "body")),
            actionButton(ns("GeneratePlots"), "Generate plots", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
            
          )
        })
      }
    })
    
    # Render dynamic plots based on analysis method and user input
    output$plottype <- renderUI({
      req(deResults(), LastExecutedMethod())
      
      if (LastExecutedMethod() == "FindAllMarkers") {
        tagList(
          if (!is.null(heatmapplt())) {
            selectInput(ns("HeatMapDE"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")
          },
        )
      } else if (LastExecutedMethod() == "FindMarkers") {
        tagList(
          if (!is.null(volcanoplt())) {
            selectInput(ns("VolcanoDE"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png")
          },
        )
      }
    })
    
    # Render dynamic plots based on analysis method and user input
    output$dynamicPlots <- renderUI({
      req(deResults(), LastExecutedMethod())
      
      if (LastExecutedMethod() == "FindAllMarkers") {
        tagList(
          if (!is.null(heatmapplt())) {
            downloadButton(ns("download_heatmap"), "Save plot", style='color:#fff; background-color:#1c82e0')
          },
          plotOutput(ns("HeatMapPlot"), width = 600, height = 600)
        )
      } else if (LastExecutedMethod() == "FindMarkers") {
        tagList(
          if (!is.null(volcanoplt())) {
            downloadButton(ns("download_volcano"), "Save plot", style='color:#fff; background-color:#1c82e0')
          },
          plotlyOutput(ns("volcanoPlot"), width = 600, height = 600)
        )
      }
    })
    
    # Reactive values to store plots
    volcanoplt <- reactiveVal(NULL)
    heatmapplt <- reactiveVal(NULL)
    
    # Generate plots
    observeEvent(input$GeneratePlots, {
      
      req(deResults(), LastExecutedMethod())
      results <- deResults()
      results$gene <- rownames(results)
      
      if (LastExecutedMethod() == "FindAllMarkers") {
        # Heatmap plotting
        heatmapplt({
          heatmap_de_results <- deResults()
          top <- heatmap_de_results %>%
            group_by(cluster) %>%
            filter(avg_log2FC > input$logFCThreshold_heatmap) %>%
            slice_head(n = input$nHeatmap) %>%
            ungroup()
          
          # Check for intersection between top genes and rownames of Scale.data
          if (length(intersect(top$gene, rownames(DESeuratObj()@assays$RNA$scale.data))) == 0) {
            shinyalert("Warning", "No features found in Scale Data. Consider running normalization module with all features.", type = "warning", confirmButtonCol = "#1c82e0")
            return(NULL)
          }
          
          DoHeatmap(DESeuratObj(), features = top$gene)
        })
        
        # Render plots
        output$HeatMapPlot <- renderPlot({
          heatmapplt()
        })
        
        output$download_heatmap <- downloadHandler(
          filename = function() {
            paste0("CogentDS_v2.0_TopMarkers_Heatmap", "-", Sys.Date(), input$HeatMapDE)
          },
          content = function(file) {
            heatmapplt()
            # top <- results %>% group_by(cluster) %>% filter(avg_log2FC > input$logFCThreshold) %>% slice_head(n = input$nHeatmap) %>% ungroup()
            # DoHeatmap(seuratObject(), features = top$gene) + RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
            ggsave(file, plot = heatmapplt(), width = 10, height = 10)
          },
          contentType = 'image'
        ) 
        
      } else if (LastExecutedMethod() == "FindMarkers") {
        
        # Initialize an empty vector for genes of interest
        interestedGenes <- character(0)
        
        # Determine the source of gene names based on user input
        if (!is.null(input$FileUpload) && file.exists(input$FileUpload$datapath)) {
          # Read genes from uploaded file
          uploadedGenes <- read.csv(input$FileUpload$datapath, header = FALSE, stringsAsFactors = FALSE)
          interestedGenes <- unlist(uploadedGenes)
        } else if (nzchar(input$GenesofInterest)) {
          # Use genes entered in the text input
          interestedGenes <- trimws(unlist(strsplit(input$GenesofInterest, ",")))
        }
        
        # Filter results if there are any genes of interest
        if (length(interestedGenes) > 0) {
          filteredResults <- results[results$gene %in% interestedGenes, ]
          if (nrow(filteredResults) == 0) {
            showNotification("No genes found. Did you enter the gene names correctly?", type = "error")
            return()
          }
          results <- filteredResults
        }
        
        # Volcano plot generation
        interactive_vlplot <-  reactive({
          results$gene <- rownames(results)
          Title <- paste("Volcano Plot for", unique(results$compared_idents))
          
          results$category <- ifelse(results$p_val_adj < input$adjPValThreshold & results$avg_log2FC > input$logFCThreshold_volcano, "Upregulated",
                                     ifelse(results$p_val_adj < input$adjPValThreshold & results$avg_log2FC < -input$logFCThreshold_volcano, "Downregulated", "Non-significant"))
          results$hover_info <- paste("Gene:", results$gene, "<br>Log2 FC:", round(results$avg_log2FC, 3), "<br>Adj P-value:", results$p_val_adj)
          p <- ggplot(results, aes(x = avg_log2FC, y = -log10(p_val_adj), color = category, text = hover_info)) +
            geom_point(alpha = 0.5, size = input$pointSize) +
            scale_color_manual(values = c("Upregulated" = "red", "Downregulated" = "blue", "Non-significant" = "grey")) +
            labs(x = "Log2 Fold Change", y = "-Log10 Adjusted P-value", title = Title) + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
          ggplotly(p, tooltip = "text")
        })
        
        # Update Volcano Plot
        volcanoplt(interactive_vlplot())
        
        # Render plots
        output$volcanoPlot <- renderPlotly({
          volcanoplt()
        })
        
        # Save static volcano plot with labels
        static_volcanoplt = reactive({
          results$gene <- rownames(results)
          Title <- paste("Volcano Plot for", unique(results$compared_idents))
          
          results$category <- ifelse(results$p_val_adj < input$adjPValThreshold & results$avg_log2FC > input$logFCThreshold_volcano, "Upregulated",
                                     ifelse(results$p_val_adj < input$adjPValThreshold & results$avg_log2FC < -input$logFCThreshold_volcano, "Downregulated", "Non-significant"))
          
          # Identify top genes for labeling
          top_genes <- results %>%
            filter(p_val_adj < input$adjPValThreshold & abs(avg_log2FC) > input$logFCThreshold_volcano) %>%
            arrange(p_val_adj) %>%
            head(n = 20) # Adjust the number of top genes to label as needed
          
          ggplot(results, aes(x = avg_log2FC, y = -log10(p_val_adj), color = category)) +
            geom_point(alpha = 0.5, size = input$pointSize) +
            scale_color_manual(values = c("Upregulated" = "red", "Downregulated" = "blue", "Non-significant" = "grey")) +
            labs(x = "Log2 Fold Change", y = "-Log10 Adjusted P-value", title = Title) +
            guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) +
            theme(axis.text = element_text(colour = "black", size = 15, face = "bold"), 
                  axis.title = element_text(colour = "black", size = 15, face = "bold"), 
                  legend.position = "right", legend.box = "vertical", 
                  panel.background = element_rect(colour = "white")) +
            geom_text_repel(data = top_genes, aes(label = gene), size = 3, max.overlaps = 10)
        })
        
        output$download_volcano <- downloadHandler(
          filename = function() {
            paste0("CogentDS_v2.0_Markers_VolcanoPlot", "-", Sys.Date(), input$VolcanoDE)
          },
          content = function(file) {
            static_volcanoplt()
            ggsave(file, plot = static_volcanoplt(), width = 10, height = 10)
          },
          contentType = 'image'
        ) 
      }
    })
    
    # Return deResults
    return(deResults)
    
  })
}
