#' UI for Differential Expression Analysis Module
#'
#' Generates UI elements for a differential expression (DE) analysis module within
#' a Shiny application. This module allows users to select a DE analysis method,
#' set parameters for the analysis, and initiate the analysis process. It supports
#' various methods for finding differential genes and provides options for downloading results,
#' adjusting visualization thresholds, and generating plots to visualize the DE analysis
#' results.
#'
#' @param id A unique identifier for the UI module, necessary for distinguishing
#'           this module's input from others in the same Shiny application.
#'
#'
#' @return A tag list containing the UI elements required for the DE analysis module,
#'         including selection inputs, action buttons, and outputs for visualization.
#'         
#' @export
#'
#' @examples
#' diffExpressionUI("differentialExpressionModule")
#' UI for Differential Expression Analysis Module
#'
#' Generates UI elements for a differential expression (DE) analysis module within
#' a Shiny application. This module allows users to select a DE analysis method,
#' set parameters for the analysis, and initiate the analysis process. It supports
#' various methods for finding differential genes and provides options for downloading results,
#' adjusting visualization thresholds, and generating plots to visualize the DE analysis
#' results.
#'
#' @param id A unique identifier for the UI module, necessary for distinguishing
#'           this module's input from others in the same Shiny application.
#'
#'
#' @return A tag list containing the UI elements required for the DE analysis module,
#'         including selection inputs, action buttons, and outputs for visualization.
#'         
#' @export
#'
#' @examples
#' diffExpressionUI("differentialExpressionModule")
diffExpressionUI <- function(id) {
  ns <- NS(id)
  tagList(
    h2("Differential Expression Analysis"),
    fluidRow(
      column(
        width = 4,
        radioButtons(ns("deAnalysisMethod"), "DE analysis method",
                     choices = list("Find Markers" = "FindMarkers",
                                    "Find All Markers" = "FindAllMarkers"),
                     selected = "FindMarkers"),
        bsPopover(ns("deAnalysisMethod"), "Choose method for performing differential gene expression analysis", "Options include Find Markers, Find All Markers, and Find Conserved Markers", placement = "right", trigger = "hover", options = list(container = "body")),
        
        #actionButton(ns("runDEAnalysis"), "Select DE Method", class = "btn-primary"),
        radioButtons(ns("identityType"), "Identity type",
                     choices = c("Clusters", "Cell Type", 
                                 "Custom Selection Clusters", 
                                 "Custom Selection Cell Types"),
                     selected = "Clusters"),
        bsPopover(ns("identityType"), "Identity Type", "Select either clusters, cell type or custom selection to use for the analysis", placement = "right", trigger = "hover", options = list(container = "body")),
        
        uiOutput(ns("deParams")),
        tags$hr(), 
        uiOutput(ns("PlotSettings")),
      ),
      
      column(width = 6,
        uiOutput(ns("plottype")),
        uiOutput(ns("dynamicPlots")) 
      )
    )
  )
  
}