#' Feature Selection Server Logic
#'
#' This server-side module for a Shiny application facilitates the identification
#' and visualization of highly variable features within a Seurat object. It allows
#' users to select a method for feature selection and, depending on the
#' method, specify the number of features to consider. The results are visualized
#' through a variable feature plot, highlighting the top variable features.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'           
#' @param seuratObject A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#'
#' @seealso \code{\link[Seurat]{FindVariableFeatures}}, \code{\link[Seurat]{VariableFeatures}}, and
#'          \code{\link[Seurat]{VariableFeaturePlot}} for understanding the Seurat functionalities utilized.
#'
#' @examples
#' 
#' featureSelectionServer("featureSelectionModule", seuratObject)
featureSelectionServer <- function(id, seuratObject) { 
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    # Reactive value to indicate completion of variable feature finding
    variableFeaturesFound <- reactiveVal(FALSE)
    
    # Modal dialog for variable feature method selection and parameters
    shiny::observeEvent(input$openVariableFeaturesModal, {
      variableFeaturesFound(FALSE) # Reset the flag when modal is opened
      showModal(modalDialog(
        title = "Find Highly Variable Features",
        # Method selection for finding variable features
        selectInput(ns("variableFeaturesMethod"), "Variable Features Method", choices = c("vst" = "vst", 
                                                                                          "mean.var.plot" = "mean.var.plot", 
                                                                                          "dispersion" = "dispersion"), 
                    selected = "vst"),
        
        # Conditional input for number of features based on method selection
        conditionalPanel(
          condition = "input.variableFeaturesMethod != 'dispersion'",
          numericInput(ns("nFeatures"), "Number of Features", value = 2000)
        ),
        # Button to initiate the variable feature finding process
        actionButton(ns("findVariableFeatures"), "Find Variable Features", class = "btn-primary")
      ))
    })
    
    # Finding variable features based on user inputs
    shiny::observeEvent(input$findVariableFeatures, {
      req(seuratObject())
      
      # Start timing
      timeStart <- Sys.time()
      
      # Initialize progress bar
      withProgress(
        message = 'Finding Variable Features...',
        value = 0, {
          tryCatch({
            method <- input$variableFeaturesMethod
            nFeatures <- if (method == "dispersion") NULL else input$nFeatures
            # Perform feature selection with Seurat
            updatedSeuratObject <- Seurat::FindVariableFeatures(seuratObject(), selection.method = method, nfeatures = nFeatures)
            seuratObject(updatedSeuratObject) # Update seuratObject
            
            variableFeaturesFound(TRUE) # Mark variable features as found
            
            # Increment progress bar value
            incProgress(1)
            
            showNotification(paste("Variable features identified"), type = "message")
          }, error = function(e) {
            showNotification(paste("Error finding variable features:", e$message), type = "error")
          })
        })
      
      removeModal()
    })
    
    # Render plot for variable features
    output$variableFeaturePlot <- renderPlot({
      req(seuratObject(), variableFeaturesFound()) # Ensure plot is rendered after features are found
      
      tryCatch({
        top10 <- head(Seurat::VariableFeatures(seuratObject()), 10)
        plot1 <- Seurat::VariableFeaturePlot(seuratObject())
        plot2 <- Seurat::LabelPoints(plot = plot1, points = top10, repel = TRUE)
        plot1 + plot2
      }, error = function(e) {
        showNotification(paste("Error rendering variable features plot:", e$message), type = "error")
      })
    })
  })
}
