#' Gene fusion analysis module server logic
#' 
#' This server module handles overlaying fusion genes on UMAP
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param seuratObj A reactive expression containing the Seurat object.
#'
#' @return
#' @export
#'
#' @examples
#' geneFusionServer("geneFusionModule", seuratObject)
geneFusionServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    
    ns <- session$ns
    shinyjs::hide("FusionPlt")
    
    output$downloadPlotUI <- renderUI({
      if (input$overlay_btn > 0 && !is.null(static_overlay_plot()) && !is.null(input$fusion_gene)) {
        downloadButton(ns('downloadPlot'), 'Save plot', style='color:#fff; background-color:#1c82e0')
      }
    })
    
    # Reactive value to hold metadata columns
    metadataColumns <- reactive({
      req(seuratObj())
      
      # Get the names of metadata columns
      all_metadata <- colnames(seuratObj()@meta.data)
      # Columns that are either factors or characters
      relevant_columns <- all_metadata[sapply(seuratObj()@meta.data[, all_metadata], function(col) is.factor(col) || is.character(col))]
      
      # Columns to be potentially removed
      column_not_needed <- c("Barcode")
      
      # Check and remove columns if they exist
      if (any(column_not_needed %in% relevant_columns)) {
        relevant_columns <- relevant_columns[!(relevant_columns %in% column_not_needed)]
      }
      
      # Return the filtered relevant columns
      relevant_columns
    })
    
    # Function to get fusion genes with additional information
    get_fusion_genes_with_info <- function() {
      req(seuratObj())
      
      if (!("span_frags" %in% names(seuratObj()@assays)) || !("junction_reads" %in% names(seuratObj()@assays))) {
        return(NULL)
      }
      
      spanning_reads <- seuratObj()@assays[["span_frags"]]@counts
      junction_reads <- seuratObj()@assays[["junction_reads"]]@counts
      
      if (is.null(spanning_reads) || is.null(junction_reads)) {
        return(NULL)
      }
      
      fusion_genes_info <- lapply(rownames(spanning_reads), function(gene) {
        cells_with_span <- which(spanning_reads[gene, ] > 0)
        cells_with_junction <- which(junction_reads[gene, ] > 0)
        
        # Count of cells with spanning reads and junction reads respectively
        span_count <- length(cells_with_span)
        junction_count <- length(cells_with_junction)
        
        list(gene = gene, span_count = span_count, junction_count = junction_count)
      })
      
      fusion_genes_info
    }
    
    fusion_genes_with_info <- reactive({
      get_fusion_genes_with_info()
    })
    
    output$ConditionalParamteres <- renderUI({
      if (is.null(fusion_genes_with_info()) || length(fusion_genes_with_info()) == 0) {
        tagList(
          HTML("<p>No fusion gene data available</p>")
        )
      } else {
        # Sorting choices by frequency in descending order
        sorted_choices <- fusion_genes_with_info()[order(sapply(fusion_genes_with_info(), function(gene_info) -gene_info$span_count))]
        
        column(4,
               selectInput(ns("fusion_gene"), label = "Select fusion genes:",
                           choices = lapply(sorted_choices, function(gene_info) {
                             paste(gene_info$gene, " (Span:", gene_info$span_count, ", Junction:", gene_info$junction_count, ")", sep = "")
                           }), multiple = TRUE),
               radioButtons(ns("overlay_type"), "Overlay type:",
                            choices = c("Span Fragments", "Junction Reads", "Span Fragments + Junction Reads"),
                            selected = "Span Fragments"), 
               selectInput(ns("umapColorBy"), "Color by:", choices = metadataColumns()),
               actionButton(ns("overlay_btn"), "Overlay Plot", style='color:#fff; background-color:#1c82e0')
        )
      }
    })
    
    
    # Reactive expression to subset umap coordinates with cells having the selected fusion genes
    umap_coords_with_fusion <- reactive({
      req(input$fusion_gene, seuratObj())
      umap_coords_with_fusion <- data.frame(matrix(ncol = 2, nrow = 0))
      selected_genes <- gsub("\\s*\\(.*", "", input$fusion_gene)
      
      for (gene in selected_genes) {
        spanning_reads <- seuratObj()@assays[["span_frags"]]@counts
        junction_reads <- seuratObj()@assays[["junction_reads"]]@counts
        
        cells_with_fusion <- switch(input$overlay_type,
                                    "Span Fragments" = which(spanning_reads[gene, ] > 0),
                                    "Junction Reads" = which(junction_reads[gene, ] > 0),
                                    "Span Fragments + Junction Reads" = which(spanning_reads[gene, ] > 0 | junction_reads[gene, ] > 0))
        
        if("umap" %in% Reductions(seuratObj())) {
          umap_coords <- data.frame(seuratObj()[["umap"]]@cell.embeddings)
          umap_coords_with_fusion_gene <- data.frame(umap_coords[rownames(umap_coords) %in% colnames(spanning_reads)[cells_with_fusion], ])
          
          umap_coords_with_fusion_gene$gene <- gene
          umap_coords_with_fusion_gene$gene <- as.factor(umap_coords_with_fusion_gene$gene)
          
          umap_coords_with_fusion <- rbind(umap_coords_with_fusion, umap_coords_with_fusion_gene)
        }
        if("tsne" %in% Reductions(seuratObj())) {
          umap_coords <- data.frame(seuratObj()[["tsne"]]@cell.embeddings)
          umap_coords_with_fusion_gene <- data.frame(umap_coords[rownames(umap_coords) %in% colnames(spanning_reads)[cells_with_fusion], ])
          
          umap_coords_with_fusion_gene$gene <- gene
          umap_coords_with_fusion_gene$gene <- as.factor(umap_coords_with_fusion_gene$gene)
          
          umap_coords_with_fusion <- rbind(umap_coords_with_fusion, umap_coords_with_fusion_gene)
        }
      }
      umap_coords_with_fusion
    })
    
    # Reactive expression for overlay plot
    overlay_plot <- eventReactive(input$overlay_btn, {
      
      if (input$overlay_btn > 0) {
        if (is.null(input$fusion_gene)) {
          shinyalert("Error", "Please select fusion genes to overlay.", type = "error", confirmButtonCol = "#1c82e0")
          return()
        }
      }
      req(umap_coords_with_fusion(), seuratObj(), input$fusion_gene, input$umapColorBy)
      
      # Check if either UMAP or tSNE reductions are present
      if (!"umap" %in% Reductions(seuratObj()) && !"tsne" %in% Reductions(seuratObj())) {
        shinyalert("Neither UMAP nor t-SNE reduction is available. Please run clustering Module.", type = "warning", confirmButtonCol = "#1c82e0")
        return()
      }
      
      # Create the UMAP plot
      p1 <- DimPlot(seuratObj(), pt.size = 1.5, label = FALSE, label.size = 5 ,group.by = input$umapColorBy) + 
        RotatedAxis() + 
        guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + 
        theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), 
              axis.title = element_text(colour = "black", size = 15, face = "bold"), 
              legend.position = "right", legend.box = "vertical", 
              panel.background = element_rect(colour = "white"))
      
      # Convert ggplot to plotly
      plot <- ggplotly(p1)
      
      selected_genes <- gsub("\\s*\\(.*", "", input$fusion_gene)
      for (gene in selected_genes) {
        if("umap" %in% Reductions(seuratObj())) {
          plot <- plot %>% add_trace(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            x = ~umap_1,
            y = ~umap_2,
            type = "scatter",
            mode = "markers",
            marker = list(
              size = 7,
              symbol = "circle-open",
              color = "black"
            ),
            hoverinfo = "text",
            text = ~gene,
            name = gene
          )
        }
        if("tsne" %in% Reductions(seuratObj())) {
          plot <- plot %>% add_trace(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            x = ~tSNE_1,
            y = ~tSNE_2,
            type = "scatter",
            mode = "markers",
            marker = list(
              size = 7,
              symbol = "circle-open",
              color = "black"
            ),
            hoverinfo = "text",
            text = ~gene,
            name = gene
          )
        }
      }
      plot
    })
    
    # Reactive expression to generate the static overlay plot
    static_overlay_plot <- reactive({
      req(umap_coords_with_fusion(), seuratObj(), input$fusion_gene, input$umapColorBy)
      
      p1 <- DimPlot(seuratObj(), pt.size = 1.5, label = FALSE, label.size = 5, group.by = input$umapColorBy) + 
        RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) + 
        theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"), axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right", legend.box = "vertical", panel.background = element_rect(colour = "white"))
      
      selected_genes <- gsub("\\s*\\(.*", "", input$fusion_gene)
      
      for (gene in selected_genes) {
        if("umap" %in% Reductions(seuratObj())) {
          p1 <- p1 + geom_point(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            aes(x = umap_1, y = umap_2, shape = gene),
            size = 2,
            fill = NA,
            color = "black") + scale_shape_manual(values = rep(21, length(input$fusion_gene))) +
            guides(color = guide_legend(order=1),
                   shape = guide_legend(order=2, title = "")) +
            theme(legend.spacing.y = unit(-0.2, "cm"))
          
        }
        if("tsne" %in% Reductions(seuratObj())) {
          p1 <- p1 + geom_point(
            data = umap_coords_with_fusion()[umap_coords_with_fusion()$gene == gene, ],
            aes(x = tSNE_1, y = tSNE_2, shape = gene),
            size = 2,
            fill = NA,
            color = "black") + scale_shape_manual(values = rep(21, length(input$fusion_gene))) +
            guides(shape = guide_legend(order=2, title = "")) +
            theme(legend.spacing.y = unit(-0.2, "cm"))
          
        }
      }
      p1
    })
    
    # Render fusion gene plot
    output$fusion_plot <- renderPlotly({
      overlay_plot()
    })
    
    observe({
      if (!is.null(overlay_plot())) {
        shinyjs::show("FusionPlt")
      }
    })
    
    output$downloadPlot <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_FusionOverlayPlot", "-", Sys.Date(), input$FusionPlt)
      },
      content = function(file) {
        ggsave(file, plot = static_overlay_plot(), width = 10, height = 10)
      },
      contentType = 'image'
    )    
    
  })
}
