#' Normalization and Scaling UI Component
#' 
#' Generates UI elements specifically for initiating normalization and scaling 
#' operations on dataset within a Shiny application. This function creates action 
#' buttons that, when activated, trigger server-side logic to perform these 
#' data preprocessing steps
#'
#' @param id A unique identifier for the UI module, necessary for distinguishing
#'           this module's input from others in the same Shiny application. 
#'
#' @return UI elements comprising two action buttons: one for starting the 
#'         normalization process and another for initiating data scaling. These buttons 
#'         are designed to trigger respective preprocessing steps which should be 
#'         defined in the server logic of the Shiny application.
#'
#' @export
#' @examples
#' normalizationScalingUI("normalizationScalingModule")
normalizationScalingUI <- function(id) {
  ns <- NS(id)
  tagList(
    fluidRow(column(4,
                    h2("Normalization, Feature Selection & Scaling"),
                    h5("*Required", style="color:#1c82e0"),
                    selectInput(ns("normMethod"), "Normalization method", choices = c("LogNormalize" = "LogNormalize", "SCT" = "SCT"), selected = "LogNormalize"),
                    bsPopover(ns("normMethod"), "Normalization method", "Method for normalization; LogNormalize: Feature counts for each cell are divided by the total counts for that cell and multiplied by the scale.factor. This is then natural-log transformed using log1p; SCT: scTransform. Please check Seurat documentation for more information", placement = "right", trigger = "hover", options = list(container = "body")),
                    conditionalPanel(
                      condition = "input.normMethod == 'LogNormalize' || input.normMethod == 'SCT'",
                      numericInput(ns("featureselection_number"), "Number of variable features", value = 2000),
                      bsPopover(ns("featureselection_number"), "Number of features to select", "Number of features to select as top variable features; only used when selection.method is set to VST", placement = "right", trigger = "hover", options = list(container = "body")), ns = NS(id)),
                    conditionalPanel(
                      condition = "input.normMethod == 'LogNormalize'",
                      numericInput(ns("scaleFactor"), "Scale factor", value = 10000, step = 100),
                      bsPopover(ns("scaleFactor"), "Scale Factor", "Sets the scale factor for cell-level normalization", placement = "right", trigger = "hover", options = list(container = "body")),
                      selectInput(ns("featureselection"), "Feature selection", choices = c("vst"), selected = "vst"),
                      bsPopover(ns("featureselection"), "Feature selection method", "VST: First, fits a line to the relationship of log(variance) and log(mean) using local polynomial regression (loess). Then standardizes the feature values using the observed mean and expected variance (given by the fitted line). Feature variance is then calculated on the standardized values after clipping to a maximum (see clip.max parameter)", placement = "right", trigger = "hover", options = list(container = "body")),
                      selectInput(ns("usefeatures"), "Features to use for scaling", choices = c("Variable features (Recommended for Large Datasets)", "All features"), selected = "Variable Features (Recommended for Large Datasets)"), ns = NS(id)),
                    actionButton(ns("performNormalization"), "Perform normalization, feature selection & scaling", class = "btn-primary", style='color:#fff; background-color:#1c82e0')),
             column(6,
                    conditionalPanel(
                      condition = "input.performNormalization > 0",
                      selectInput(ns("varfeature"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
                      downloadButton(ns('downloadPlot_varfeature'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
                      plotOutput(ns("Varfeatures"), width = "700", height = "600"), ns = NS(id))))
  )
}
