#' Pre-QC and QC Server Logic
#'  
#' Handles server-side logic for pre-quality control (Pre-QC) and quality control (QC)
#' operations. It includes functionalities for displaying
#' pre-QC plots, applying QC filters, and showing post-QC plots.
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'           
#' @param seuratObject A reactive expression containing the Seurat object to be processed.
#'
#' @importFrom shiny moduleServer observeEvent plotOutput
#' @importFrom Seurat VlnPlot FeatureScatter
#' @import shiny
#' @import Seurat
#' 
#' @seealso Refer to \code{\link[Seurat]{VlnPlot}}, \code{\link[Seurat]{FeatureScatter}}
#'          for more details on these functions.
#'
#' @return
#' @export
#' @examples
#' preQCQCServer("preQCQCModule", seuratObject)
preQCQCServer <- function(id, seuratObject) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    # Create a copy of the Seurat object for pre-QC analysis
    preQCSeuratObject <- reactiveVal(NULL)
    
    # Create a copy of the Seurat object for post-QC analysis
    postQCSeuratObject <- reactiveVal(NULL)
    
    # Reactive values for storing pre-QC and post-QC scatter plots
    p1 <- reactiveValues(scatterPlot = NULL)
    p2 <- reactiveValues(scatterPlot = NULL)
    
    observe({
      # Initialize the pre-QC Seurat object
      req(seuratObject())
      if("Original_RNA" %in% names(seuratObject())) {
        #message("Ambient RNA Correction is performed")
        sc_temp = seuratObject()
        #print(sc_temp@assays$user_parameter_tracker)
        DefaultAssay(sc_temp) = "RNA"
        Idents(sc_temp) <- "orig.ident"
        preQCSeuratObject(sc_temp)
      } else {
        #message("Ambient RNA Correction is skipped")
        sc_temp = seuratObject()
        #print(sc_temp@assays$user_parameter_tracker)
        DefaultAssay(sc_temp) = "RNA"
        Idents(sc_temp) <- "orig.ident"
        preQCSeuratObject(sc_temp)
      }
    })
    
    
    # Render plot selection input
    output$plotTypeInput <- renderUI({
      tagList(
        selectInput(ns("plotType"), "Select plot type", c("Violin plot", "Scatter plot"), selected = "Violin plot"),
      )
    })
    
    # Render pre-QC plots
    output$preQCPlots <- renderUI({
      tagList(
        h3("Pre-QC Plots"),
        selectInput(ns("preQC"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
        downloadButton(ns('downloadPlot_preQC'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
        uiOutput(ns("preQCPlot"))
      )
    })
    
    output$preQCPlot <- renderUI({
      if (input$plotType == "Violin plot") {
        tagList(
          shiny::plotOutput(ns("violinPlotPreQC"), height = 1000),
        )
      } else {
        shiny::plotOutput(ns("scatterPlotPreQC"))
      }
    })
    
    # Render pre-QC violin plots
    output$violinPlotPreQC <- renderPlot({
      req(preQCSeuratObject())
      
      tryCatch({
        # Generate violin plots for key metrics 
        preQCobj <- preQCSeuratObject()
        Idents(preQCobj) <- "orig.ident"  
        Seurat::VlnPlot(preQCobj, 
                        features = c("nFeature_RNA", "nCount_RNA", "percent.mt", "Ribosomal_reads_percentage", "Intergenic_reads_percentage"), 
                        same.y.lims = FALSE, stack = T, flip = T) + NoLegend() + scale_y_continuous(n.breaks = 10) +
          theme(axis.text.y = element_text(size = 10),
                plot.margin = unit(c(1, 1, 1, 2), "cm"))
      }, error = function(e) {
        showNotification(paste("Error generating pre-QC violin plot:", e$message), type = "error")
      })
    })
    
    # Render pre-QC scatter plot
    output$scatterPlotPreQC <- renderPlot({
      req(preQCSeuratObject())
      tryCatch({
        preQCobj <- preQCSeuratObject()
        Idents(preQCobj) <- "orig.ident"
        
        # Check the number of levels in Idents
        num_levels <- length(levels(Idents(preQCobj)))
        
        if (num_levels > 1) {
          # Generate scatter plots
          p1$scatterPlot <- Seurat::FeatureScatter(preQCobj, feature1 = "nCount_RNA", feature2 = "percent.mt") + ggtitle(paste("Pearson correlation", round(cor(preQCobj$nCount_RNA, preQCobj$percent.mt),2), sep = ":")) + guides(color = guide_legend(nrow = 6)) + theme(legend.position = "bottom") + 
            Seurat::FeatureScatter(preQCobj, feature1 = "nCount_RNA", feature2 = "nFeature_RNA") + NoLegend()  +  ggtitle(paste("Pearson correlation", round(cor(preQCobj$nCount_RNA, preQCobj$nFeature_RNA),2), sep = ":")) 
          print(p1$scatterPlot)
        } else {
          # Generate scatter plots
          p1$scatterPlot <- Seurat::FeatureScatter(preQCobj, feature1 = "nCount_RNA", feature2 = "percent.mt") +  NoLegend() + ggtitle(paste("Pearson correlation", round(cor(preQCobj$nCount_RNA, preQCobj$percent.mt),2), sep = ":")) +
            Seurat::FeatureScatter(preQCobj, feature1 = "nCount_RNA", feature2 = "nFeature_RNA")  + ggtitle(paste("Pearson correlation", round(cor(preQCobj$nCount_RNA, preQCobj$nFeature_RNA),2), sep = ":")) 
          print(p1$scatterPlot)
        }
        
        
      }, error = function(e) {
        showNotification(paste("Error generating pre-QC scatter plots:", e$message), type = "error")
      })
    })
    
    # Save pre-QC plot
    output$downloadPlot_preQC <- downloadHandler(
      filename = function() {
        paste0("CogentDS_v2.0_PreQCPlots_", Sys.Date(), input$preQC)
      },
      content = function(file) {
        if (input$plotType == "Violin plot") {
          req(preQCSeuratObject())
          preQCobj <- preQCSeuratObject()
          Idents(preQCobj) <- "orig.ident"
          plot_to_save <-  Seurat::VlnPlot(preQCobj, 
                                           features = c("nFeature_RNA", "nCount_RNA", "percent.mt", "Ribosomal_reads_percentage", "Intergenic_reads_percentage"), 
                                           same.y.lims = FALSE, stack = T, flip = T) + NoLegend() + scale_y_continuous(n.breaks = 10) +
            theme(axis.text.y = element_text(size = 10),
                  plot.margin = unit(c(1, 1, 1, 2), "cm"))
        } else if (input$plotType == "Scatter plot") {
          # Save the scatter plot that is rendered in the UI
          plot_to_save <- p1$scatterPlot
        }
        
        ggsave(file, plot = plot_to_save, width = 10, height = 10)
      },
      contentType = 'image'
    )
    
    # QC Filters
    output$qcfiltersUI <- renderUI({
      tagList(h3("QC Parameters"),
              fluidRow(
                column(4,
                       numericInput(ns("nFeatureLow"), "Minimum number of genes/features per cell", min = 0, max = round(max(preQCSeuratObject()$nFeature_RNA)), value = 0),
                       numericInput(ns("nFeatureHigh"), "Maximum number of genes/features per cell", min = 200, max = round(max(preQCSeuratObject()$nFeature_RNA)), value = round(quantile(preQCSeuratObject()@meta.data$nFeature_RNA, 0.95, na.rm = TRUE))),
                       checkboxInput(ns("skipnFeatureHigh"), "Skip maximum genes/features per cell threshold", value = FALSE),
                       #numericInput(ns("exonicfrac"), "Maximum percentage of Mapped Ambiguous Exonic Reads", min = 0, max = 100, value = round((outliers_mad(preQCSeuratObject()$ExonicReadFraction)[['UL_CI_MAD']]+outliers_mad(preQCSeuratObject()$ExonicReadFraction)[['MAD']])*100)),
                       #numericInput(ns("intronicfrac"), "Maximum percentage of Mapped Ambiguous Intronic Reads", min = 0, max = 100, value = round((outliers_mad(preQCSeuratObject()$IntronicReadFraction)[['UL_CI_MAD']]+outliers_mad(preQCSeuratObject()$IntronicReadFraction)[['MAD']])*100)),
                       numericInput(ns("intergenicfrac"), "Maximum Intergenic reads percentage", min = 0, max = 100, value = round((outliers_mad(preQCSeuratObject()$Intergenic_reads_percentage)[['UL_CI_MAD']]+outliers_mad(preQCSeuratObject()$Intergenic_reads_percentage)[['MAD']]))),
                ),
                column(4,
                       numericInput(ns("ribofrac"), "Maximum Ribosomal reads percentage", min = 0, max = 100, value = round((outliers_mad(preQCSeuratObject()$Ribosomal_reads_percentage)[['UL_CI_MAD']]+outliers_mad(preQCSeuratObject()$Ribosomal_reads_percentage)[['MAD']]))),
                       numericInput(ns("percentMt"), "Maximum percentage of mitochondrial genes", min = 0, max = round(max(preQCSeuratObject()$percent.mt, na.rm = TRUE)), value = round(quantile(preQCSeuratObject()$percent.mt, 0.90, na.rm = TRUE))),
                       actionButton(ns("applyQC"), "Apply QC filters", class = "btn-primary", style='color:#fff; background-color:#1c82e0'),
                       actionButton(ns("skipQC"), "Skip QC", class = "btn-primary", style='color:#fff; background-color:#1c82e0'))
              )
      )
    })
    
    
    # Apply QC Filters and Show Post-QC Plots
    observeEvent(input$applyQC, {
      req(preQCSeuratObject())
      
      if("DecontX_contamination" %in% colnames(preQCSeuratObject()@meta.data)) {
        parameterList_use_sc = preQCSeuratObject()
        DefaultAssay(parameterList_use_sc) = "RNA"
        min_genes = paste0('Minimum number of genes/features per cell: ', input$nFeatureLow)
        max_genes = paste0('Maximum number of genes/features per cell: ', input$nFeatureHigh)
        ribofrac_cells = paste0('Ribosomal Reads percentage: ', input$ribofrac)
        percentMt_cells = paste0('Mitochondrial fraction: ', input$percentMt)
        intergenic_cells = paste0('Intergenic Reads percentage: ', input$intergenicfrac)
        qc_status = paste0('QC analysis: ', 'Performed')
        parameterList_use = parameterList_use_sc@assays$user_parameter_tracker
        parameterList_use = c(parameterList_use, min_genes, max_genes, percentMt_cells, ribofrac_cells, intergenic_cells, qc_status)
        parameterList_use_sc@assays$user_parameter_tracker = NULL
        DefaultAssay(parameterList_use_sc) = "RNA"
      } else {
        parameterList_use_sc = preQCSeuratObject()
        DefaultAssay(parameterList_use_sc) = "RNA"
        ambient_status = paste0('Ambient RNA correction: ', "Skipped")
        min_genes = paste0('Minimum number of genes/features per cell: ', input$nFeatureLow)
        max_genes = paste0('Maximum number of genes/features per cell: ', input$nFeatureHigh)
        ribofrac_cells = paste0('Ribosomal Reads percentage: ', input$ribofrac)
        percentMt_cells = paste0('Mitochondrial fraction: ', input$percentMt)
        intergenic_cells = paste0('Intergenic Reads percentage: ', input$intergenicfrac)
        qc_status = paste0('QC analysis: ', 'Performed')
        parameterList_use = parameterList_use_sc@assays$user_parameter_tracker
        parameterList_use = c(parameterList_use, ambient_status, min_genes, max_genes, percentMt_cells, ribofrac_cells, intergenic_cells, qc_status)
        parameterList_use_sc@assays$user_parameter_tracker = NULL
        DefaultAssay(parameterList_use_sc) = "RNA"
      }
      
      tryCatch({
        if (input$skipnFeatureHigh) {
          # Applying user-defined QC filters to the Seurat object
          updatedSeuratObject <- parameterList_use_sc %>%
            subset(subset = nFeature_RNA >= input$nFeatureLow & percent.mt <= input$percentMt & Ribosomal_reads_percentage <= input$ribofrac & Intergenic_reads_percentage <= input$intergenicfrac)
          
        } else {
          updatedSeuratObject <- parameterList_use_sc %>%
            subset(subset = nFeature_RNA >= input$nFeatureLow & nFeature_RNA <= input$nFeatureHigh & percent.mt <= input$percentMt & Ribosomal_reads_percentage <= input$ribofrac & Intergenic_reads_percentage <= input$intergenicfrac)
        }
        updatedSeuratObject@assays$user_parameter_tracker = parameterList_use
        DefaultAssay(updatedSeuratObject) = "RNA"
        postQCSeuratObject(updatedSeuratObject)
        
        shinyalert("QC is complete", paste("Number of cell:", ncol(updatedSeuratObject), "\n" ,"Number of genes:", nrow(updatedSeuratObject), "\n" ,"Let's Get Started with Normalization, Feature Selection & Scaling"), type = "success", confirmButtonCol = "#1c82e0")
        
        # QC stats table
        output$postQCStatsUI <- renderUI({
          tagList(
            fluidRow(
              h3("Post-QC Stats"),
              DT::dataTableOutput(ns("postQCStats")), style = "height:500px; overflow-y: scroll;overflow-x: scroll;")
          )
        })
        
        # Table for QC stats
        output$postQCStats <- DT::renderDataTable({
          req(postQCSeuratObject())
          
          # Calculate statistics
          num_cells <- ncol(postQCSeuratObject())
          num_genes <- nrow(postQCSeuratObject())
          genes_per_cell <- postQCSeuratObject()$nFeature_RNA
          quantile_range <- round(quantile(genes_per_cell, probs = c(0.025, 0.975)))
          
          # Compute IQR x 1.5
          IQR_value <- IQR(genes_per_cell)
          Q1 <- quantile(genes_per_cell, probs = 0.25)
          Q3 <- quantile(genes_per_cell, probs = 0.75)
          gene_lower_bound <- round(Q1 - 1.5 * IQR_value)
          gene_upper_bound <- round(Q3 + 1.5 * IQR_value)
          
          # Create a data frame with the statistics
          stats_df <- data.frame(
            Stats = c("Number of cells", "Number of features", 
                      "95% percentile range for number of features per cell",
                      "IQR * 1.5 range for number of features per cell"),
            Value = c(num_cells, num_genes, 
                      paste(quantile_range[1], quantile_range[2], sep = "-"),
                      paste(gene_lower_bound, gene_upper_bound, sep = "-"))
          )
          
          # Return the data frame for rendering as a table
          DT::datatable(stats_df, options = list(pageLength = 20, dom = "t"))
        })
        
        
        # Render post-QC plots
        output$postQCPlots <- renderUI({
          tagList(
            h3("Post-QC Plots"),
            selectInput(ns("postQC"), "File Type Selection:", choices = list("PNG" = ".png", "PDF" = ".pdf", "SVG" = ".svg", "JPEG" = ".jpeg"), selected = ".png"),
            downloadButton(ns('downloadPlot_postQC'), 'Save plot', style='color:#fff; background-color:#1c82e0'),
            uiOutput(ns("postQCPlot")),
            uiOutput(ns("postQCStatsUI"))
          )
        })
        
        output$postQCPlot <- renderUI({
          if (input$plotType == "Violin plot") {
            tagList(
              shiny::plotOutput(ns("violinPlotPostQC"), height = 1000),
            )
          } else {
            shiny::plotOutput(ns("scatterPlotPostQC"))
          }
        })
        
        # Render post-QC violin plots
        output$violinPlotPostQC <- renderPlot({
          tryCatch({
            # Generate violin plots for key metrics after QC
            postQCobj <- postQCSeuratObject()
            Idents(postQCobj) <- "orig.ident"
            Seurat::VlnPlot(postQCobj, 
                            features = c("nFeature_RNA", "nCount_RNA", "percent.mt", "Ribosomal_reads_percentage", "Intergenic_reads_percentage"), 
                            same.y.lims = FALSE, stack = T, flip = T) + NoLegend() + scale_y_continuous(n.breaks = 10) +
              theme(axis.text.y = element_text(size = 10),
                    plot.margin = unit(c(1, 1, 1, 2), "cm"))
          }, error = function(e) {
            showNotification(paste("Error generating post-QC violin plot:", e$message), type = "error")
          })
        })
        
        # Render post-QC scatter plot
        output$scatterPlotPostQC <- renderPlot({
          tryCatch({
            postQCobj <- postQCSeuratObject()
            Idents(postQCobj) <- "orig.ident"  
            
            # Check the number of levels in Idents
            num_levels <- length(levels(Idents(postQCobj)))
            
            if (num_levels > 1) {
              # Generate scatter plots 
              p2$scatterPlot <- Seurat::FeatureScatter(postQCobj, feature1 = "nCount_RNA", feature2 = "percent.mt") + ggtitle(paste("Pearson correlation", round(cor(postQCobj$nCount_RNA, postQCobj$percent.mt),2), sep = ":")) + guides(color = guide_legend(nrow = 6)) + theme(legend.position = "bottom") +
                Seurat::FeatureScatter(postQCobj, feature1 = "nCount_RNA", feature2 = "nFeature_RNA") + NoLegend() + ggtitle(paste("Pearson correlation", round(cor(postQCobj$nCount_RNA, postQCobj$nFeature_RNA),2), sep = ":"))
              print(p2$scatterPlot)    
            } else {
              # Generate scatter plots
              p2$scatterPlot <- Seurat::FeatureScatter(postQCobj, feature1 = "nCount_RNA", feature2 = "percent.mt") + NoLegend() + ggtitle(paste("Pearson correlation", round(cor(postQCobj$nCount_RNA, postQCobj$percent.mt),2), sep = ":")) +
                Seurat::FeatureScatter(postQCobj, feature1 = "nCount_RNA", feature2 = "nFeature_RNA") + ggtitle(paste("Pearson correlation", round(cor(postQCobj$nCount_RNA, postQCobj$nFeature_RNA),2), sep = ":"))
              print(p2$scatterPlot)    
            }
          }, error = function(e) {
            showNotification(paste("Error generating post-QC scatter plots:", e$message), type = "error")
          })
        })
        
        
        # Save post-QC plot
        output$downloadPlot_postQC <- downloadHandler(
          filename = function() {
            paste0("CogentDS_v2.0_PostQCPlots_", Sys.Date(), input$postQC)
          },
          content = function(file) {
            req(preQCSeuratObject())
            preQCobj <- preQCSeuratObject()
            Idents(preQCobj) <- "orig.ident"
            if (input$plotType == "Violin plot") {
              postQCobj <- postQCSeuratObject()
              Idents(postQCobj) <- "orig.ident"
              plot_to_save <-  VlnPlot(postQCobj, 
                                       features = c("nFeature_RNA", "nCount_RNA", "percent.mt", "Ribosomal_reads_percentage", "Intergenic_reads_percentage"), 
                                       same.y.lims = FALSE, stack = T, flip = T) + NoLegend() + scale_y_continuous(n.breaks = 10) +
                theme(axis.text.y = element_text(size = 10),
                      plot.margin = unit(c(1, 1, 1, 2), "cm"))
              
            } else if (input$plotType == "Scatter plot") {
              plot_to_save <- p2$scatterPlot
            }
            
            ggsave(file, plot = plot_to_save, width = 10, height = 10)
          },
          contentType = 'image'
        )
        
      }, error = function(e) {
        shinyalert("QC Failed", "Please check your input data", type = "error", confirmButtonCol = "#1c82e0")
      })
    })
    
    observeEvent(input$skipQC, {
      req(seuratObject())
      
      if("DecontX_contamination" %in% colnames(seuratObject()@meta.data)) {
        parameterList_use_sc = seuratObject()
        parameterList_use = parameterList_use_sc@assays$user_parameter_tracker
        qc_skipped = paste0('QC analysis : ', 'Skipped')
        parameterList_use = c(parameterList_use, qc_skipped)
        parameterList_use_sc@assays$user_parameter_tracker = parameterList_use
        DefaultAssay(parameterList_use_sc) = "RNA"
        Idents(parameterList_use_sc) <- "orig.ident"
        postQCSeuratObject(parameterList_use_sc)
      } else {
        parameterList_use_sc = seuratObject()
        parameterList_use = parameterList_use_sc@assays$user_parameter_tracker
        ambient_status = paste0('Ambient RNA correction: ', 'Skipped')
        qc_skipped = paste0('QC analysis : ', 'Skipped')
        parameterList_use = c(parameterList_use, ambient_status, qc_skipped)
        parameterList_use_sc@assays$user_parameter_tracker = parameterList_use
        DefaultAssay(parameterList_use_sc) = "RNA"
        Idents(parameterList_use_sc) <- "orig.ident"
        postQCSeuratObject(parameterList_use_sc)
      }
      # Remove post-QC plots from UI
      output$postQCPlots <- NULL  
      output$postQCPlot <- NULL   
      output$violinPlotPostQC <- NULL  
      shinyalert("Skip QC", "No QC was performed", type = "info", confirmButtonCol = "#1c82e0")  # Notify user that QC was skipped
    })
    
    observeEvent(input$nFeatureHigh, {
      req(preQCSeuratObject(), seuratObject(), input$nFeatureLow, input$percentMt, input$ribofrac)
      if((is.null(postQCSeuratObject())) && (!is.null(preQCSeuratObject())) && (!is.null(seuratObject()))) {
        if("DecontX_contamination" %in% colnames(seuratObject()@meta.data)) {
          parameterList_use_sc = seuratObject()
          parameterList_use = parameterList_use_sc@assays$user_parameter_tracker
          qc_skipped = paste0('QC analysis : ', 'Skipped')
          parameterList_use = c(parameterList_use, qc_skipped)
          parameterList_use_sc@assays$user_parameter_tracker = parameterList_use
          DefaultAssay(parameterList_use_sc) = "RNA"
          Idents(parameterList_use_sc) <- "orig.ident"
          postQCSeuratObject(parameterList_use_sc)
        } else {
          parameterList_use_sc = seuratObject()
          parameterList_use = parameterList_use_sc@assays$user_parameter_tracker
          ambient_status = paste0('Ambient RNA correction: ', 'Skipped')
          qc_skipped = paste0('QC analysis : ', 'Skipped')
          parameterList_use = c(parameterList_use, ambient_status, qc_skipped)
          parameterList_use_sc@assays$user_parameter_tracker = parameterList_use
          DefaultAssay(parameterList_use_sc) = "RNA"
          Idents(parameterList_use_sc) <- "orig.ident"
          postQCSeuratObject(parameterList_use_sc)
        }
      }
    })
    return(postQCSeuratObject)
  })
}