#' Server logic for Lasso selection & DE Analysis
#' 
#' Implements server side functionality for UMAP lasso selection. It includes handling
#' of lasso selection on the UMAP plot and differential gene expression for selected points
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#'
#' @examples 
#' umapLassoSelectionServer("umapLassoSelectionModule", seuratObjectQC)
#' Server logic for Lasso selection & DE Analysis
#' 
#' Implements server side functionality for UMAP lasso selection. It includes handling
#' of lasso selection on the UMAP plot and differential gene expression for selected points
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#' @param seuratObj A reactive expression containing the Seurat object to be processed.
#'
#' @return
#' @export
#'
#' @examples 
#' umapLassoSelectionServer("umapLassoSelectionModule", seuratObjectQC)
umapLassoSelectionServer <- function(id, seuratObj) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns
    
    selected_data <- reactiveVal(data.frame())
    combined_selection <- reactiveVal(NULL)
    
    
    # Define seurat copy
    LassoSeuratObj <- reactiveVal(NULL)
    
    observe({
      LassoSeuratObj(seuratObj())
    })
  
    observe({
      req(LassoSeuratObj())
      if (!is.null(input$Selectionchoice) && !(input$Selectionchoice %in% c("Clusters", "Cell Types"))) {
        updateSelectInput(session, "Selectionchoice", choices = c("Clusters", "Cell Types"))
      }
    })
    
    observeEvent(input$Selectionchoice, {
      req(LassoSeuratObj())
      seurat_object <- LassoSeuratObj()
      combined_selection(NULL)
      selected_data(data.frame())
    })
    
    observeEvent(input$Selectionchoice, {
      req(LassoSeuratObj())
      combined_selection(NULL)
      selected_data(data.frame())
      
      # Reset custom clusters or cell types to their original values
      seurat_object <- LassoSeuratObj()
      seurat_object@meta.data$custom_clusters <- as.character(seurat_object@meta.data$seurat_clusters)
      if ("SingleR_Celltypes" %in% colnames(seurat_object@meta.data)) {
        seurat_object@meta.data$custom_cell_types <- as.character(seurat_object@meta.data$SingleR_Celltype)
      }
      LassoSeuratObj(seurat_object)
    })
    
    output$reductionPlot <- renderPlotly({
      req(LassoSeuratObj())
      
      if (input$Selectionchoice == "Clusters" && !"seurat_clusters" %in% colnames(LassoSeuratObj()@meta.data)) {
        shinyalert("Error", "Clusters not available. Please run clustering module first.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      if (input$Selectionchoice == "Cell Types" && !"SingleR_Celltypes" %in% colnames(LassoSeuratObj()@meta.data)) {
        shinyalert("Error", "Cell type data not available. Please run cell type annotation first.", type = "error", confirmButtonCol = "#1c82e0")
        return(NULL)
      }
      
      seurat_object <- LassoSeuratObj()
      
      if (!"custom_clusters" %in% colnames(seurat_object@meta.data)) {
        seurat_object@meta.data$custom_clusters <- as.character(seurat_object@meta.data$seurat_clusters)
      }
      
      if ("SingleR_Celltypes" %in% colnames(seurat_object@meta.data) && !"custom_cell_types" %in% colnames(seurat_object@meta.data)) {
        seurat_object@meta.data$custom_cell_types <- as.character(seurat_object@meta.data$SingleR_Celltypes)
      }
      
      LassoSeuratObj(seurat_object)
      
      group.by <- if (input$Selectionchoice == "Clusters") {
        "custom_clusters"
      } else {
        "custom_cell_types"
      }
      
      plot <- DimPlot(LassoSeuratObj(), group.by = group.by, label = FALSE, label.size = 5) +
        RotatedAxis() + guides(colour = guide_legend(ncol = 1, override.aes = list(size = 5))) +
        theme(aspect.ratio = 1, axis.text = element_text(colour = "black", size = 15, face = "bold"),
              axis.title = element_text(colour = "black", size = 15, face = "bold"), legend.position = "right",
              legend.box = "vertical", panel.background = element_rect(colour = "white")) + ggtitle("")
      
      ggplotly(plot, source = "lassoplot") %>% layout(dragmode = 'lasso')
    })
    
    observeEvent(event_data("plotly_selected", source = "lassoplot"), {
      req(LassoSeuratObj())

      if("umap" %in% Reductions(LassoSeuratObj())) {
        embedding <- data.frame(LassoSeuratObj()[["umap"]]@cell.embeddings)
        embedding$original_index <- 1:nrow(embedding)
      }
      if("tsne" %in% Reductions(LassoSeuratObj())) {
        embedding <- data.frame(LassoSeuratObj()[["tsne"]]@cell.embeddings)
        embedding$original_index <- 1:nrow(embedding)
      }
      selected_points <- event_data("plotly_selected", source = "lassoplot")
      if (!is.null(selected_points) && length(selected_points) > 0 && !is.null(selected_points$x) && !is.null(selected_points$y)) {
        selected_coords <- selected_points[c("x", "y")]
        
        tolerance <- 1e-5
        
        matched_indices <- lapply(1:nrow(selected_coords), function(i) {
          which(abs(embedding[, 1] - selected_coords$x[i]) < tolerance & 
                  abs(embedding[, 2] - selected_coords$y[i]) < tolerance)
        })
        
        matched_indices <- unlist(matched_indices)
        
        if (length(matched_indices) == 0) {
          shinyalert("No points selected", "Please select points on the plot.", type = "warning", confirmButtonCol = "#1c82e0")
        } else {
          valid_indices <- embedding$original_index[matched_indices]
          
          combined_selection(valid_indices)
          
          # Show modal for naming the selection
          showModal(modalDialog(
            title = "Save Lasso Selection",
            textInput(ns("lassoNameModal"), "Name your selection", ""),
            footer = tagList(
              modalButton("Cancel"),
              actionButton(ns("saveLassoSelectionModal"), "Save", style='color:#fff; background-color:#1c82e0')
            )
          ))
        }
      } else {
        shinyalert("No points selected", "Please select points on the plot.", type = "warning", confirmButtonCol = "#1c82e0")
      }
    })
    
    observeEvent(input$saveLassoSelectionModal, {
      req(combined_selection())
      lassoName <- input$lassoNameModal
      
      if (lassoName == "") {
        shinyalert("Error", "Please provide a name for the lasso selection.", type = "error", confirmButtonCol = "#1c82e0")
      } else {
        seurat_object <- LassoSeuratObj()
        
        if (input$Selectionchoice == "Clusters") {
          seurat_object@meta.data$custom_clusters[combined_selection()] <- lassoName
        } else if (input$Selectionchoice == "Cell Types") {
          seurat_object@meta.data$custom_cell_types[combined_selection()] <- lassoName
        }
        
        LassoSeuratObj(seurat_object) # Update the object
        seuratObj(LassoSeuratObj()) # update seurat object
        
        shinyalert("Success", paste("Lasso selection saved as", lassoName), type = "success", confirmButtonCol = "#1c82e0")
        
        output$downloadUI <- renderUI({
          if (input$saveLassoSelectionModal > 0){
            downloadButton(ns("downloadMetadata"), "Download metadata", style='color:#fff; background-color:#1c82e0')
          }
        })
        
        output$downloadMetadata <- downloadHandler(
          filename = function() {
            paste("CogentDS_v2.0_metadata_", Sys.Date(), ".csv", sep="")
          },
          content = function(file) {
            write.csv(seurat_object@meta.data, file, row.names = TRUE)
          }
        )
      
        # Remove modal after saving
        removeModal()
      }
    })
  })
}
