#' Upload Object Server Component
#'  
#' This server module handles the server-side logic for seurat v5 processed object upload
#'
#' @param id Unique identifier for the server module, ensuring isolated scope
#'           within the Shiny application.
#'
#' @return A reactive value containing the Seurat object uploaded by the user
#'         
#' Detailed Parameter Description:
#' - object: A processed seurat v5 object either using CogentDS v2.0 or self analysis
#'
#' @seealso \code{\link[Seurat]{CreateSeuratObject}}, \code{\link[Seurat]{PercentageFeatureSet}}
#' for details on Seurat object creation and estimation of % of all counts that belong to set of features. 
#'         
#' @export
#'
#' @examples
#' # In the server definition of a Shiny app
#' seuratObject_processed <- uploadObjectUI("uploadObjectModule")


uploadObjectServer <- function(id) {
  shiny::moduleServer(id, function(input, output, session) {
    ns <- session$ns
    data <- reactiveVal(NULL)
    
    shiny::observeEvent(input$uploadobject, {
      req(input$seurat_obj)
      tryCatch({
        timeStart <- Sys.time()
        
        if (grepl(".rda$", input$seurat_obj$name)) {
          seuratObject_processed = load(input$seurat_obj$datapath)
        } else if (grepl(".rds$", input$seurat_obj$name)) {        ### added the .rds acceptance
          seuratObject_processed <- readRDS(input$seurat_obj$datapath)
        } else {
          stop("Unsupported file type")
        }
        
        if (!is.null(seuratObject_processed)) {
          if(("umap" %in% Reductions(seuratObject_processed)) || ("tsne" %in% Reductions(seuratObject_processed))) {
            output$seuratupload <- DT::renderDataTable({
              avg_nfeature = cbind('Average nFeature_RNA', round(mean(seuratObject_processed@meta.data$nFeature_RNA, na.rm = T)))
              avg_nCount = cbind('Average nCount_RNA', round(mean(seuratObject_processed@meta.data$nCount_RNA, na.rm = T)))
              avg_percentmt = cbind('Average percent-mt', round(mean(seuratObject_processed@meta.data$percent.mt, na.rm = T)))
              numberofcells = cbind('Number of Cells', ncol(seuratObject_processed))
              numberoffeatures = cbind('Number of Features', nrow(seuratObject_processed))
              df_show = rbind(numberofcells, numberoffeatures, avg_nfeature, avg_nCount, avg_percentmt)
              df_show = data.frame(df_show)
              colnames(df_show) = c("Seurat Object Stats", "Value")
              df_show$Value <- as.numeric(df_show$Value)
              DT::datatable(df_show, options = list(pageLength = 20, dom = "t"))
            })
            
            
            shiny::observeEvent(input$reset_btn, {
              if (!is.null(seuratObject_processed)) {
                if(("umap" %in% Reductions(seuratObject_processed)) || ("tsne" %in% Reductions(seuratObject_processed))) {
                  output$seuratupload <- DT::renderDataTable({
                    avg_nfeature = cbind('Average nFeature_RNA', round(mean(seuratObject_processed@meta.data$nFeature_RNA, na.rm = T)))
                    avg_nCount = cbind('Average nCount_RNA', round(mean(seuratObject_processed@meta.data$nCount_RNA, na.rm = T)))
                    avg_percentmt = cbind('Average percent-mt', round(mean(seuratObject_processed@meta.data$percent.mt, na.rm = T)))
                    numberofcells = cbind('Number of Cells', ncol(seuratObject_processed))
                    numberoffeatures = cbind('Number of Features', nrow(seuratObject_processed))
                    df_show = rbind(numberofcells, numberoffeatures, avg_nfeature, avg_nCount, avg_percentmt)
                    df_show = data.frame(df_show)
                    colnames(df_show) = c("Seurat Object Stats", "Value")
                    df_show$Value <- as.numeric(df_show$Value)
                    DT::datatable(df_show, options = list(pageLength = 20, dom = "t"))
                  })
                }
              }
              
            })
            
            data(seuratObject_processed) # Update reactive value
            
            timeEnd <- Sys.time()
            elapsedTime <- difftime(timeEnd, timeStart, units = 'mins')
            
            ####### Added Shiny alert instead of the notification!
            shinyalert("Seurat v5 Object Successfully Loaded", paste("Number of cell:", ncol(seuratObject_processed), "\n" ,"Number of genes:", nrow(seuratObject_processed), "\n" ,"Let's Get Started with Visualization"), type = "success", confirmButtonCol = "#1c82e0")
          } else {
            shinyalert("Dimension Reductions - UMAP or tSNE not found", "Please provide processed seurat object analzed with CogentDS v2.0 or RunUMAP / RunTSNE should be completed", type = "error", confirmButtonCol = "#1c82e0")
            seuratObject_processed = NULL
            stop("dimension reductions not found!")
          }
        }
      }, error = function(e) {
        shinyalert("Seurat v5 Object Load Failed", "Please check your inputs and parameters", type = "error", confirmButtonCol = "#1c82e0")
      })
    })
    return(data)
  })
}